/** @file debughelp.h
 *
 * Hilfs-Konstruktionen zum Debugging und Profiling
 *
 * $Date: 2020-08-13 11:43:34 +0200 (Do, 13 Aug 2020) $
 * $Revision: 69 $
 *
 * @date 07.03.2018 @author Nicolas */

#ifndef DEBUGHELP_H_INCLUDED
    #define DEBUGHELP_H_INCLUDED

    #include <stdio.h>      /* Printf-Debugging */
    #include "io.h"         /* Pinwackeln */
    #include "hw_config.h"
    #include "rect.h"
    #include "color.h"


    /** Debug-Output (Rechtecke/Boxen), Debug-Level */
    enum debugoutput_e
    {
        debugoutput_none  = 0x0,
        debugoutput_boxes = 0x01,
        debugoutput_all   = 0xFF,
    };

    void debugoutput_enable(enum debugoutput_e c);
    void debugoutput_disable(enum debugoutput_e c);
    ufast8_t debugoutput_getstatus(void);

    void debugoutput_showBox(Rect_t BB, color_t color );


    /** Debug-Hilfsfunktion: Rechteck-Koodinaten in der Konsole ausgeben
     *
     * @param[in] R: Rechteck */
    static inline void rect_print(Rect_t R)
    {
        printf("R = {%i, %i, %i, %i}\n", R.x, R.y, R.w, R.h);
    }




	/** Profiling/Taktzyklen messen
	 *
	 * Die benoetigten Register-Adressen sind im Standard-Header nicht enthalten.
     * @{ */

    /** Zyklenzaehler-Register */
    #define DWT_CYCCNT   ((volatile uint32_t *) 0xE0001004)

    #define DWT_CONTROL  ((volatile uint32_t *) 0xE0001000)
    #define DWT_LAR      ((volatile uint32_t *) 0xE0001FB0)
    #define SCB_DEMCR    ((volatile uint32_t *) 0xE000EDFC)



    /** Taktzaehler aktivieren
     *
     * Auf einigen Plattformen wird der Taktzaehler vom Startup-Code initialisiert.
     *
     * Anschliessend kann er ausgelesen werden mit
     *   uint32_t cnt0 = *(DWT_CYCCNT); */
    static inline void cyccnt_init(void)
    {
        *DWT_LAR = 0xC5ACCE55;    /* unlock (CM7) */
        *SCB_DEMCR |= 0x01000000;
        *DWT_CYCCNT = 0;          /* reset the counter */
        *DWT_CONTROL |= 1 ;       /* enable the counter */
    }



    /**
     *
     * Die folgenden Code-Schnipsel koennen direkt an die entsprechden Quelltext-Stellen kopiert
     * werden, um Overhead durch Funktionsaufrufe zu vermeiden.
     * @{

    volatile struct
    {
        int32_t min;
        int32_t max;
    }
    Cycles = {.max = 0, .min = INT32_MAX};


    __disable_irq();
    volatile uint32_t startcyc = *DWT_CYCCNT;
    asm volatile("": : :"memory");


    asm volatile("": : :"memory");
    volatile uint32_t stopcyc = *DWT_CYCCNT;
    int32_t cycles = (int32_t) (stopcyc - startcyc);
    Cycles.max = MAX( Cycles.max, cycles );
    Cycles.min = MIN( Cycles.min, cycles );
    __enable_irq();


     @} */

#endif /* DEBUGHELP_H_INCLUDED */
