/** @file main.c
**
** \mainpage CNC Schrittzaehler myStep
**
** Hardware:
**   - Schrittzaehlermodul Rev. 0.2
**   - STM32F103CBT6 mit 8 MHz-QUarz
**   - LCD mit PCD8544 (Nokia 5510-Display-Modul)
**   - EEPROM 24C02
**
** Funktion:
**    Dekodierung von Takt-Richtungs-Signalen fuer Schrittmotorendstufen und Darstellung auf dem
**    LCD (Schrittzahl und Koordinate).
**
**
** Bedienung:
**    - Zuruecksetzen des Zaehlers ueber den Reset-Taster
**    - Einstellungs-Menue ueber einen langen Tastendruck
**
**   Last committed:     $Revision: 68 $
**   Last changed by:    $Author: NicolasQRL $
**   Last changed date:  $Date: 2020-08-13 11:28:54 +0200 (Do, 13 Aug 2020) $
**   ID:                 $Id: main.c 68 2020-08-13 09:28:54Z NicolasQRL $
**
** @date 01.07.2020 @author Nicolas
**
***************************************************************************************************/

#include "main.h"
#include "sw_config.h"
#include "debughelp.h"
#include "delay.h"
#include "encoder_soft.h"
#include "gfx_draw.h"
#include "glcd_generic.h"
#include "glcd_pcd8544.h"
#include "i2c.h"
#include "menues.h"
#include "settings.h"
#include "text_window.h"
#include "pushbutton_time.h"
#include "svn_revision.h"
#include "u8g2_fontdata.h"
#include "ui_nvram.h"
#include "userinterface.h"
#include <inttypes.h>


#if defined( STM32F10x )

#endif

#if defined( INCLUDE_MOCKUP_SDL )
    #include "mocksdl_input.h"
    #include "mocksdl_timer.h"
#endif

#if defined( INCLUDE_UNIT_TESTS )
    #include "unity_fixture.h"
    extern void RunAllTests(void);
#endif




static_inline void realMain(void);
void timer_fastloop_init(void);
void SysTick_Handler(void);
void blackScreen(void);
void mainScreen(void);




/** @name Projektweit verwendete globale Variablen
 * @{ */
uint32_t volatile counter_milliseconds = 0;

/* Ein Flankenzaehler mit 32 Bit Breite laeuft bei 200 kHz nach schon 2 h 57 min 58 und 57 s ueber */
volatile int64_t glCounter0 = 0; /**< Pulszaehler Richtungsabhaengig */
volatile int64_t glTotal0 = 0;   /**< Pulszaehler Richtungsunabhaengig */

/** Text-Puffer */
char glTextbuffer[TEXTBUFFERSIZE_BYTES];

/** Monochrom-Pixel-Puffer fuer Fonts und andere Bitmaps */
uint8_t glPixelbuffer[PIXELBUFFERSIZE_BYTES];

/** Build-Revisions-Info */
const char *buildRevisionString =
    "Rev.:  " SVN_DATE "\n"
    "Rev.:  " SVN_REVISION "\n"
    "Build: " BUILD_DATE "\n";

/** @} */



/* Zusammenstellung der entry-Funktion mit der passenden Signatur */
#if defined( STM32F10x ) || defined( STM32F4XX ) || DOXYGEN
    /** Main-Function fuer Zielplattform (Debug, Release, Modultests) */
    int main(void)
    {
        realMain();
        return 0;
    }

#elif defined( TARGET_MOCKUP_SDL )
    /** Main-Function SDL-Mockup */
    int SDL_main(int argc, char *argv[])
    {
        (void) argc;
        (void) argv;
        mocksdl_timer_init();
        realMain();
        mocksdl_timer_deinit();
        return 0;
    }

#elif defined( TARGET_UNIT_TEST_PC )
    /** Fuer Modultests auf dem PC werden nur die Modultests aufgerufen. Diese koennen mit
     * Parametern konfiguriert werden. */
    int main(int argc, const char * argv[])
    {
        if( argv[0] != NULL )
        {
            printf("arg=%s\n", argv[0]);
        }
        if( argv[1] != NULL )
        {
            printf("arg=%s\n", argv[1]);
        }

        return UnityMain(argc, argv, RunAllTests);
    }

#else
    #error "Unknow Target"

#endif



/** Main-Funktion fuer Zielplattform und PC-Mockup
 *
 * Die Main-Funktion benoetigt je nach Zielplattform eine unterschiedliche Signatur. */
static_inline void realMain(void)
{
    #if defined( STM32F4XX ) || defined( STM32F10x )

        /* Timer und Interrupt-Prioritaeten einstellen */
        SystemInit();
        SysTick_Config(SystemCoreClock/1000);
        assert( SystemCoreClock == SYSTEMCORECLOCK_HZ );

        /* SysTick-IRQ deaktivieren.
         In manchen Implementierungen der Standard Peripheral Library funktioniert
         NVIC_DisableIRQ() nicht mit SysTick_IRQn */
        uint32_t saveSysTickCtrl = SysTick->CTRL;
        SysTick->CTRL  = 0;

        /* NVIC/ISRs konfigurieren */
        NVIC_PriorityGroupConfig(0);
        NVIC_SetPriority(PendSV_IRQn, 2);
        NVIC_SetPriority(SysTick_IRQn, 1);
        NVIC_SetPriority(HIGHSPEED_TIMER_IRQn, 0);


        /* Alle Interrupts sperren
        Interrupts duerfen erst erlaubt werden, wenn das LCD und die anderen Fehlerausgabekanaele
        (Morsen) initialisiert sind, damit eventuelle Fehlermeldungen nicht verloren gehen. */
        __disable_irq();

        /* SysTick wieder einschalten (laeuft noch nicht los) */
        SysTick->CTRL = saveSysTickCtrl;

       /* Zyklenzaehler fuer Profiling aktivieren */
        cyccnt_init();


        timer_fastloop_init();

        #ifdef STM32F10x
            /* Pins des Debug-Interfaces freigeben */
            AFIO->MAPR &= ~AFIO_MAPR_SWJ_CFG; // RM0008 S. 176
            AFIO->MAPR |=  AFIO_MAPR_SWJ_CFG_2;
        #endif

    #endif



    /* GPIOs vorbereiten (Takt fuer GPIO, Takt fuer Alternate Functions etc.) */
    initGpioClock(GPIOA);
    initGpioClock(GPIOB);
    initGpioClock(GPIOC);
    initGpioClock(GPIOD);
    initGpioClock(GPIOE);



    /* Zaehler-Pins initialisieren */
    io_setInputPU(SAMPLE_GPIO, SAMPLE_PIN_STEP_PLUS);
    io_setInputPU(SAMPLE_GPIO, SAMPLE_PIN_STEP_MINUS);
    io_setInputPU(SAMPLE_GPIO, SAMPLE_PIN_DIR_PLUS);
    io_setInputPU(SAMPLE_GPIO, SAMPLE_PIN_DIR_MINUS);


    /* Initialisierung GLCD, Textfenster, Schriftdarstellung */
	glcd_init();
    textwindow_setFont(font_breadAndButter);
	textwindow_setColors((Colors_t) {.fg=1, .bg=0});
	Rect_t Window = {.x = 0, .y = 0, .w = 86, .h = 48};
	text_setWindow(Window);
	ui_setcursor(0, 0);


    /* Komplett dunkler Bildschirm. Hier Breakpoint setzen, um das Display zu vermessen. */
    blackScreen();


   /* Alle Fehler-Ausgabe-Kanaele sind aktiv -> Interrupts erlauben */
    #if defined(STM32F10x) || defined(STM32F4XX)
        __enable_irq();
    #endif

    encoder_soft_init();
    buttonTime_init();
    i2c_init(EEPROM_C24X_I2C_DEVICE);
    ui_read_eeprom();

    /* Initialisierung fertig */

    while(1)
    {
        mainScreen();
        menuSettings();
    }
}



/** Hauptbildschirm: Darstellung der Schrittdaten */
void mainScreen(void)
{

    Textwindowdump_t Dump = textwindow_dumpSettings();


    double mmPerStep = 1./(double) (1<<Settings.spmmExp);
    mmPerStep /= (double) Settings.spmmNom;
    mmPerStep *= (double) Settings.spmmDenom;


    bool loop = true;
    while( loop )
    {
        /* Auswertung lange oder kurzer Tastendruck */
        switch( ui_getKey0Action() )
        {
            case 0:
                /* Kein Tastendruck - normaler Durchlauf */
                break;

            case 1:
                /* Kurzer Tastendruck - Beide Zhler nullen
                   Das Nullen bewirkt einen nicht-atomaren schreibenden Zugriff auf die globlaen
                   Variablen. Das koennte zu Fehlern fuehren, wenn der Zaehler waehrend der
                   Bewegung genullt wird. */
                #if defined(STM32F10x) || defined(STM32F4XX)
                    __disable_irq();
                    glCounter0 = 0;
                    glTotal0 = 0;
                    __enable_irq();
                #endif

                break;

            case 2:
                /* Langer Tastendruck - Hauptprogramm beenden, Einstellungs-Menue */
                loop = false;
                break;
        }



        /* Auf die globalen Zaehler wird nicht-atomar lesend zugegriffen. Dies kann im Fall des
           Uebertrags ueber die 32-Bit-Grenze in einem einzelnen Schleifendurchlauf zu zu einem
           inplausiblen Anzeigewert fuehren. Es wird allerdings erwartet, dass niemand schnell
           genug einen fehlerhaften Wert ablesen kann. Deswegen wird auf Massnahmen verzichtet. */


        /* Eine double-Multiplikation ist immer noch ein absoluter Bruchteil des Aufwandes, den
           die Darstellung nur eines einzigen Buchstabens benoetigt. */
        double millimeters = (double) glCounter0 * mmPerStep;



        /* Darstellung der Daten */
        gfx_clear();

        const uint8_t *smallFont = font_smallnice;
        const uint8_t *largeFont = font_largedigit;
        uint8_t largeFontHeight = 16;



        /* Textfelder haben alle die Breite Null, damit nicht das ganze Feld ueberschrieben wird */
        Rect_t R = {.x = 0, .y = 0, .w = 0, .h = largeFontHeight};
        textwindow_setFont(smallFont);
        textfield(R, "count", align_left | align_middle);

        R.x = GFX_MAX_WIDTH_PX;
        textwindow_setFont(largeFont);
        snprintf(glTextbuffer, TEXTBUFFERSIZE_BYTES-1, "%" PRId64 "", glCounter0);
        textfield(R, glTextbuffer, align_right | align_middle);

        R.x = 0;
        R.y = 16;
        textwindow_setFont(smallFont);
        textfield(R, "total", align_left | align_middle);

        R.x = GFX_MAX_WIDTH_PX;
        textwindow_setFont(largeFont);
        snprintf(glTextbuffer, TEXTBUFFERSIZE_BYTES-1, "%" PRId64 "", glTotal0);
        textfield(R, glTextbuffer, align_right | align_middle);

        R.x = 0;
        R.y = 32;
        textwindow_setFont(smallFont);
        textfield(R, "mm", align_left | align_middle);

        R.x = GFX_MAX_WIDTH_PX;
        textwindow_setFont(largeFont);
        snprintf(glTextbuffer, TEXTBUFFERSIZE_BYTES-1, "%7.2f", millimeters);
        textfield(R, glTextbuffer, align_right | align_top);


        glcd_drawnow();
    }; /* Ende Hauptschleife */

    textwindow_restoreSettings(Dump);
}



/** SysTick-Handler
 *
 *  - Millisekunden-Zaehler inkrementieren
 *  - Schalter und Taster abfragen
 *
 * Der SysTick-Timer ist der langsamste zyklische Interrupt. */
void SysTick_Handler(void)
{
    encoder_soft_get(0);
    counter_milliseconds += SYSTICK_TIMER_INTERVAL_MS;
    buttonTime_poll();
}



/** Timer und ISRs fuer die schnelle Schleife initialisieren und Timer einschalten
 *
 * "Tasks"/ISRs fuer die beiden Schleifen vorbereiten */
void timer_fastloop_init(void)
{
    #if defined( TARGET_HARDWARE_STEPCNT_REV01 )

        /* Takt fr Timer 4 einschalten */
        RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, ENABLE);

        /* Timer4 konfigurieren (200kHz)  */
        RCC_ClocksTypeDef RCC_Clocks;
        RCC_GetClocksFreq(&RCC_Clocks);
        uint32_t prescaler = 1U;
        uint32_t divider = RCC_Clocks.PCLK1_Frequency/(SIGNAL_SAMPLE_RATE_HZ*prescaler)-1;

        TIM_TimeBaseInitTypeDef TIM_TimeBase_InitStructure;
        TIM_TimeBase_InitStructure.TIM_ClockDivision = TIM_CKD_DIV1;
        TIM_TimeBase_InitStructure.TIM_CounterMode = TIM_CounterMode_Up;
        TIM_TimeBase_InitStructure.TIM_Period = divider;
        TIM_TimeBase_InitStructure.TIM_Prescaler = 1;
        TIM_TimeBaseInit(TIM4, &TIM_TimeBase_InitStructure);
        TIM_ITConfig(TIM4, TIM_IT_Update, ENABLE);

        /* Interruptcontroller konfigurieren */
        NVIC_InitTypeDef NVIC_InitStructure;
        NVIC_InitStructure.NVIC_IRQChannel = TIM4_IRQn;
        NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
        NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x0F;
        NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x0F;
        NVIC_Init(&NVIC_InitStructure);

        /* Timer4 einschalten */
        TIM_Cmd(TIM4, ENABLE);

    #endif
}



/** Schnelle Schleife (Taktabtastung) */
void TIM4_IRQHandler(void)
{
    #if defined( TARGET_HARDWARE_STEPCNT_REV01 )
        TIM_ClearITPendingBit(TIM4, TIM_IT_Update);


        /* Schrittmotortreiber-Taktzaehler. Es werden beide Halbschritte gezaehlt. */
        static bool lastClock0 = 0;

        /* Alles in einem Takt samplen */
        uint16_t sample = SAMPLE_GPIO->IDR;

        /* Clock- und Dir-Signal sind aktiv, wenn der Plus-Pin positiv und der Minus-Pin Null ist */
        bool thisClock0 = (sample & SAMPLE_PIN_STEP_PLUS) && !(sample & SAMPLE_PIN_STEP_MINUS);
        bool thisDir0   = (sample & SAMPLE_PIN_DIR_PLUS)  && !(sample & SAMPLE_PIN_DIR_MINUS);

        /* Richtungslogik */
        if( lastClock0 != thisClock0 )
        {
            if( thisDir0 )
            {
                glCounter0++;
            }
            else
            {
                glCounter0--;
            }
            glTotal0++;
        }

        lastClock0 = thisClock0;
    #endif
}



/** Abort: Die Hardware besitzt/benoetigt keinen sicheren Zustand. */
void abort(void)
{
    #ifdef __MINGW64__
        __builtin_trap();
    #else
        while(1);
    #endif
}



/** Hard Fault: Die Hardware besitzt/benoetigt keinen sicheren Zustand. */
void HardFault_Handler(void)
{
    while(1);
}



/** Hilfsfunktion: Dunkler Bildschirm
 *
 * Diese Darstellung ist insbesondere geeignet, um die tatsaechlichen Abmessungen des  Grafik-LCDs
 * zu vermessen. */
void blackScreen(void)
{
    glcd_setpos(0);
    uint8_t byte = 0x55;
    for (int i = 0; i<GLCD_END; i++)
    {
        glcd_putdat(byte);
        byte = ~byte;
    }
	glcd_drawnow();
}
