/** @file userinterface.c
 *
 * Benutzerschnittstelle: Buttons, Textausgaben, etc.
 *
 * $Date: 2020-08-12 16:39:52 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 66 $
 *
 * @date 26.04.2018 @author Nicolas */

#include "userinterface.h"
#include "sw_config.h"
#include <stdarg.h>
#include <stdio.h>
#include "pushbutton_time.h"
#include "text_window.h"
#include "text_cursor.h"
#include "main.h"
#include "nstdlib.h"


#if defined STM32F4XX || defined STM32F10x
    #include "encoder_soft.h"
//    #include "encoder_hard.h"
#else
    #include "mocksdl_input.h"
    #include "encoder_mocksdl.h"
    #include "mocksdl_key.h"
#endif



/** User-Interface: Initialisierung
 *  Vorhergehende Einstellungen sichern und Konsole neu initialisieren
 *
 * @param[out] Ui: User-Interface-Objekt
 * @param[in] title: Titel (wird ignoriert)
 * @return EXIT_SUCCESS */
exit_t ui_init(Ui_t *Ui, const char *title)
{
    (void) title;

    Ui->Dump =  textwindow_dumpSettings();
    textwindow_setColors((Colors_t)  {.fg = TEXT_COLOR, .bg = BACKGROUND_COLOR});
    ui_clear();
    cursor_home();
    return EXIT_SUCCESS;
}



/** User-Interface: Alte Konsolen-Einstellungen wiederherstellen
 *
 * @param[in] Ui: Ui-Objekt */
void ui_deinit(Ui_t *Ui)
{
    textwindow_restoreSettings(Ui->Dump);
}



/** User-Interface: Text-Ausgabe (Ersatz fuer puts() )
 *
 * @param[in] s: Darzustellender String
 * @return Anzahl der darstellten Zeichen */
int ui_putstr(const char *s)
{
    return tft_putstr(s);
}



/** User-Interface: printf()-Ersatz
 *
 * @param[in] format: Format-String (siehe printf() )
 * @param[in] ... Weitere Parameter
 * @return Anzahl der dargestellten Zeichen */
int ui_printf(constflash char * format, ... )
{
    va_list args;

    int ret;
    va_start(args, format);
    ret = vsnprintf(glTextbuffer, TEXTBUFFERSIZE_BYTES, format, args);
    va_end(args);

    ui_putstr(glTextbuffer);

    return ret;
}



/** Cursor setzen (Zeile/Spalte)
 *
 * @param[in] line: Zeile [0...n]
 *            Die Anzahl der Zeilen haengt von der Zeilenhoehe und damit der eingestellten
 *            Schriftart ab.
 * @param[in] col: Spalte [0...79] */
void ui_setcursor(uint8_t line, uint8_t col)
{
    uint_fast8_t lineheight = window_getLineHeight();
    Rect_t R = text_getWindow();
    int_fast8_t nLine = R.h/lineheight;
    line = line%nLine;
    int_fast8_t nCol = 80;
    int_fast16_t columnWidth = R.w/nCol;
    Coord2_t CursorPos;
    CursorPos.x = R.x + columnWidth * col;
    CursorPos.y = R.y + lineheight * line;
    cursor_setPosition(CursorPos);
}



/** Textfenster-Inhalt loeschen */
void ui_clear(void)
{
    gfx_clear();
    ui_setcursor(0, 0);
}



/** Tastendruecke Key0 interpretieren (non-reentrant)
 *
 * Der Aktionswert (langer oder kurzer Tastendruck wird beim ersten Aufruf nach dem Loslassen
 * der Taste zurueckgegeben.
 *
 * @return 0: kein Tastendruck, 1: kurzer Tastendruck, 2: langer Tastendruck */
int ui_getKey0Action(void)
{
    static int lastButtonTime = 0;

    /* Drehgeber-Taster */
    int buttonTime = ui_getButtonTime();
    int ret = 0;

    if( (buttonTime <= 0) && lastButtonTime > 0 )
    {
        if( lastButtonTime > T_LONGKEYPRESS_MS )
        {
            /* Langer Tastendruck */
            ret = 2;
        }
        else
        {
            /* Kurzer Tastendruck */
            ret = 1;
        }
    }

    lastButtonTime = buttonTime;
    return ret;
}



/** Tastendruecke Taster + MPG-Taster interpretieren (non-reentrant)
 *
 *
 * Der Aktionswert (langer oder kurzer Tastendruck wird beim ersten Aufruf nach dem Loslassen
 * der Taste zurueckgegeben.
 *
 * @return 0: kein Tastendruck, 1: kurzer Tastendruck, 2: langer Tastendruck */
int ui_getAction(void)
{
    return ui_getKey0Action();
}



/** Inkrement abfragen: Drehgeber + MPG
 *
 * Der mechanische Drehgeber wird direkt abgefragt, das Handrad um den Faktor 4 dezimiert
 * @return Inkrement seit der letzten Aufruf der Funktion */
int ui_getIncrementFast(void)
{
    #if defined( INCLUDE_MOCKUP_SDL )
        return mocksdl_encoder_get_diff();
    #elif defined( TARGET_UNIT_TEST_PC )
        /* Roher Hack: Gibt immer 1 zurueck */
        return 1;
    #else

        /* Mit normalem Drehgeber-Inkrement verrechnen */
        return encoder_soft_get_diff(0);

    #endif
}



/** Inkrement abfragen: Nur Drehgeber
 *
 * @return Inkrement seit der letzten Aufruf der Funktion */
int ui_getIncrementSlow(void)
{
    #if defined( TARGET_MOCKUP_SDL )
        return mocksdl_encoder_get_diff();
    #elif defined( TARGET_UNIT_TEST_PC )
        /* Roher Hack: Gibt immer 1 zurueck */
        return 1;
    #else
        return encoder_soft_get_diff(0);

    #endif
}



/** Dauer eines Tastendrucks abfragen */
int ui_getButtonTime(void)
{
    #if defined( INCLUDE_MOCKUP_SDL )
        return mocksdl_buttonTime_get();

    #elif defined( PC_UNITTEST )
        /* Roher Hack: Gibt immer 1 zurueck */
        return 1;

    #else
        return buttonTime_get();

    #endif
}
