/** @file delay.h
 *
 * Delays: Hilfsdefinitionen fuer die Kompatibilitaet mit AVR-Code
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 22.12.2013 @author Nicolas  */

#ifndef DELAY_H
    #define DELAY_H

    #include <stdint.h>

    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif



    /** Hilfsfunktion: Nanosekunden in Anzahl der Prozessortakte (aufgerundet) umrechnen
     * @param[in] ns: Nanosekunden
     * @param[in] clockrate: Prozessortaktfrequenz
     * @return Ausgerundete Anzahl an Prozessortakten */
    static_inline uint32_t nanosecondToCycle(uint32_t ns, uint32_t clockrate)
    {
        /* Aufgerundete Division */
        uint64_t dividend = (uint64_t) clockrate * ns;
        uint64_t divisor = 1000000000ULL;
        uint64_t q = (dividend + divisor - 1)/divisor;

        /* Da ein Zyklus bei < 1GHz immer laenger als eine Nanosekunde dauert, ist das
        Ergebnis immer kleiner als der Eingabewert fuer Nanosekunden */
        return q;
    }



    /** Hilfsfunktion: Mikrosekunden in Anzahl der Prozessortakte (aufgerundet) umrechnen
     * @param[in] us: Mikrosekunden
     * @param[in] clockrate: Prozessortaktfrequenz
     * @return Ausgerundete Anzahl an Prozessortakten */
    static_inline uint64_t microsecondToCycle(uint32_t us, uint32_t clockrate)
    {
        /* Aufgerundete Division */
        uint64_t dividend = (uint64_t) clockrate * us;
        uint64_t divisor = 1000000ULL;
        uint64_t q = (dividend + divisor - 1)/divisor;

        return q;
    }



    /** Hilfsfunktion: Millisekunden in Anzahl der Prozessortakte (aufgerundet) umrechnen
     * @param[in] ms: Millisekunden
     * @param[in] clockrate: Prozessortaktfrequenz
     * @return Ausgerundete Anzahl an Prozessortakten */
    static_inline uint64_t millisecondToCycle(uint32_t ms, uint32_t clockrate)
    {
        /* Aufgerundete Division */
        uint64_t dividend = (uint64_t) clockrate * ms;
        uint64_t divisor = 1000ULL;
        uint64_t q = (dividend + divisor - 1)/divisor;

        return q;
    }



    #ifdef __AVR__

    	/* Beim AVR werden die Routinen des AVR-GCC verwendet */
    	#include <util/delay.h>
    	#define delay_us _delay_us
		#define delay_ms _delay_ms

	#elif defined(STM32F10x) || defined(STM32F4XX)
    	/* Ersatz fuer delay-Routinen */
    	#include <stdint.h>
        #include "sw_config.h"



        /** Hilfsfunktion: Schleife ausfuehren, um Rechenzeit zu verbrauchen
         * Verbraucht CYCLE_PER_LOOP * nLoop + LOOP_OVERHEAD Takte
         * @param[in] nLoop: Anzahl der Schleifendurchlaeufe (Compilezeit-Konstante) */
	  	static_inline void loopSomeCycles(uint64_t nLoop)
	  	{
	  	    // assert( nLoop != 0 )
            asm volatile (\
                        "1:   subs %0, %0, #1  \n\t" \
                        "     bne  1b\n\t" \
                        :  "+r" (nLoop) );
	  	}


        #if defined( DOXYGEN )
            /** @def CYCLE_PER_LOOP
             *  Anzahl der Takte pro Schleife fuer die obenstehende Hilfsfunktion */
            #define CYCLE_PER_LOOP 4

            /** @def LOOP_OVERHEAD
             * Schleifen-Overhead fuer die obenstehende Hilfsfunktion */
            #define LOOP_OVERHEAD 4   /* geraten */

        #elif SYSTEMCORECLOCK_HZ == 36000000ULL && defined( ARM_MATH_CM3 )
            #define CYCLE_PER_LOOP 4
            #define LOOP_OVERHEAD 4   /* geraten */

        #elif SYSTEMCORECLOCK_HZ == 72000000ULL && defined( ARM_MATH_CM3 )
            #define CYCLE_PER_LOOP 6
            #define LOOP_OVERHEAD 4   /* geraten */

        #elif SYSTEMCORECLOCK_HZ == 168000000ULL && defined( ARM_MATH_CM4 )

            #define CYCLE_PER_LOOP 3  /* gemessen */
            #define LOOP_OVERHEAD 4   /* geraten */

        #else
            #error "delay.h not prepared for this SYSTEMCORECLOCK value"

        #endif



        /** Millisekunden verzoegern
         * @param[in] milliseconds: Verzoegerungsdauer, Compilezeit-Konstante */
        static_inline void delay_ms(uint32_t milliseconds)
        {
            int64_t nCycle = millisecondToCycle(milliseconds, SYSTEMCORECLOCK_HZ);

            int64_t nLoop = (nCycle)/CYCLE_PER_LOOP;
            if( nLoop <= 0 )
            {
                nLoop = 1;
            }

            if( nLoop != 0 )
            {
                loopSomeCycles(nLoop);
            }
        }



        /** Mikrosekunden verzoegern
         * @param[in] microseconds: Verzoegerungsdauer, Compilezeit-Konstante */
        static_inline void delay_us(uint32_t microseconds)
        {
            int64_t nCycle = microsecondToCycle(microseconds, SYSTEMCORECLOCK_HZ);

            int64_t nLoop = (nCycle)/CYCLE_PER_LOOP;
            if( nLoop <= 0 )
            {
                nLoop = 1;
            }

            if( nLoop != 0 )
            {
                loopSomeCycles(nLoop);
            }
        }



        /** Nanosekunden verzoegern
         *
         * @param[in] nanoseconds: Verzoegerungsdauer, Compilezeit-Konstante */
        static_inline void delay_ns(uint32_t nanoseconds)
        {
            const int32_t loopOverhead = LOOP_OVERHEAD;
            int32_t nCycle = nanosecondToCycle(nanoseconds, SYSTEMCORECLOCK_HZ);
            if( nCycle == 0 )
            {
                nCycle = 1;
            }

            int32_t nLoop = (nCycle - loopOverhead)/CYCLE_PER_LOOP;
            if( nLoop < 0 )
            {
                nLoop = 0;
            }

            int32_t rem = nCycle - CYCLE_PER_LOOP * nLoop;
            if( nLoop != 0 )
            {
                loopSomeCycles(nLoop);
                rem -= loopOverhead;
            }

            switch( rem )
            {
                /* Anzahl cases min. 3 + loopOverhead */
                case 8:
                    asm volatile("nop;");
                case 7:
                    asm volatile("nop;");
                case 6:
                    asm volatile("nop;");
                case 5:
                    asm volatile("nop;");
                case 4:
                    asm volatile("nop;");
                case 3:
                    asm volatile("nop;");
                case 2:
                    asm volatile("nop;");
                case 1:
                    asm volatile("nop;");
                case 0:
                    break;
            }
        }



      #elif defined( __MINGW64__ )

          void delay_ns(uint32_t a);
	  	  void delay_us(uint32_t a);
	  	  void delay_ms(uint32_t a);

  	  #else
          #error "No MCU defined"

      #endif

      void wait_ms(uint32_t milliseconds);


#endif // DELAY_H

