/** @file encoder_soft.c
 *
 * Auslesen eines oder mehrerer Drehgeber
 *
 * Die Funktion encoder_get() pollt den Drehgeber ohne !negative!
 * Nebenwirkungen und kann beliebig oft aufgerufen werden. encoder_get_diff()
 * gibt Werte zwischen aufeinanderfolgenden Aufrufen zurueck.
 * Mit encoder_get() in einer schnellen Programmschleife oder ISR und
 * encoder_get_diff() in einer langsamen Programmschleife laesst sich eine
 * schnelle Benutzerschnittstelle realisieren.
 *
 * Funktionen fuer AVR und STM32
 *
 * Auslesen der Drehgeber mit reinen GPIO-Operationen, geeignet fuer Panasonic-Encoder mit wackligen
 * Rastpunkten auf Spur B.
 *
 * Fuer die Panasonic-Encoder als Eingabegeraet sollte der Drehgeber
 * mindestens alle 2 Millisekunden gepollt werden.
 *
 * Die Zustandtabellen werden statisch hinterlegt. Dies erlaubt das Auslesen jedes Drehgebers
 * ueber eine Integer-Variable ohne die Uebergabe eines Zeigers
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 09.07.2013 19:43:55 @author Nicolas */


#include "encoder_soft.h"
#include "io.h"
#include "pgmspace.h"
#include "myassert.h"
#include "nstdlib.h"

/* In hw_config.h muss definiert sein:
 * 	ENCx_A_GPIO, ENCx_A_Pin
 *	ENCx_B_GPIO, ENCx_B_Pin
 *   #define ENC_N
 *
 * Mit x = 0...ENC_N-1 */
#include "hw_config.h"





/** Dekodertabelle fr wackeligen Rastpunkt (halbe Aufloesung) */
constflash int_fast8_t table_panasonic[16] = {0,0,-1,0,0,0,0,1,1,0,0,0,0,-1,0,0};

/* Dekodertabelle fr normale Drehgeber (volle Aufloesung) */
constflash int_fast8_t table_fullres[16] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};



/** Merker/Zustandvariaben Software-Gray-Dekoder */
typedef struct EncoderState_s
{
    uint_fast16_t lastval;
    uint8_t lastbits;
}
EncoderState_t;

static EncoderState_t Encstat[ENC_N];



/** Drehgeber pollen und/oder auslesen
 *
 * Drehgeber pollen, Absolutwert abfragen und Statusupdate vollziehen. Die Differenz
 * zum vorherigen Aufruf ist nie groesser als 1. Haeufige Aufrufe haben keine
 * negativen Nebenwirkungen, deshalb eignet sie sich fuer Status-Updates in
 * Interrupt-Routinen und Programmschleifen gleichermassen.
 *
 * @param[in] i: Nummer des Encoders 0...ENC_N-1
 * @return: Aktueller Wert  */
uint_fast16_t encoder_soft_get(uint_fast8_t i)
{
    uint_fast8_t enc_savebits = Encstat[i].lastbits;
    uint_fast8_t enc_val = Encstat[i].lastval;


    enc_savebits = (enc_savebits << 2)  & 0x0F;

    switch( i ) {
        case 0:
            if( io_bitIsSet(ENC0_A_PIN) ){ enc_savebits |=2; }
            if( io_bitIsSet(ENC0_B_PIN) ){ enc_savebits |=1; }
            break;
        #if ENC_N > 1
        case 1:
            if( io_bitIsSet(ENC1_A_PIN) ){ enc_savebits |=2; }
            if( io_bitIsSet(ENC1_B_PIN) ){ enc_savebits |=1; }
            break;
        #endif
    }

    enc_val -= (uint_fast8_t) table_panasonic[enc_savebits];

    Encstat[i].lastbits = enc_savebits;
    Encstat[i].lastval  = enc_val;

    return enc_val;
}



/** Alle per Pins definierten Soft-Encoder initialisieren
 *
 * Keine Eingabeparameter. Alle Pinouts sind in hwconfig.h hinterlegt. */
exit_t encoder_soft_init(void)
{
    /* I/O einstellen */
    #if( ENC_N > 0 )
        #if( !defined( ENC0_A_PIN ) || !defined( ENC0_B_PIN ) )
            #error "undefined macro for ENC0"
        #endif

        io_setInputPU(ENC0_A_PIN);
        io_setInputPU(ENC0_B_PIN);

        /* Startwerte */
        Encstat[0].lastbits = 0x01;
        Encstat[0].lastval = 0x00;

        encoder_soft_get(0); // Dummy-Read
    #endif

    #if( ENC_N > 1 )
        #if( !defined( ENC1_A_PIN ) || !defined( ENC1_B_PIN ) )
            #error "undefined macro for ENC1"
        #endif

        io_setInputPU(ENC1_A_PIN);
        io_setInputPU(ENC1_B_PIN);

        /* Startwerte */
        Encstat[1].lastbits = 0x01;
        Encstat[1].lastval = 0x00;

        encoder_soft_get(1); // Dummy-Read
    #endif


    return EXIT_SUCCESS;
}



/** Differenz des Drehgeber-Standes zwischen aufeinanderfolgenden Aufrufen der Funktion auslesen.
 * NICHT re-entrant!
 *
 * @param[in] i: Nummer des Encoders 0...ENC_N-1
 * @return: Differenz zu vorherigem Aufruf */
int_fast8_t encoder_soft_get_diff(uint_fast8_t i)
{
	static uint_fast16_t enc_last;
	uint_fast16_t enc_this;
	int_fast16_t enc_diff;

	static_assert(sizeof(uint_fast8_t)==sizeof(int_fast8_t), \
			"variable must be of same size");

	enc_this = encoder_soft_get(i);
	enc_diff = (int_fast8_t) (enc_this-enc_last);
	enc_last = enc_this;

	return enc_diff;
}
