/** @file gfx_draw.h
 *
 * Darstellung von Grafikelementen
 *
 * Vermittlungsschicht zwischen zeichnenden Funktionen und Darstellungsfunktionen von TFT oder GLCD
 *
 * $Date: 2020-08-12 18:19:08 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 67 $
 *
 * @date 30.07.2020 @author Nicolas */

#ifndef GFX_DRAW_H
    #define GFX_DRAW_H

    #include "hw_config.h"
    #include "sw_config.h"

    #include "nstdint.h"
    #include "color.h"
    #include "rect.h"
    #include "nstdlib.h"



    #if defined( GLCD_FRAMEBUFFER )
        #include "glcd_generic.h"
        #include "glcd_framebuffer.h"

    #else
        //#include "tft_draw.h"

    #endif



    Rect_t gfx_cliprect(Rect_t R);
    void gfx_fillscreen(color_t color);



    #if defined( GLCD_FRAMEBUFFER )
        #define GFX_MAX_WIDTH_PX   GLCD_WIDTH_PX
        #define GFX_MAX_HEIGHT_PX  GLCD_HEIGHT_PX

    #elif defined( GLCD_NONE ) || defined( TFT_NONE )
        #define GFX_MAX_WIDTH_PX   0
        #define GFX_MAX_HEIGHT_PX  0
        #define GFX_DRAW_NONE

    #else
        #define GFX_MAX_WIDTH_PX   TFT_WIDTH_PX
        #define GFX_MAX_HEIGHT_PX  TFT_HEIGHT_PX

    #endif



    /** Bildschirmabmessungen als Rechteck
     *
     * @return (Rect_t) {0, 0, Breite, Hoehe} */
    static_inline Rect_t gfx_getDisplayDimensions(void)
    {
        return (Rect_t) {.x=0, .y=0, .w=GFX_MAX_WIDTH_PX, .h=GFX_MAX_HEIGHT_PX};
    }



    /** Bildschirm loeschen */
    static_inline void gfx_clear(void)
    {
        #if defined( GLCD_FRAMEBUFFER )
            glcd_clear();
        #else
            gfx_fillscreen(BACKGROUND_COLOR);
        #endif
    }



    /** Einzelnes Pixel auf dem LCD darstellen
     *
     * Passen die Koordinaten nicht auf das Display, wird der Pixel nicht dargestellt.
     *
     * @param[in] Pos: Pixel-Position (x, y)
     * @return EXIT_SUCCESS|EXIT_FAILURE */
    static_inline void drawPixel(Coord2_t Pos, color_t color)
    {
        #if defined( GFX_DRAW_NONE )
            (void) Pos;
            (void) color;

        #elif defined( GLCD_FRAMEBUFFER )
            glcd_frmbuf_drawPixel(Pos, color);
        #else
            if( coord_in_rect(Pos, gfx_getDisplayDimensions()) )
            {
                tft_drawPixel_noclip(Pos, color);
            }
        #endif
    }



    /** Einfarbiges, gefuelltes Rechteck auf dem LCD darstellen
     *
     * Die Rechteck-Koordinaten (Pos/Size) muessen schon normalisiert sein. Rechtecke, die ueber den
     * Display-Rand hinausgehen, werden abgeschnitten.
     *
     * @param[in] R: Rechteck (x, y, w, h).
     * @return Anzahl der gezeichneten Pixel */
    static_inline void drawRectF(Rect_t R, uint16_t color)
    {
        #if defined( GFX_DRAW_NONE )
            (void) R;
            (void) color;

        #elif defined( GLCD_FRAMEBUFFER )
            glcd_frmbuf_drawRectF(R, color);

        #else
            R = gfx_cliprect(R);
            tft_drawRectF_noclip(R, color);
        #endif
    }



    /** Wrapper @see tft_drawBitmap() oder @see glcd_drawLinearBitmap()
     *
     * @param[in] scale: Wird ignoriert bei Monochrom-GLCD*/
    static_inline void drawBitmap(const uint8_t *bitmap, Rect_t R, Colors_t Color, ufast8_t scale)
    {
        #if defined( GFX_DRAW_NONE )
            (void) bitmap;
            (void) R;
            (void) Color;
            (void) scale;

        #elif defined( GLCD_FRAMEBUFFER )
            /* Da Monochrom-GLCD ueblicherweise eine deutlich kleinere Aufloesung als TFTs besitzen,
               wird auf eine Skalierungs-Funktion verzichtet */
            (void) scale;
            glcd_drawLinearBitmap(bitmap, R, Color);
        #else
            tft_drawBitmap(bitmap, R, Color, scale);
        #endif
    }



    /** Rechteck mit Farbverlauf darstellen
     *
     * Die Rechteck-Koordinaten (Pos/Size) muessen schon normalisiert sein. Rechtecke, die ueber den
     * Display-Rand hinausgehen, werden abgeschnitten.
     *
     * Auf einem Monochrom-GLCD wird ein schwarzes Rechteck darstellt.
     *
     * @param[in] R: Position und Groesse
     * @param[in] Color: Einfaerbe-Farbkombination fg/bg
     * @param[in] dir: 0|1 vertikal|horizontal */
    static_inline void drawRectGradient(Rect_t R, Colors_t Color, ufast8_t dir)
    {
        #if defined( GFX_DRAW_NONE )
            (void) R;
            (void) Color;
            (void) dir;

        #elif defined( GLCD_FRAMEBUFFER )
            /* Ersatz: Schwarzes Rechteck */
            (void) Color;
            (void) dir;
            color_t black = 1;
            glcd_frmbuf_drawRectF(R, black);
        #else
            R = gfx_cliprect(R);
            tft_drawRectGradient_noclip(R, Color, dir);
        #endif
    }



    /** Mit Rauschen gefuelltes randloses Rechteck auf dem LCD darstellen
     *
     * Die Rechteck-Koordinaten (Pos/Size) muessen schon normalisiert sein. Rechtecke, die ueber den
     * Display-Rand hinausgehen, werden abgeschnitten.
     *
     * Auf einem Monochrom-GLCD wird ein schwarzes Rechteck darstellt.
     *
     * @param[in] R: Rechteck (x, y, w, h).
     * @return Anzahl der gezeichneten Pixel */
    static_inline void drawRectNoise(Rect_t R)
    {
        #if defined( GFX_DRAW_NONE )
            (void) R;

        #elif defined( GLCD_FRAMEBUFFER )
            /* Ersatz: Schwarzes Rechteck */
            color_t black = 1;
            glcd_frmbuf_drawRectF(R, black);

        #else
            R = gfx_cliprect(R);
            tft_drawRectNoise_noclip(R);
        #endif
    }



    /** Wrapper @see tft_drawRectTransparent() */
    static_inline void drawRectTransparent(Rect_t R, color_t color, uint8_t opacity)
    {
        #if defined( GFX_DRAW_NONE )
            (void) R;
            (void) color;
            (void) opacity;

        #elif defined( GLCD_FRAMEBUFFER )
            /* Ersatz: Schwarzes Rechteck */
            (void) color;
            (void) opacity;
            color_t black = 1;
            glcd_frmbuf_drawRectF(R, black);

        #else
            tft_drawRectTransparent(R, color, opacity);

        #endif
    }

#endif
