/** @file gfx_draw_circles.c
 *
 * Grafikfunktionen fuer GLCD und TFT: Kreise
 *
 * Die Funktionen eigenen sich fuer Farb- und Monochrom-LCD gleichermassen, da der Farbvektor
 * einfach an die Low-Level-Grafikfunktion durchgereicht wird.
 *
 * Die Darstellungsfunktionen sind ein wenig undurchsichtig geworden. Korrektur erst bei Bedarf!
 *
 * $Date: 2020-08-02 19:08:25 +0200 (So, 02 Aug 2020) $
 * $Revision: 32 $
 *
 * @date 4.11.2017 @author Nicolas */


#include "gfx_draw_circles.h"
#include "gfx_draw_lines.h"
#include "gfx_draw.h"
#include "intmath.h"
#include "myassert.h"



/* ************************************************************************************ *
 *       Gefuellte Quadranten plotten                                                   *
 * ************************************************************************************ */

/** Bresenham-Algorithmus fuer Kreise: Statusvariable */
typedef struct BresenhamCircleState_s
{
    ifast16_t f;
    ifast16_t x;
    ifast16_t y;
    ifast16_t dfdx;
    ifast16_t dfdy;
}
BresenhamCircleState_t;



BresenhamCircleState_t bresenham_circle_init( ifast16_t radius )
{
    BresenhamCircleState_t Bh;
    Bh.f = 1 - radius;
    Bh.dfdx = 0;
    Bh.dfdy = -2 * radius;
    Bh.x = 0;
    Bh.y = radius;

    return Bh;
}



bool bresenham_circle_step(BresenhamCircleState_t *Bh)
{
    assert( Bh != NULL );

    if( Bh->x < Bh->y )
    {
        if( Bh->f >= 0 )
        {
            Bh->y--;
            Bh->dfdy += 2;
            Bh->f += Bh->dfdy;
        }
        Bh->x++;
        Bh->dfdx += 2;
        Bh->f += Bh->dfdx + 1;
        return true;
    }
    else
    {
        return false;
    }
}



/* Linienabschnitt der Kreisflaeche und des Randes darstellen */
void drawAreaLine(Coord2_t X, Coord2_t X0, Colors_t Color )
{
    X0.x = X.x; // Punkt auf Basislinie
    drawLine(X0, X, Color.bg);        // Kreis ausfuellen
}



/* Hilfsfunktion: Vollstaendige Kreissegmente (Achtelkreise) nach dem
 * Bresenham-Algorithmus plotten
 * Implementierung nach Wikipedia
 * X0.x, X0.y: Kreismittelpunkt, radius: Radius
 * quadrant: bitfeld mit 1: Viertelkreisabschnitt wird geplottet, 0: Abschnitt fehlt.
 *
 * Achtung: Aufgrund des Koordinatensystems    3 | 4
 * x-> rechts, y->unten sind die Quadranten   ---+---
 * ungewohnt angeordnet:                       2 | 1
 */
void drawCircleQuadrant(Coord2_t Center, ifast16_t radius, ufast8_t quadrant, Colors_t Color)
{
    Coord2_t X0 = Center;
    Coord2_t Pc;    // Punkt auf Kreisbogen (point on circumference)
    ifast16_t y = 0;


    BresenhamCircleState_t Bh = bresenham_circle_init(radius);
    bool drawArea;
    if( Color.bg == transparent )
        drawArea = false;
    else
        drawArea = true;

    bool drawBorder;
    if( Color.fg == transparent )
        drawBorder = false;
    else
        drawBorder = true;



    // Mit do-while-Schleife werden die Polarpunkte (= Startbedingung) aus gezeichnet
    do
    {
        Coord2_t X = {Bh.x, Bh.y};
        bool isStepX = true; // Immer. Aufgrund des Aufbaus des Bresenham-Algorithmus
        bool isStepY = ( X.y != y );
        y = X.y;

        // 1. Quadrant (unten rechts)
        if( quadrant & 1U<<0 )
        {
            Pc.x = X0.x + X.x;
            Pc.y = X0.y + X.y;
            if( drawArea && isStepX ) {  drawAreaLine(Pc, X0, Color);  }// Flaeche
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie

            Pc.x = X0.x + X.y;
            Pc.y = X0.y + X.x;
            if( drawArea && isStepY ) {  drawAreaLine(Pc, X0, Color);  }// Flaeche
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie
        }

        // 2. Quadrant (unten links)
        if( quadrant & 1U<<1 )
        {
            Pc.x = X0.x - X.y;
            Pc.y = X0.y + X.x;
            if( drawArea && isStepY ) {  drawAreaLine(Pc, X0, Color);  }
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie

            Pc.x = X0.x - X.x;
            Pc.y = X0.y + X.y;
            if( drawArea && isStepX ) {  drawAreaLine(Pc, X0, Color);  }
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie
        }

        // 3. Quadrant (oben links)
        if( quadrant & 1U<<2 )
        {
            Pc.x = X0.x - X.x;
            Pc.y = X0.y - X.y;
            if( drawArea && isStepX ) {  drawAreaLine(Pc, X0, Color);  }
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie

            Pc.x = X0.x - X.y;
            Pc.y = X0.y - X.x;
            if( drawArea && isStepY ) {  drawAreaLine(Pc, X0, Color);  }
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie
        }

        // 4. Quadrant (oben rechts)
        if( quadrant & 1U<<3 )
        {
            Pc.x = X0.x + X.y;
            Pc.y = X0.y - X.x;
            if( drawArea && isStepY ) {  drawAreaLine(Pc, X0, Color);  }
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie

            Pc.x = X0.x + X.x;
            Pc.y = X0.y - X.y;
            if( drawArea && isStepX ) {  drawAreaLine(Pc, X0, Color);  }
            if( drawBorder) {  drawPixel(Pc, Color.fg);                }// Umrandungslinie
        }
    }
    while( bresenham_circle_step(&Bh) );
}





/* ************************************************************************************ *
 *       Kreise / Kreisabschnitte plotten                                               *
 * ************************************************************************************ */

/* Hilfsfunktion: Vollstaendige Kreissegmente (Achtelkreise) nach dem
 * Bresenham-Algorithmus plotten
 * Implementierung nach Wikipedia
 * x0, y0: Kreismittelpunkt, radius: Radius
 * seg: bitfeld mit 1: Achtelkreisabschnitt wird geplottet, 0: Abschnitt fehlt. */
static void drawCircleQuadrantF(int x0, int y0, int radius, uint8_t seg, color_t color)
{
    int f = 1 - radius;
    int ddF_x = 0;
    int ddF_y = -2 * radius;
    int x = 0;
    int y = radius;


    while( x < y )
    {
        if( f >= 0 )
        {
            y--;
            ddF_y += 2;
            f += ddF_y;
        }
        x++;
        ddF_x += 2;
        f += ddF_x + 1;

        if ( seg & 1U<<0 ) drawPixel((Coord2_t) {x0 - y, y0 + x}, color); // 0. Achtelkreisabschnitt
        if ( seg & 1U<<1 ) drawPixel((Coord2_t) {x0 - x, y0 + y}, color); // 1. Achtelkreisabschnitt
        if ( seg & 1U<<2 ) drawPixel((Coord2_t) {x0 + x, y0 + y}, color); // 2. Achtelkreisabschnitt
        if ( seg & 1U<<3 ) drawPixel((Coord2_t) {x0 + y, y0 + x}, color); // 3. Achtelkreisabschnitt
        if ( seg & 1U<<4 ) drawPixel((Coord2_t) {x0 + y, y0 - x}, color); // 4. Achtelkreisabschnitt
        if ( seg & 1U<<5 ) drawPixel((Coord2_t) {x0 + x, y0 - y}, color); // 5. Achtelkreisabschnitt
        if ( seg & 1U<<6 ) drawPixel((Coord2_t) {x0 - x, y0 - y}, color); // 6. Achtelkreisabschnitt
        if ( seg & 1U<<7 ) drawPixel((Coord2_t) {x0 - y, y0 - x}, color); // 7. Achtelkreisabschnitt
    }
}



/** Kreislinie plotten
 *
 * Implementierung des Bresenham-Algorithmus frei nach Wikipedia
 *
 * @param[in] center: Kreismittelpunkt (x/y)
 * @param[in] radius: Radius
 * @param[in] color: Farbe */
void drawCircle(Coord2_t center, int radius, color_t color)
{
    int x0 = center.x;
    int y0 = center.y;


    /* Luecken fuellen (Punkte an NO, SO, SW, NW muessen nicht gefuellt werden) */
    drawPixel((Coord2_t) {x0 - radius, y0}, color); /* N */
    drawPixel((Coord2_t) {x0, y0 + radius}, color); /* O */
    drawPixel((Coord2_t) {x0 + radius, y0}, color); /* S */
    drawPixel((Coord2_t) {x0, y0 - radius}, color); /* W */


    /* Vollstaendige Segmente zeichnen */
    drawCircleQuadrantF(x0, y0, radius, 0xFF, color);
}



