/** @file gfx_draw_lines.c
 *
 * Grafikfunktionen fuer GLCD und TFT: Linien-Objekte
 *
 * Die Funktionen eigenen sich fuer Farb- und Monochrom-LCD gleichermassen, da der Farbvektor
 * einfach an die Low-Level-Grafikfunktion durchgereicht wird.
 *
 * $Date: 2020-08-02 19:08:25 +0200 (So, 02 Aug 2020) $
 * $Revision: 32 $
 *
 * @date 1.11.2017 @author Nicolas */


#include "gfx_draw_lines.h"
#include "gfx_draw.h"
#include "intmath.h"
#include "coord.h"
#include "rect.h"
#include "nstdint.h"



/** Linie zeichnen
 * Die Endpunkte Pos0, Pos1 muessen sich innerhalb der Bildflaeche befinden.
 *
 * Vertikale und horizontale Linien werden unmittelbar als Rechtecke auf den Bildschirm geschrieben,
 * Linien mit beliebiger Steigung mit dem Bresenham-Algorithmus pixelweise gesetzt.
 *
 * @param[in] Pos0, Pos1: Endpunkte
 * @param[in] color: Linienfarbe */
void drawLine(Coord2_t Pos0, Coord2_t Pos1, uint16_t color )
{

    if( Pos0.x == Pos1.x && Pos0.y == Pos1.y)
    {
        /* Einzelnes Pixel
           (vertikale Linie behandelt diesen Sonderfall nicht, da Linienlaenge = Null */
        drawRectF(coordToRect(Pos0, (Coord2_t) {1, 1}), color );
    }
    else if( Pos0.x == Pos1.x )
    {
        /* Vertikale Linie */
        ifast16_t len = Pos1.y - Pos0.y;
        len += SIGN(len);
        drawRectF(coordToRect(Pos0, (Coord2_t) {1, len}), color );
    }
    else if( Pos0.y == Pos1.y )
    {
        /* Horizontale Linie */
        ifast16_t len = Pos1.x - Pos0.x;
        len += SIGN(len);
        drawRectF(coordToRect(Pos0, (Coord2_t) {len, 1 }), color);
    }
    else
    {
        /* Linien beliebiger Steigung mit dem Bresenham-Algorithmus */
        ifast16_t x0 = Pos0.x;
        ifast16_t y0 = Pos0.y;
        ifast16_t x1 = Pos1.x;
        ifast16_t y1 = Pos1.y;

        ifast16_t dx =  ABS(x1-x0), sx = x0<x1 ? 1 : -1;
        ifast16_t dy = -ABS(y1-y0), sy = y0<y1 ? 1 : -1;
        ifast16_t err = dx+dy, e2; /* error value e_xy */

        while(1)
        {
            /* drawPixel prueft selbststaendig, ob Pixel innerhalb der Bildflaeche */
            drawPixel((Coord2_t) {x0, y0}, color);

            if (x0 == x1 && y0 == y1) break;
            e2 = 2*err;
            if (e2 > dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
            if (e2 < dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
        }
    }
}



/** Hilfsfunktion: Horizontale gepunktete Linie zeichnen
 *
 * Ob diese Funktion die Darstellung wirklich gegenueber dem Bresenham beschleunigt, ist
 * zweifelhaft. Bei Speicherplatzmangel ist sie ein Loeschkandidat.
 *
 * @param[in] Pos0, Pos1: Endpunkte
 * @param[in] color: Linienfarbe
 * @return Anzahl der gezeichneten Pixel */
ufast16_t drawDottedLineHorizontal(Coord2_t Pos0, Coord2_t Pos1, uint16_t color)
{
    ufast16_t nPixel = 0;
    assert( Pos0.y == Pos1.y );
    ifast16_t x0 = MIN(Pos0.x, Pos1.x);
    ifast16_t x1 = MAX(Pos0.x, Pos1.x);
    Coord2_t Pos = { .x = x0, .y = Pos0.y, };

    while( Pos.x <= x1 )
    {
        if( !(nPixel & 1) )
        {
            drawPixel(Pos, color);
        }
        Pos.x++;
        nPixel++;
    }
    return nPixel;
}



/** Hilfsfunktion: Vertikale gepunktete Linie zeichnen
 *
 * Ob diese Funktion die Darstellung wirklich gegenueber dem Bresenham beschleunigt, ist
 * zweifelhaft. Bei Speicherplatzmangel ist sie ein Loeschkandidat.
 *
 * @param[in] Pos0, Pos1: Endpunkte
 * @param[in] color: Linienfarbe
 * @return Anzahl der gezeichneten Pixel */
ufast16_t drawDottedLineVertical(Coord2_t Pos0, Coord2_t Pos1, uint16_t color)
{
    ufast16_t nPixel = 0;
    assert( Pos0.x == Pos1.x );
    Coord2_t Pos = { .x = Pos0.x, .y = MIN(Pos0.y, Pos1.y), };
    int16_t yMax = MAX(Pos0.y, Pos1.y);

    while( Pos.y <= yMax )
    {
        if( !(nPixel & 1) )
        {
            drawPixel(Pos, color);
        }
        Pos.y++;
        nPixel++;
    }
    return nPixel;
}



/** Gepunktete Linie zeichnen
 * Die Endpunlte Pos0, Pos1 muessen sich innerhalb der Bildflaeche befinden.
 *
 * Alle Linien werden pixelweise gesetzt.
 *
 * @param[in] Pos0, Pos1: Endpunkte
 * @param[in] color: Linienfarbe */
void drawDottedLine(Coord2_t Pos0, Coord2_t Pos1, color_t color)
{
    uint_fast16_t nPixel = 0;

    if( Pos0.x == Pos1.x && Pos0.y == Pos1.y)
    {
        /* Einzelnes Pixel
           (vertikale Linie behandelt diesen Sonderfall nicht, da Linienlaenge = Null */
        drawPixel(Pos0, color );
    }
    else if( Pos0.x == Pos1.x )
    {
        drawDottedLineVertical(Pos0, Pos1, color);
    }
    else if( Pos0.y == Pos1.y )
    {
        drawDottedLineHorizontal(Pos0, Pos1, color);
    }
    else
    {
        /* Linien beliebiger Steigung mit dem Bresenham-Algorithmus */
        ifast16_t x0 = Pos0.x;
        ifast16_t y0 = Pos0.y;
        ifast16_t x1 = Pos1.x;
        ifast16_t y1 = Pos1.y;

        ifast16_t dx =  ABS(x1-x0), sx = x0<x1 ? 1 : -1;
        ifast16_t dy = -ABS(y1-y0), sy = y0<y1 ? 1 : -1;
        ifast16_t err = dx+dy, e2; /* error value e_xy */

        while(1)
        {
            // drawPixel prueft selbststaendig, ob Pixel innerhalb der Bildflaeche
            if( !(nPixel++ & 1) )
            {
                drawPixel((Coord2_t) {x0, y0}, color);
            }

            if (x0 == x1 && y0 == y1) break;
            e2 = 2*err;
            if (e2 > dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
            if (e2 < dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
        }
    }
}



