/** @file gfx_draw_rectangles.c
 *
 * Grafikfunktionen fuer TFT: Rechtecke (ohne direkten Hardware-Bezug)
 *
 * Die Funktionen eigenen sich fuer Farb- und Monochrom-LCD gleichermassen, da der Farbvektor
 * einfach an die Low-Level-Grafikfunktion durchgereicht wird.
 *
 * $Date: 2020-08-02 19:08:25 +0200 (So, 02 Aug 2020) $
 * $Revision: 32 $
 *
 * @date 04.03.2017 @author Nicolas */


#include "gfx_draw_rectangles.h"
#include "gfx_draw_circles.h"
#include "gfx_draw_lines.h"
#include "gfx_draw.h"



/** Leeres oder mit anderer Farbe gefuelltes Rechteck zeichen (Strichstaerke 1 px)
 *
 * @param[in] R: Rechteck"=Koordinaten
 * @param[in] Color: (fg: Rand, bg: Flaeche) */
exit_t drawRect(Rect_t R, Colors_t Color)
{
    R = gfx_cliprect(R);
    if( R.w < 1 || R.h < 1 )
    {
        return EXIT_FAILURE;
    }


    /* Hintergrund */
    if( Color.bg != transparent )
    {
        drawRectF(R, Color.bg);
    }

    /* Rand */
    if( Color.fg != transparent )
    {
        Rect_t Rr = R;
        Rr.h = 1;
        drawRectF(Rr, Color.fg);
        Rr.y += R.h - 1;
        drawRectF(Rr, Color.fg);
        Rr = R;
        Rr.w = 1; Rr.y += 1; Rr.h -= 2;
        drawRectF(Rr, Color.fg);
        Rr.x += R.w - 1;
        drawRectF(Rr, Color.fg);
    }

    return EXIT_SUCCESS;
}



/** Leeres oder mit anderer Farbe gefuelltes Rechteck mit gepunkteter Linie zeichen
 * (Strichstaerke 1 px)
 *
 * @param[in] R: Rechteck"=Koordinaten
 * @param[in] Color: (fg: Rand, bg: Flaeche) */
exit_t drawDottedRect(Rect_t R, Colors_t Color)
{
    R = gfx_cliprect(R);
    if( R.w < 1 || R.h < 1 )
    {
        return EXIT_FAILURE;
    }


    /* Hintergrund */
    if( Color.bg != transparent )
    {
        drawRectF(R, Color.bg);
    }

    /* Rand */
    if( Color.fg != transparent )
    {
        Coord2_t C0, C1, C2, C3;
        C0.x = R.x;
        C0.y = R.y;
        C1.x = R.x;
        C1.y = R.y+R.h-1;
        C2.x = R.x+R.w-1;
        C2.y = R.y+R.h-1;
        C3.x = R.x+R.w-1;
        C3.y = R.y;

        drawDottedLine(C0, C1, Color.fg);
        drawDottedLine(C1, C2, Color.fg);
        drawDottedLine(C2, C3, Color.fg);
        drawDottedLine(C3, C0, Color.fg);
    }

    return EXIT_SUCCESS;
}



/** Rechteck mit abgerundeten Ecken
 *
 * @param[in] R: Rechteck"=Koordinaten
 * @param[in] radius: Rundungsradius
 * @param[in] Color: (fg: Rand, bg: Flaeche) */
exit_t drawRectRounded(Rect_t R, int16_t radius, Colors_t Color)
{
    assert( radius >= 0);
    R = gfx_cliprect(R);

    int16_t minRadius = MIN(R.w, R.h)/2;
    clamp_s16(&radius, 0, minRadius);

    if( radius == 0 || radius == 1 )
    {
        drawRect(R, Color);
    }
    else
    {
        // Hintergrund - nur fuer Debug
        //drawRect(R, (Colors_t) {yellow, white});

        // Ecken
        Coord2_t CenterBr = {R.x + R.w - radius, R.y + R.h - radius};
        drawCircleQuadrant(CenterBr, radius-1, 1U<<0 , Color);

        Coord2_t CenterBl = {R.x + radius-1, R.y + R.h - radius};
        drawCircleQuadrant(CenterBl, radius-1, 1U<<1 , Color);

        Coord2_t CenterTl = {R.x + radius -1, R.y + radius -1};
        drawCircleQuadrant(CenterTl, radius-1, 1U<<2 , Color);

        Coord2_t CenterTr = {R.x + R.w - radius, R.y + radius-1};
        drawCircleQuadrant(CenterTr, radius-1, 1U<<3 , Color);

        // Eckpunkte der Kreise
        //drawPixel(CenterBr, green);
        //drawPixel(CenterBl, green);
        //drawPixel(CenterTl, green);
        //drawPixel(CenterTr, green);

        /* Flaechen */
        if( Color.bg != transparent )
        {
            Rect_t FillerM = {.x = R.x + radius, .y = R.y, .w = R.w - 2*radius, .h = R.h };
            drawRectF(FillerM, Color.bg); // Nur Hintergrund

            Rect_t FillerL = {.x = R.x, .y = R.y + radius, .w = radius, .h = R.h - 2*radius };
            drawRectF(FillerL, Color.bg); // Nur Hintergrund

            Rect_t FillerR = {.x = R.x + R.w - radius, .y = R.y + radius, .w = radius, .h = R.h - 2*radius };
            drawRectF(FillerR, Color.bg); // Nur Hintergrund
        }


        /* Gerade Raender */
        if( Color.fg != transparent )
        {
            Rect_t BorderL = {.x = R.x, .y = R.y + radius, .w = 1, .h = R.h - 2*radius };
            drawRectF(BorderL, Color.fg);

            BorderL.x += R.w - 1;
            drawRectF(BorderL, Color.fg);

            Rect_t BorderT = {.x = R.x + radius, .y = R.y, .w = R.w - 2*radius, .h = 1 };
            drawRectF(BorderT, Color.fg);

            BorderT.y += R.h - 1;
            drawRectF(BorderT, Color.fg);
        }
    }

    return EXIT_SUCCESS;
}
