/** @file rect.h
 *
 * Funktionen mit Rechtecken
 *
 * $Date: 2020-08-01 09:01:46 +0200 (Sa, 01 Aug 2020) $
 * $Revision: 27 $
 *
 * @date 9.12.2017 @author Nicolas */

#ifndef RECT_H_INCLUDED
    #define RECT_H_INCLUDED

    #include <stdint.h>
    #include "myassert.h"
    #include "intmath.h"
    #include "coord.h"

    /** Rechteckt: Koordinaten, Breite, Hoehe x <-> w, y <-> h */
    typedef struct Rect_s
    {
        int16_t x; /**< X-Koordinate */
        int16_t y; /**< Y-Koordinate */
        int16_t w; /**< Breite (in X-Richtung) */
        int16_t h; /**< Hoehe (in Y-Richtung) */
    }
    Rect_t;



    /** Kleines Rechteck
     *
     * Nur positive Breiten und Hoehen sind erlaubt */
    typedef struct Rect8_s
    {
        int8_t x; /**< X-Koordinate */
        int8_t y; /**< Y-Koordinate */
        uint8_t w; /**< Breite (in X-Richtung) */
        uint8_t h; /**< Hoehe (in Y-Richtung) */
    }
    Rect8_t;


    /** Ausrichtungen
     * Ausrichtungsn lassen sich auch mit ODER kombinieren, z.B. align_left | align_middle */
    enum rect_orientation_e
    {
        aling_none   = 0x00,    /**< Keine Ausrichtung */
        align_left   = 0x01,    /**< X-Ausrichtung links */
        align_center = 0x02,    /**< X-Ausrichtung zentriert */
        align_right  = 0x04,    /**< X-Ausrichtung rechts */
        align_top    = 0x10,    /**< Y-Ausrichtung oben */
        align_middle = 0x20,    /**< Y-Ausrichtung mitte */
        align_bottom = 0x40,    /**< Y-Ausrichtung unten */
        align_all    = 0xFF,    /**< Alles Seiten (z.B. zum Clippen) */
    };



    Rect_t rect_token(Rect_t *Rest, size_t nPixel);



    /** Rechteck normalisieren, d.h. x,y beschreibt die obere, linke Ecke, w und h sind positiv
     * @param[in] R: Rechteck
     * @return: normalisiertes Rechteck */
    static inline Rect_t rect_normalize(Rect_t R)
    {
        if( R.w < 0 )
        {
            /* Beispiel: Ein -3 Pixel breites Rechteck an der Stelle x = 3 entspricht einem 3 Pixel
               breiten Rechteck an der Stelle 1 */
            R.x += R.w + 1;
            R.w = -R.w;
        }

        if( R.h < 0 )
        {
            R.y += R.h + 1;
            R.h = -R.h;
        }

        return R;
    }



    /** Schnittmenge zweier normalisierter Rechtecke R1, R2
     * Beruehren sich die Rechtecke nicht, wird ein leeres Rechteck {0, 0, 0, 0} zurueckgegeben.
     * @param[in] R1: Rechteckt 1
     * @param[in] R2: Rechteckt 2
     * @return Schnittmenge aus R1 und R2 */
    static inline Rect_t rect_intersect(Rect_t R1, Rect_t R2)
    {
        /*Es wird davon ausgegangen, dass beide Rechtecke normalisiert sind */
        //R1 = rect_normalize(R1);
        //R2 = rect_normalize(R2);
        assert( R1.w >= 0 && R1.h >= 0 && R2.w >= 0 && R2.h >= 0 );

        /* Untere rechte Ecke */
        struct { int16_t x, y; } V1 = {R1.x + R1.w, R1.y + R1.h};
        struct { int16_t x, y; } V2 = {R2.x + R2.w, R2.y + R2.h};

        /* Pruefen, ob R1 und R2 ueberhaupt eine Schnittflaeche haben */
        if( V1.x < R2.x || /* R1 komplett links von R2 */
            V1.y < R2.y || /* R1 komplett oberhalb von R2 */
            V2.x < R1.x || /* R2 komplett links von R1 */
            V2.y < R1.y)   /* R2 komplett oberhalb von R1 */
        {
            return (Rect_t) {0, 0, 0, 0}; /* Schnittmenge leer */
        }

        /* Schnittflaeche berechnen */
        R1.x = MAX(R1.x, R2.x);
        R1.y = MAX(R1.y, R2.y);
        V1.x = MIN(V1.x, V2.x);
        V1.y = MIN(V1.y, V2.y);
        R1.w = V1.x - R1.x;
        R1.h = V1.y - R1.y;

        return R1;
    }



    /** Umhuellendes Rechteck zweier Einzel-Rechtecke berechnen
     * @param[in] R1: Rechteckt 1
     * @param[in] R2: Rechteckt 2
     * @return Rechteck um R1 und R2 */
    static inline Rect_t rect_outline(Rect_t R1, Rect_t R2)
    {
        Rect_t R =
        {
            .x = MIN(R1.x, R2.x),
            .y = MIN(R1.y, R2.y),
            .w = MAX(R1.x + R1.w, R2.x + R2.w) - MIN(R1.x, R2.x),
            .h = MAX(R1.y + R1.h, R2.y + R2.h) - MIN(R1.y, R2.y),
        };
        return R;
    }



    /** Befindet sich Koordinate innerhalb des Rechtecks?
     * Es wird davon ausgegangen, das das Rechteck normalisiert ist
     * @param[in] C: Koordinate X/Y
     * @param[in] R: Rechteck */
    static inline bool coord_in_rect(Coord2_t C, Rect_t R)
    {
        struct { int16_t x, y; } V = {R.x + R.w, R.y + R.h};
        if( C.x >= R.x && C.x < V.x && C.y >= R.y && C.y < V.y )
            return true;
        else
            return false;
    }



    /** Hilfsfunktion: Zwei Koordinaten in normalisiertes Rechteck zusammenschreiben
     * Dabei wird auch (bei Bedarf) die X-Y-Logik mit umgedreht.
     * @param[in] Pos: X/Y-Koordinate
     * @pream[in] Size: W/H-Koordinate
     * @return Resultierendes Rechteck */
    static inline Rect_t coordToRect(Coord2_t Pos, Coord2_t Size)
    {
        Rect_t R;
        R.x = Pos.x;
        R.y = Pos.y;
        R.w = Size.x;
        R.h = Size.y;
        return rect_normalize(R);
    }



    /** Hilfsfunktion: X/Y-Koordinaten aus Rechteck extrahieren
     * @param[in] R: Rechteck
     * @return Koordinate obere linke Ecke */
    static inline Coord2_t rectToPosition(Rect_t R)
    {
        return (Coord2_t) {R.x, R.y};
    }



    /** Hilfsfunktion: Breite und Hoeje aus Rechteck extrahieren
     * @param[in] R: Rechteck
     * @return Breite und Hoehe */
    static inline Coord2_t rectToSize(Rect_t R)
    {
        return (Coord2_t) {R.w, R.h};
    }



    /** Rechteck (Object) innerhalb anderen Rechtecks (Canvas) ausrichten
     * Beide Rechtecke muessen schon im Vorherein normalisiert sein.
     * @param[in] Canvas: Rahmen
     * @param[in] Object: Auszurichtendes Rechteck
     * @param[in] orientation: Ausrichtung
     * @return Ausgerichtetes Rechteck */
    static_inline Rect_t rect_align(Rect_t Canvas, Rect_t Object, enum rect_orientation_e orientation )
    {
        /* Seitliche Ausrichtung */
        if( orientation & align_center )
        {
            Object.x = Canvas.x + (Canvas.w - Object.w)/2;
        }
        else if( orientation & align_right )
        {
            Object.x = Canvas.x + (Canvas.w - Object.w);
        }
        else /* linksbuendig */
        {
            Object.x = Canvas.x;
        }

        /* Hoehenausrichtung */
        if( orientation & align_middle )
        {
            Object.y = Canvas.y + (Canvas.h - Object.h)/2;
        }
        else if( orientation & align_bottom )
        {
            Object.y = Canvas.y + (Canvas.h - Object.h);
        }
        else
        {
            Object.y = Canvas.y;
        }

        return Object;
    }



    /** Rechteck zuschneiden
     * Das Rechteck wird um `width' Pixel an einer oder mehreren Kanten abgeschnitten
     * @param[in] R: Ursprungs-Rechteckt
     * @param[in] width: Breite des abgeschnittenen Streifens
     * @param[in] side: Auswahl der abgeschnittenen Seiten (links, rechts, oben, unten, alle)
     * @return Zugeschnittenes Rechteck */
    static_inline Rect_t rect_clip(Rect_t R, ifast16_t width, enum rect_orientation_e side )
    {
        if( side & align_left )
        {
            R.x += width;
            R.w -= width;
        }

        if( side & align_right )
        {
            R.w -= width;
        }

        if( side & align_top )
        {
            R.y += width;
            R.h -= width;
        }

        if( side & align_bottom )
        {
            R.h -= width;
        }

        return R;
    }



    /** 8-Bit-Rechteck in normales 16-Bit-Rechteck umwandeln
     * Keine Einschraenkungen
     * @param[in] R: 8-Bit-Rechteck
     * @return Normales Rechteck (normalisiert) */
    static inline Rect_t rect8ToRect(Rect8_t R)
    {
        return (Rect_t) {.x = R.x, .y = R.y, .w = R.w, .h = R.h};
    }



    /** Normales 16-Bit-Rechteck in 8-Bit-Rechteck umwandeln
     * Das Rechteck wird dabei normalisiert.
     * Einschraenkungen: Groesse und Koordinate muessen in 8 Bit passen.
     * @param[in] R: 8-Bit-Rechteck
     * @return Normales Rechteck (normalisiert) */
    static inline Rect8_t rectToRect8(Rect_t R)
    {
        R = rect_normalize(R); // positive widths and heights

        /* Limitations */
        assert( INT8_MIN <= R.x && R.x <= INT8_MAX );
        assert( INT8_MIN <= R.y && R.y <= INT8_MAX );
        assert( R.w <= UINT8_MAX );
        assert( R.h <= UINT8_MAX );

        return (Rect8_t) {.x = R.x, .y = R.y, .w = R.w, .h = R.h};
    }



   /** 8-Bit-Rechtecke: Umhuellendes Rechteck zweier Einzel-Rechtecke berechnen
     * @param[in] R1: Rechteckt 1
     * @param[in] R2: Rechteckt 2
     * @return Rechteck um R1 und R2 */
    static inline Rect8_t rect8_outline(Rect8_t R1, Rect8_t R2)
    {
        /* play it safe */
        return rectToRect8(rect_outline(rect8ToRect(R1), rect8ToRect(R2)));
    }


#endif /* RECT_H_INCLUDED */
