/** @file glcd_framebuffer.c
 *
 * Frame-Buffer-Funktionalitaet aller Grafik-LCDs mit 8-Bit-Bloecken (z.B. KS0108, PCD8544 etc.)
 *
 * $Date: 2020-08-11 15:56:48 +0200 (Di, 11 Aug 2020) $
 * $Revision: 57 $
 *
 * @date 11.06.2020 @author Nicolas */


#include "glcd_framebuffer.h"
#include "glcd_generic.h"
#include "hw_config.h"
#include "rect.h"


#if defined( GLCD_FRAMEBUFFER )

    uint8_t gl_framebuffer[GLCD_END];
    uint16_t gl_pos = 0;



    /** Byte an Display zur Darstellung senden
     * @param[in] value: 8-Bit-Pixel-Block*/
    void glcd_frmbuf_putdat(uint8_t value)
    {
        gl_framebuffer[gl_pos] = value;
        gl_pos++;
        gl_pos%=GLCD_END;
    }


    void glcd_frmbuf_setpos(uint16_t pos)
    {
        gl_pos = pos % GLCD_END;
    }



    uint16_t glcd_frmbuf_getpos(void)
    {
        return gl_pos;
    }



    /** Der Inhalt des Displays wird geloescht, indem das ganze Bild mit leeren Pixeln
        ueberschrieben wird.                                    */
    void glcd_frmbuf_clear(void)
    {
        uint_fast16_t i;
        for (i = 0; i<GLCD_END; i++)
        {
            gl_framebuffer[i] = 0x00;
        }
        glcd_frmbuf_setpos(0); // Aus Kompatibilitaetsgruenden zu direktem Schreiben
    }



    /** Einzelnes Pixel im Framebuffer setzen
     *
     * Greift unmittelbar auf den Framebuffer zu.
     *
     * @param[in] R: Koordinaten x/y/h/w
     * @param[in] Color: Farbe 0|1 */
    void glcd_frmbuf_drawPixel(Coord2_t Pos, color_t color)
    {
        uint8_t *framebuffer = gl_framebuffer;

        if( Pos.x >= 0 && Pos.x < GLCD_WIDTH_PX && Pos.y >= 0 && Pos.y < GLCD_HEIGHT_PX )
        {
            uint_fast16_t adressOffset = Pos.x + Pos.y/8*GLCD_WIDTH_PX;
            uint_fast8_t bit = Pos.y%8;
            uint8_t byte = 1<<bit;

            if( color )
            {
                /* Pixel setzen */
                *(framebuffer + adressOffset) |= byte;
            }
            else
            {
                /* Pixel loeschen */
                *(framebuffer + adressOffset) &= ~byte;
            }
        }
    }



    /** Gefuelltes Rechteck im Framebuffer setzen
     *
     * Greift unmittelbar auf den Framebuffer zu. Die Implementierung ist sehr ineffizient, da jedes
     * Pixel einzeln gesetzt wird.
     *
     * @param[in] R: Koordinaten x/y/h/w
     * @param[in] Color: Farbe 0|1 */
    void glcd_frmbuf_drawRectF(Rect_t R, uint16_t color)
    {
        /* Rechteck ist normalisiert */
        assert(R.w >= 0);
        assert(R.h >= 0);
        if( R.w == 0 || R.h == 0 )
        {
            return;
        }


        uint8_t *framebuffer = gl_framebuffer;


        /* Jedes zu setzendes Pixel einzeln durchgehen, Adresse + Bit bestimmen, Bit setzen oder
           loeschen. */
        int_fast16_t x, y, xMin, xMax, yMin, yMax;
        xMin = MAX(R.x, 0);
        xMax = MIN( (R.x + R.w -1), GLCD_WIDTH_PX-1);
        yMin = MAX(R.y, 0);
        yMax = MIN( (R.y + R.h -1), GLCD_HEIGHT_PX-1);
        for( x = xMin; x <= xMax; x++ )
        {
            for( y = yMin; y <= yMax; y++ )
            {
                uint_fast16_t adressOffset = x + y/8*GLCD_WIDTH_PX;
                uint_fast8_t bit = y%8;
                uint8_t byte = 1<<bit;

                if( color )
                {
                    /* Pixel setzen */
                    *(framebuffer + adressOffset) |= byte;
                }
                else
                {
                    /* Pixel loeschen */
                    *(framebuffer + adressOffset) &= ~byte;
                }
            }
        }
    }



    /** Lineare Bitmap in den Framebuffer sortieren
     *
     * Geschrieben werden reine Bitmaps in linearer Ordnung (keine Block-Ordnung, keine Groessenangabe)
     * Die Implementierung ist sehr ineffizient. Jedes Pixel und damit jedes Bit wird einzeln kopiert
     *
     * @param[in] bitmap
     * @param[in] R: Position und Groesse
     * @param[in] Color: Einfaerbe-Farbkombination fg/bg
     * @param[in] scale: Skalierung */
    void glcd_drawLinearBitmap(const uint8_t *bitmap, Rect_t R, Colors_t Color)
    {
        /* Rechteck ist normalisiert */
        assert(R.w >=0 );
        assert(R.h >= 0);
        if( R.w == 0 || R.h == 0 )
        {
            return;
        }


        uint8_t *framebuffer = gl_framebuffer;

        /* Jedes zu setzendes Pixel einzeln durchgehen, Adresse + Bit bestimmen, Bit setzen oder
           loeschen. */
        int_fast16_t x, y, xMin, xMax, yMin, yMax;
        xMin = R.x;
        xMax = R.x + R.w -1;
        yMin = R.y;
        yMax = R.y + R.h -1;

        uint8_t srcByte = 0;
        size_t iBit = 0;

        for( x = xMin; x <= xMax; x++ )
        {
            for( y = yMin; y <= yMax; y++ )
            {
                if( !(iBit++ & 0x07) )
                {
                    srcByte = *bitmap++;
                }
                bool pixel = srcByte & 0x80;
                srcByte <<= 1;



                /* Nur Pixel darstellen, die wirklich auf der Bildschirmflaeche liegen */
                if( (x < 0) || (y < 0) || (x>=GLCD_WIDTH_PX) || (y>=GLCD_HEIGHT_PX) )
                {
                    continue;
                }

                size_t adressOffset = x + y/8*GLCD_WIDTH_PX;
                uint_fast8_t dstBit = y%8;
                uint8_t byte = 1<<dstBit;

                /* Jedes Pixel mal kurz blinken lassen (fuer Debugging) */
                /*
                *(framebuffer + adressOffset) |= byte;
                glcd_drawnow();
                delay_ms(100);
                *(framebuffer + adressOffset) &= ~byte;
                glcd_drawnow();
                delay_ms(100);
                */

                color_t color = pixel ? Color.fg : Color.bg;

                /* Endgueltig setzen oder loeschen */
                if( color )
                {
                    /* Pixel setzen */
                    *(framebuffer + adressOffset) |= byte;
                }
                else
                {
                    /* Pixel loeschen */
                    *(framebuffer + adressOffset) &= ~byte;
                }
                /* glcd_drawnow(); */
            }
        }
    }





#endif
