/** @file glcd_pcd8544.c
 *
 * Ansteuerung Grafik-LCD  mit Controller PCD8544, z.B. Displays von Nokia 3210, 3310, 5510
 *
 * Pinbezeichnungen:
 *
 * DC : Data/Command
 * CE, SCE: Chip Enable
 * SDIN, Serial Data input
 *
 * Eine direkte Ansteuerung des Displays ohne Framebuffer ist nicht (mehr) vorgesehen.
 *
 * $Date: 2020-08-11 15:56:48 +0200 (Di, 11 Aug 2020) $
 * $Revision: 57 $
 *
 * @date 08.03.2014 @author Nicolas */



#include "glcd_pcd8544.h"
#include "hw_config.h"


#if defined( GLCD_PCD8544 )

    #include "glcd_framebuffer.h"
    #include "delay.h"
    #include "io.h"
    #include "spi_bitbang.h"
    #include <stdint.h>



    /** Displaykommandos */
    enum pcd8544_cmd_e
    {
        GLCD_PCD8544_CMD_FunctionSet     = 0x20,
        GLCD_PCD8544_CMD_FSNormalMode    = 0x20,
        GLCD_PCD8544_CMD_FSPowerDown     = 0x24,
        GLCD_PCD8544_CMD_FSVertical      = 0x22,
        GLCD_PCD8544_CMD_FSExtended      = 0x21,
        GLCD_PCD8544_CMD_DisplayControl  = 0x04,
        GLCD_PCD8544_CMD_DCDisplayBlank  = 0x00,
        GLCD_PCD8544_CMD_DCDisplayNorm   = 0x0C,
        GLCD_PCD8544_CMD_DCDisplayAllOn  = 0x01,
        GLCD_PCD8544_CMD_DCDisplayInv    = 0x0D,

        /* gehen nur wenn nicht GLCD_PCD8= 544_,FSExtended */
        GLCD_PCD8544_CMD_SetYAddr        = 0x40,
        GLCD_PCD8544_CMD_SetXAddr        = 0x80,

        /* gehen nur im Extendes Command = mode, */
        GLCD_PCD8544_CMD_TempControl     = 0x04,
        GLCD_PCD8544_CMD_BiasSystem      = 0x10,
        GLCD_PCD8544_CMD_SetVop          = 0x80,
    };



    /** GLCD reseten
     *
     * GLCD reseten gemaess Datenblatt S. 16 */
    void glcd_pcd8544_reset(void)
    {
        io_clearBit(GLCD_PCD8544_RST_PIN);
        delay_ms(1);
        io_setBit(GLCD_PCD8544_RST_PIN);
        delay_ms(1);
    }



    /** LCD als Empfnger fr den SPI auswaehlen (low-aktiv -> "SS") */
    void glcd_pcd8544_aktiv(void)
    {
        io_clearBit(GLCD_PCD8544_CE_PIN);
    }



    /** LCD als Empfnger fr den SPI abwaehlen (low-aktiv) */
    void glcd_pcd8544_inaktiv(void)
    {
      io_setBit(GLCD_PCD8544_CE_PIN);
    }



    /** Dummy-Funktion: Aus dem Nokia-Display kann nicht ausgelesen werden.
     *
     * @return Dummy-Wert */
    uint8_t glcd_pcd8544_readdat(void)
    {
        return 0;
    }



    /** Zeichen oder Kommando an Display senden
     *
     * @param[in] value: Daten- oder Kommando-Byte
     * @param[in] isData: 0=Instruction 1=data */
    void glcd_pcd8544_putsth(char value, bool isData)
    {
        /* Bitbanging */
        glcd_pcd8544_aktiv();

        /* RS-Flag setzen: Data(1)/Command(0) */
        if( isData )
            io_setBit(GLCD_PCD8544_DC_PIN);
        else
            io_clearBit(GLCD_PCD8544_DC_PIN);


        /* Datenbyte auf SDIN ausgeben
           Aus dem LCD kann nicht ausgelesen werden, deswegen ist MISO-Pin undefiniert */
        spi_bitbang_msb_mode3(GLCD_PCD8544_SCLK_PIN, NO_PIN, GLCD_PCD8544_SDIN_PIN, value);
    }



    /** Kommando an Display senden
     *
     * @param[in] value: Kommando-Byte, @see pcd8544_cmd_e */
    static_inline void glcd_pcd8544_putcom(enum pcd8544_cmd_e cmd)
    {
        glcd_pcd8544_putsth(cmd, false);
    }



    /** Byte an Display zur Darstellung senden (Low Level)
     *
     * Wird in dieser Form eigentlich nie benoetigt, weil es sinnvoller ist, den kompletten
     * Framebuffer ins Display zu kopieren @see glcd_pcd8544_drawnow()
     * @param[in] value: Pixel-Bitmuster */
    void glcd_pcd8544_putdat(uint8_t value)
    {
        glcd_pcd8544_putsth(value, true);

        /* Weil das Senden von Daten den internen Positionszaehler des Displays inkrementiert,
           wird auch der lokale Positionszaehler inkrementiert. */
        gl_pos++;
        gl_pos%=GLCD_PCD8544_END;
    }



    /** Inhalt des Displays loeschen (Low Level)
     * Der Inhalt des Displays wird geloescht, indem das ganze Bild mit leeren Pixeln ueberschrieben
     * wird.                                    */
    void glcd_pcd8544_clear()
    {
        uint16_t i;
        glcd_pcd8544_setpos(0);
        for(i = 0; i<GLCD_PCD8544_END; i++)
        {
            glcd_pcd8544_putdat(0x00);
        }
    }



    /** Initialisierung: Ports vorbereiten, Display einschalten und loeschen
     *
     * Laut Datenblatt S. 15 muss bei der Initialisierung die Reihenfolge (Reset vor dem ersten
     * Kommando) genau eingehalten werden, weil sonst der Chip zerstoert werden koennte. */
    void glcd_pcd8544_init(void)
    {
        /* 100ms, Sicherheitsabstand zu Power on */
        delay_ms(100);

        /* Ports auf Ausgang (GPIO-Takt muss schon vorher eingeschaltet sein) */
        io_setOutput(GLCD_PCD8544_RST_PIN);
        io_setOutput(GLCD_PCD8544_CE_PIN);
        io_setOutput(GLCD_PCD8544_DC_PIN);

        spi_bitbang_init(GLCD_PCD8544_SCLK_PIN, NO_PIN, GLCD_PCD8544_SDIN_PIN, spi_cpol1_cpha1);


        glcd_pcd8544_reset();

        /* Display einschalten (extended Modus fuer erweiterte Einstellungen) */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_FSExtended);       /* Erweiterter Kommando-Modus */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetVop | 70);      /* Mittlerer Kontrast */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_TempControl | 2);  /* Temperaturausleich */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_BiasSystem | 3);   /* Bias 1:48 */

        /* Zurueck in normalen Modus schalten und restliche Einstellungen: */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_FSNormalMode);     /* normaler Kommandomodus */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_DCDisplayNorm);    /* nicht-invertierte Ausgabe */
        //glcd_pcd8544_putcom(GLCD_PCD8544_CMD_DCDisplayInv);   /* invertierte Ausgabe */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetXAddr | 0);
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetYAddr | 0);

        glcd_pcd8544_clear();
        gl_pos = 0;
    }



    /** Absolute Position setzen (im Displayspeicher)
     * Im Framebuffer wird die gleiche Position gesetzt.
     *
     * @param[in] pos: Zielposition [0 ... GLCD_END] */
    void glcd_pcd8544_setpos(uint16_t pos)
    {
        pos%=GLCD_PCD8544_END;

        uint8_t PosX = pos%GLCD_PCD8544_WIDTH_PX; /* x geht nach rechts */
        uint8_t PosY = pos/GLCD_PCD8544_WIDTH_PX; /* y geht nach unten */

        PosY%=GLCD_PCD8544_DISPHEIGHT;             /* erlaubter Bereich */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetXAddr|PosX);
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetYAddr|PosY);
        gl_pos = pos;
    }



    /** Framebuffer auf das Display ausgeben */
    void glcd_pcd8544_drawnow(void)
    {
        /* Die Speicheradresse wird automatisch inkrementiert */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetXAddr |0); /* X-Adresse,Reihe 0 */
        glcd_pcd8544_putcom(GLCD_PCD8544_CMD_SetYAddr |0); /* Y-Adresse,Spalte 0 */

        for( uint_fast16_t i=0; i<GLCD_PCD8544_END; i++)
        {
            glcd_pcd8544_putsth(gl_framebuffer[i], 0x01);
        }
    }



    /** Display abschalten:  Speicher mit Nullen fuellen und anschliessend abschalten */
    void glcd_pcd8544_powerdown(void)
    {
          glcd_pcd8544_clear();
          glcd_pcd8544_drawnow();
          glcd_pcd8544_putcom(GLCD_PCD8544_CMD_FSPowerDown);
    }


#endif
