/** @file io_stm32f10x.h
 *
 * GPIO-Operationen fuer STM32F10x
 *
 * Erlaubt portablen Bitbanging Treiber fuer STM32 und AVR
 * GPIOx und pin_x entstrechen den Definitionen der Standard Peripheral Library
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 22.02.2014 @author Nicolas */



#ifndef IO_STM32F10x_H
    #define IO_STM32F10x_H

	#include "bitmath.h"
    #include "stm32f10x_gpio.h"

    #ifndef STM32F10x
		#error "This include file is for STM32F10x only";
	#endif


    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif



    /** GPIO: Abkuerzung des Typnamen */
    typedef	GPIO_TypeDef GPIO_t;

    /** IOPin */
    typedef	uint16_t IOPin_t;



    /** Einzelnes oder mehrere Bits eines Ports setzen (push-pull)
     *
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske */
    static_inline void io_setBit(GPIO_t * const GPIOx, const IOPin_t pin)
    {
        GPIOx->BSRR = pin;
    }



    /** Einzelnes oder mehrere Bits eines Ports setzen (open drain)
     *
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske */
    static_inline void io_setBitOD(GPIO_t * const GPIOx, const IOPin_t pin)
    {
        GPIOx->BSRR = pin;
    }



    /** Einzelnes oder mehrere Bits eines Ports loeschen  (push-pull)
     *
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske */
    static_inline void io_clearBit(GPIO_t * const GPIOx, const IOPin_t pin)
    {
        GPIOx->BSRR = (uint32_t) pin<<16U;
    }



    /** Einzelnes oder mehrere Bits eines Ports loeschen  (open drain)
     *
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske */
    static_inline void io_clearBitOD(GPIO_t * const GPIOx, const IOPin_t pin)
    {
        GPIOx->BSRR = (uint32_t) pin<<16U;
    }



    /** Einzelnes oder mehrere Bits eines Ports umschalten (push-pull)
     *
     * Diese Funktion ist im Gegensatz zu setBit/clearBit keine atomare Operation!
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske */
    static_inline void io_toggleBit(GPIO_t * const GPIOx, const IOPin_t pin)
    {
        GPIOx->ODR ^= (uint32_t) pin;
    }



    /** Einzelnes oder mehrere Bits eines Ports setzen (push-pull)
     *
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske
     * @param[in] value 1|0 - high|low */
    static_inline void io_setBitTo(GPIO_t * const GPIOx, const IOPin_t pin, bool value)
    {
        uint16_t mask = -((uint16_t) value) & pin;
        GPIOx->BSRR = mask | ((uint32_t) pin)<<0x10U;
    }



    /** Acht zusammenhaengende Bits eines Ports setzen oder loeschen
     *
     * @param[in] GPIOx: GPIO
     * @param[in] byte  : Zielwerte
     * @param[in] offset: [0..8], Offset des Bytes innerhalb des 16-Bits-Ports */
    static_inline void
        io_setByte(GPIO_t * const GPIOx, const uint8_t byte, const uint8_t offset)
    {
        GPIOx->BSRR = (uint32_t) (((uint16_t) byte)<<offset) | (0x00FFU<<(offset+0x10U));
    }



    /** Acht zusammenhaengende Bits eines Ports lesen
     *
     * @param[in] GPIOx: GPIO
     * @param[in] offset: [0..8], Offset des Bytes innerhalb des 16-Bits-Ports
     * @return: gelesenes Byte */
    static_inline uint8_t io_getByte(GPIO_t * const GPIOx, const uint8_t offset)
    {
        return (uint8_t) (GPIOx->IDR >> offset);
    }



    /** Logischen Wert eines Bits lesen
     *
     * Liefert logische 1, wenn Bit gesetzt, aber nicht notwendigerweise 0x01.
     * Funktioniert auch mit GPIO-Pin-Paerchen-Makros, z.B. io_bisIsSet(PA0)
     *
     * @param[in] GPIOx: GPIO
     * @param[in] pin: Bitmaske
     * @return Bit-Wert */
    static_inline uint16_t io_bitIsSet(GPIO_t * const GPIOx, const IOPin_t pin)
    {
        return GPIOx->IDR & pin;
    }



    /* In- und Outputs der GPIOs per Bitmaske festlegen.
     * Die Funktion ist deutlich schneller als der Weg ueber GPIO_Init,
     * funktioniert aber aehnlich, ausser dass die anderen I/O-Register nicht
     * geloescht werden.
     * Register GPIOx->CRL und GPIOx->CRH (siehe RM0008 S. 167):
     *  0x3 Output push-pull
     *  0x4 Floating input
     *
     *  GPIOx
     *  Pin: Bitmaske zu aendernder Pins
     *  iomask: input oder output
     *    output  =  0x33333333
     *    input = 0x44444444 */
    static_inline void io_setIO(GPIO_t * GPIOx, const IOPin_t Pin, const uint32_t iomask)
    {
    	  uint8_t byte0, byte1;
    	  uint32_t mask, temp;
    	  byte0 = Pin & 0x00FF;
    	  byte1 = (Pin >> 8);

    	  if(byte0 != 0 )
          {
    		  mask = stretch4bitmask(byte0);
    		  temp = GPIOx->CRL;
    		  temp &= ~mask;
    		  temp |= (mask & iomask);
    		  GPIOx->CRL = temp;
    	  }

    	  if(byte1 != 0 )
          {
    		  mask = stretch4bitmask(byte1);
    		  temp = GPIOx->CRH;
    		  temp &= ~mask;
    		  temp |= (mask & iomask);
    		  GPIOx->CRH = temp;
    	  }
    }


    /* Fuer die AVR-Version muessen gpio_setInput und gpio_setOutput
     * getrennte Funktionen sein, da sonst selbst die entsprechend
     * geinlineten Funktionen nicht wegoptimiert werden. */



    /* Einzelne Pins eines Ports per Bitmaske als frei floatenden Eingang
     * setzen.
     *
     * GPIOx
     * Pin: Bitmaske */
    static_inline void io_setInput(GPIO_t * GPIOx, const IOPin_t Pin)
    {
        io_setIO(GPIOx, Pin, 0x44444444);
    }


    /* Einzelne Pins eines Ports per Bitmaske als Push-Pull-Ausgang setzen.
     * Die Ports-Bits werden dabei nicht beeinflusst, muessen also bei
     * Bedarf vorher oder nachher einzeln gesetzt oder geloescht werden.
     *
     * GPIOx
     * Pin: Bitmaske */
    static_inline void io_setOutput(GPIO_t * GPIOx, const IOPin_t Pin)
    {
        io_setIO(GPIOx, Pin, 0x33333333); // Entsprechend 50 MHz
    }



    /* Einzelne Pins eines Ports per Bitmaske als open-drain-Output setzen.
     *
     * GPIOx
     * Pin: Bitmaske */
    static_inline void io_setOutputOD(GPIO_t * GPIOx, const IOPin_t Pin)
    {
        io_setIO(GPIOx, Pin, 0x77777777);
    }



    /* Einzelne Pins eines Ports per Bitmaske als Eingang mit Pull-Up
     * setzen.
     *
     * GPIOx
     * Pin: Bitmaske */
    static_inline void io_setInputPU(GPIO_t * GPIOx,const IOPin_t Pin)
    {
        io_setIO(GPIOx, Pin, 0x88888888);
        GPIOx->BSRR = Pin;
    }



    /* Hilfsfunktionen fuer Clocks */
    void initGpioClock(GPIO_t * GPIOx);
    void initAfioClock(void);
    void initTimClock(TIM_TypeDef * TIMx);

#endif
