/** @file pushbutton_time.c
 *
 * Dauer eines Pin-Zustands, z.B.\ Tastendruck ermitteln.
 * Eine explizite Entprellung ist nicht notwendig.
 *
 * Poll-Funktionen lesen externen Tasterzustand aus und sollten in regelmaessigen Zeit-
 * abstaenden abgefragt werden. Get-Funktionen habe keine Nebenwirkungen.
 *
 * $Date: 2020-08-12 16:39:52 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 66 $
 *
 * @date 8.05.2018 @author Nicolas */


#include "pushbutton_time.h"
#include "nstdlib.h"
#include "myassert.h"

#include "hw_config.h"
#include "io.h"


/* TAST0_PIN muss in hw_config.h definiert sein */
#ifndef BUTTONTIME0_PIN
    #error "Undefined GPIO PIN: TAST0_PIN"
#endif


static volatile int_fast16_t pushbutton_counter = 0;


/** Initialisierung
 *
 * Pin als Eingang mit Pull-Up definieren */
void buttonTime_init(void)
{
    io_setInputPU(BUTTONTIME0_PIN);
}



/** Abfrage-Teil: Dauer des Tastendrucks in Vielfachen der Polling-Zeit
 *
 * @return > 0 : low, return < 0 : high */
int_fast16_t buttonTime_get(void)
{
    static_assert( sizeof(int) >= sizeof(int16_t) , "Unexpected integer size");
    return pushbutton_counter;
}



/** Polling-Teil
 *
 * Muss regelmaessig aufgerufen werden, z.B. im SysTick-Timer
 * Die Anzahl der Aufrufe mit dem gleichen Pin-Zustand wird akkumuliert, bis der Maximalwert
 * erreicht ist. */
void buttonTime_poll(void)
{
    int_fast16_t time = pushbutton_counter;
    bool in = io_bitIsSet(BUTTONTIME0_PIN);
    bool vz = time < 0;
    if( in ^ vz )
    {
        // Bitwechsel
        time = 0;
    }

    if( !in && (time != INT_FAST16_MAX) )
    {
        (time)++;
    }
    else if( in && (time != INT_FAST16_MIN) )
    {
        (time)--;
    }

    pushbutton_counter = time;
}


