/** @file Utest_eeprom.c
 *
 * Nutzer-interaktive Tests: Lesen und Schreiben eines externen EEPROMs 24C02 per I2C
 *
 * $Date: 2020-08-11 18:40:35 +0200 (Di, 11 Aug 2020) $
 * $Revision: 58 $
 *
 * @date 04.02.2014 @author Nicolas */


#include "Utest_eeprom.h"

#include "eeprom_24Cx.h"
#include "i2c_simple.h"
#include "hw_config.h"

#include "delay.h"
#include "gfx_draw.h"

#include "userinterface.h"
#include "text_cursor.h"




typedef Rect_t Bd_t;



/** Hifsfunktion: Low-level-Grafikfunktionen, um Bytes aus dem EEPROM als Pixelmuster auf dem LCD
    darstellen zu koennen */
Bd_t byteDraw_init(void)
{
    Coord2_t C = cursor_getPosition();
    return (Rect_t) {.x = C.x, .y = C.y, .w = 1, .h = 8};
}



/** Hifsfunktion: Low-level-Grafikfunktionen, um Bytes aus dem EEPROM als Pixelmuster auf dem LCD
    darstellen zu koennen */
static void byteDraw(Bd_t *Bd, uint8_t byte)
{
    Colors_t Color = {green, black};
    drawBitmap(&byte, *Bd, Color, 1);
    Bd->x += 1;
}



/** Test: Darstellung des gesamten EEPROM-Inhalts auf dem Grafik-LCD, aehnlich Memdump-Funktionen
 *  auf dem AVR */
void Utest_eeprom_read(void)
{
	uint8_t byte, status;

    /* User interface */
    Ui_t Uim, *Ui = &Uim;
    ui_init(Ui, "Test: read EEPROM data");
    ui_clear();
	ui_printf_P(PSTR("\fRead EEPROM data\n\a"));
	ui_setcursor(1, 0);


    Bd_t Bd = byteDraw_init();


	i2c_init(EEPROM_C24X_I2C_DEVICE);

	for(int i=0; i<EEPROM_C24X_SIZE_BYTES; i++ )
    {
		status =  ee24Cx_read_byte(i, &byte);

		if( status ) break;

        /* Daten auf LCD darstellen */
        byteDraw(&Bd, byte);
		delay_ms(1);
	}


    if( status == 0)
    {
        ui_printf_P(PSTR("\ndone\n\a"));
    }
    else
    {
        ui_printf_P(PSTR("\nCould not connect EEPROM\n"
			"byte=%i\n"
			"error code 0x%x\n"
			"press key to leave\a"), 0, status);
    }

	while(!ui_getAction());
    ui_deinit(Ui);
}



/** Test: Muster ins EEPROM byteweise schreiben - ACHTUNG! Korrumpiert EEPROM-Inhalt */
void Utest_eeprom_write(void)
{
	uint8_t byte, status;

    /* User interface */
    Ui_t Uim, *Ui = &Uim;
    ui_init(Ui, "Test: write EEPROM data");
    ui_clear();
	ui_printf_P(PSTR("\fOverwrite EEPROM data\n\a"));
	ui_setcursor(1, 0);

    Bd_t Bd = byteDraw_init();


	i2c_init(EEPROM_C24X_I2C_DEVICE);
	byte = 0xAA;

	for(int i=0; i<EEPROM_C24X_SIZE_BYTES; i++ )
    {
		byte = ~byte;
		status =  ee24Cx_write_byte(i, byte);
		if( status ) break;

        /* Daten auf LCD darstellen */
        byteDraw(&Bd, byte);

        /* Das Byteweise schreiben dauert ein Weilchen */
		delay_ms(EEPROM_C24X_WRITE_CYCLE_TIME_MS);
	}


    if( status == 0 )
    {
        ui_printf_P(PSTR("\ndone\n\a"));
    }
    else
    {
        ui_printf_P(PSTR("\nCould not connect EEPROM\nerror code 0x%x\n"\
			"press key to leave\a"), status);
    }

	while(!ui_getAction());
    ui_deinit(Ui);
}



/** Test: Darstellung des gesamten EEPROM-Inhalts auf dem Grafik-LCD, aehnlich Memdump-Funktionen
 *  auf dem AVR */
void Utest_eeprom_read_block(void)
{
	uint_fast8_t status;
	uint8_t buffer[EEPROM_C24X_SIZE_BYTES];

    /* User interface */
    Ui_t Uim, *Ui = &Uim;
    ui_init(Ui, "Test: Read EEPROM data");
    ui_clear();
	ui_printf_P(PSTR("\fRead EEPROM data block\n\a"));
	ui_setcursor(1, 0);

    Bd_t Bd = byteDraw_init();

    /* initialize I2C peripheral */
	i2c_init(EEPROM_C24X_I2C_DEVICE);

	status =  ee24Cx_read_block(buffer, 0, 3);
	status |= ee24Cx_read_block(buffer, 0, EEPROM_C24X_SIZE_BYTES);

	if( status == 0 )
    {
        /* Daten auf LCD darstellen */
        for( uint_fast16_t i=0; i<EEPROM_C24X_SIZE_BYTES; i++ )
        {
            byteDraw(&Bd, buffer[i]);
        }

        ui_printf_P(PSTR("\ndone\n\a"));
    }
    else
    {
        ui_printf_P(PSTR("\nCould not connect EEPROM\nerror code 0x%x\n"\
			"press key to leave\a"), status);

    }

	while(!ui_getAction());
    ui_deinit(Ui);
}



/** Test: Muster ins EEPROM blockweise schreiben - ACHTUNG! Korrumpiert EEPROM-Inhalt */
void Utest_eeprom_write_block(void)
{
	uint8_t status;
	uint8_t buffer[EEPROM_C24X_SIZE_BYTES];
	for ( int i=0; i<EEPROM_C24X_SIZE_BYTES; i++) buffer[i] = i;

    /* User interface */
    Ui_t Uim, *Ui = &Uim;
    ui_init(Ui, "Test: Write EEPROM data");
    ui_clear();
	ui_printf_P(PSTR("\fWrite EEPROM data block\n\a"));
	ui_setcursor(1, 0);
    ui_clear();


    /* initialize I2C peripheral */
	i2c_init(EEPROM_C24X_I2C_DEVICE);

    /* Write data */
	status =  ee24Cx_write_block(buffer, 0 ,EEPROM_C24X_SIZE_BYTES);

    if( status == 0 )
    {
        ui_printf_P(PSTR("\ndone\n\a"));
    }
    else
    {
        ui_printf_P(PSTR("\nCould not connect EEPROM\nerror code 0x%x\n"\
			"press key to leave\a"), status);
    }

	while(!ui_getAction());
    ui_deinit(Ui);
	return;
}

