/** @file eeprom_24Cx.c
 *
 * Lesen und Schreiben eines externen EEPROMs 24C02 per I2C
 *
 * $Date: 2020-08-07 16:39:11 +0200 (Fr, 07 Aug 2020) $
 * $Revision: 54 $
 *
 * @date 02.02.2014 @author Nicolas */


#include "eeprom_24Cx.h"
#include "hw_config.h"
#include <string.h>
#include "myassert.h"
#include "delay.h"
#include "i2c.h"



/** @def EEPROM_C24X_N_ADDRESS_BYTES
 * Kleine EEPROMs wie 24x02 haben nur ein Adress-Byte, grosse EEPROMs wie 24x64 benoetigen
 * zwei. Deshalb ist die Addressierung leicht unterschiedlich */
#ifndef EEPROM_C24X_N_ADDRESS_BYTES
    #error "undefined makro: EEPROM_N_ADDRESS_BYTES"

#elif(  EEPROM_C24X_N_ADDRESS_BYTES == 1 )
    /* 8-Bit-Adressen */

#elif( EEPROM_C24X_N_ADDRESS_BYTES == 2 )
    /* 16-Bit-Adressen */

#else
    #error "EEPROM: invalid number of adress bytes"
#endif



/** Einzelnes Byte aus EEPROM auslesen
 *
 * Nur getestet fuer EEPROMS vom Typ 24C02. Bei groesseren wird die Speicheradresse in die
 * I2C-Adresse eingerechnet
 *
 * @param[in] addr: Adresse
 * @param[out] Rueckgabewert
 * @return SUCCESS */
uint_fast8_t ee24Cx_read_byte(uint16_t addr, uint8_t *byte)
{
	uint_fast8_t status;


	if(addr>EEPROM_C24X_SIZE_BYTES)
    {
		return ee_invalidAddress;
	}

    uint8_t nBytes = 2;
    uint8_t buf[nBytes];
    buf[0] = addr>>8;
    buf[1] = addr & 0xFF;


	#if( EEPROM_C24X_N_ADDRESS_BYTES == 1 )
        /* 8-Bit Adressen */
        uint8_t *pBuffer = buf+1;
        nBytes -=1;
    #else
        /* 16-Bit-Adressen */
        uint8_t *pBuffer = buf;
    #endif


	/* Kommando ohne Stopbit senden */
	status = i2c_sendBuffer(EEPROM_C24X_I2C_DEVICE, EEPROM_C24X_I2C_ADDRESS, pBuffer, nBytes, noStop);
	if (status != 0)
		goto error;

	status = i2c_receiveBuffer(EEPROM_C24X_I2C_DEVICE, EEPROM_C24X_I2C_ADDRESS, byte, 1);
	if (status != 0)
		goto error;

	return status; /* = 0 */

	error:
		i2c_reset(EEPROM_C24X_I2C_DEVICE);
		return status | ee_i2cerror;
}



/** Einzelnes Byte in EEPROM schreiben
 *
 * Das Speichern eines einzelnen Bytes kann genauso lange wie das Schreiben einer
 * ganzen Page dauern (typ. 5ms).
 * Getestet fuer 24C02 und 24LC64.
 *
 * @param[in] addr: Adresse (8 oder 16 Bit)
 * @param[in] byte: Zu schreibendes Byte
 * @return Fehlercode */
uint_fast8_t ee24Cx_write_byte(uint16_t addr, uint8_t byte)
{
	uint_fast8_t status;


	if(addr>EEPROM_C24X_SIZE_BYTES)
    {
		return ee_invalidAddress;
	}

    uint8_t nBytes = 3;
    uint8_t buf[nBytes];
    buf[0] = addr>>8;
    buf[1] = addr & 0xFF;
    buf[2] = byte;


	#if( EEPROM_C24X_N_ADDRESS_BYTES == 1 )
        /* 8-Bit Adressen */
        uint8_t *pBuffer = buf+1;
        nBytes -=1;
    #else
        /* 16-Bit-Adressen */
        uint8_t *pBuffer = buf;
    #endif


	/* Kommando mit Stopbit senden */
	status = i2c_sendBuffer(EEPROM_C24X_I2C_DEVICE, EEPROM_C24X_I2C_ADDRESS, pBuffer, nBytes, stop);
	if (status != 0)
		goto error;

	return status; // = 0

	error:
		i2c_reset(EEPROM_C24X_I2C_DEVICE); /* STOP wuerde auch reichen */
		return status | ee_i2cerror;
}



/** Viele Bytes aus EEPROM auslesen
 * Nur getesten fuer EEPROMS vom Typ 24C02, bei groesseren muss die
 * Speicheradresse in die I2C-Adresse eingerechnet werden
 * size = 0 ist immer erfolgreich und fuehrt keine Busoperation aus
 *
 * @param[in] buffer: Byte-Array
 * @param[in] addr: Adresse (8 oder 16 Bit)
 * @param[in] size: Zu schreibendes Byte
 * @return Fehlercode */
uint_fast8_t ee24Cx_read_block(uint8_t *buffer, uint16_t addr, uint16_t size)
{
	uint_fast8_t status;

	if(size == 0) return ee_ok;

	/* Auf gueltige Adresse pruefen */
	if( (addr+size)>EEPROM_C24X_SIZE_BYTES )
    {
		return ee_invalidAddress;
	}

	if( (addr)>EEPROM_C24X_SIZE_BYTES )
    {
		return ee_invalidAddress;
	}

	/* Beim STM32F10x schlaegt das erste Lesen mit mehr als einem Byte direkt
	   nach der Initialisierung fehl. Deswegen wird ein einzelnes Dummy-Byte
       vorher gelesen und der Rueckgabestatus ignoriert: */
    status = ee24Cx_read_byte(0, buffer);
    (void) status;

    /* Adresse */
    uint8_t nBytes = 2;
    uint8_t buf[nBytes];
    buf[0] = addr>>8;
    buf[1] = addr & 0xFF;


	#if( EEPROM_C24X_N_ADDRESS_BYTES == 1 )
        /* 8-Bit Adressen */
        uint8_t *pBuf = buf+1;
        nBytes -=1;
    #else
        /* 16-Bit-Adressen */
        uint8_t *pBuf = buf;
    #endif


	/* Kommando ohne Stopbit senden */
	status = i2c_sendBuffer(EEPROM_C24X_I2C_DEVICE, EEPROM_C24X_I2C_ADDRESS, pBuf, nBytes, noStop);
	if (status != 0)
		goto error;

	status = i2c_receiveBuffer(EEPROM_C24X_I2C_DEVICE, EEPROM_C24X_I2C_ADDRESS, buffer, size);
	if (status != 0)
		goto error;

	return status;

	error:
	i2c_reset(EEPROM_C24X_I2C_DEVICE); /* STOP wuerde meist auch reichen */
	return status | ee_i2cerror;
}



/** Ganze Seite in EEPROM schreiben
 *
 * Nur getestet fuer EEPROMS vom Typ 24C02, bei groesseren muss die
 * Speicheradresse in die I2C-Adresse eingerechnet werden. Nur einzelne
 * Speicherseiten (16 bytes) werden unterstuetzt.
 * Wegen des malloc duerfte es eine der langsamsten und speicherintensivsten
 * Funktionen des gesamten Projekts sein.
 *
 * @param[in] buffer: Byte-Array
 * @param[in] addr: Adresse (8 oder 16 Bit)
 * @param[in] size: Zu schreibendes Byte
 * @return Fehlercode */
static uint_fast8_t ee24Cx_write_page(uint8_t *buffer, uint8_t addr, uint8_t size)
{
	uint_fast8_t status, page_start, page_end;
    size_t nBytes = size+2U;
	uint8_t bigbuffer[nBytes];

	/* Block mit Groesse Null ist immer OK und wird nicht auf den Bus geschickt */
	if(size==0) return ee_ok;

	/* Pruefen, ob der Block auf einer Seite liegt */
	page_start = addr>>4;
	page_end   = (addr + size - 1U)>>4U;
	if( !(page_start == page_end) )
    {
		return ee_invalidAddress;
	}

	/* Pruefen, ob Block gueltige Adresse ist */
	if( addr+size > EEPROM_C24X_SIZE_BYTES )
    {
		return ee_invalidAddress;
	}


	*(bigbuffer+0) = addr >> 8;
	*(bigbuffer+1) = addr & 0xFF;
	memcpy(bigbuffer+2, buffer, size);
    uint8_t offset = 0;

	#if( EEPROM_C24X_N_ADDRESS_BYTES == 1 )
        /* 8-Bit Adressen */
        offset = 1;
        nBytes -=1;
    #endif


	/* Kommando mit Stopbit senden */
	status = i2c_sendBuffer(EEPROM_C24X_I2C_DEVICE, EEPROM_C24X_I2C_ADDRESS, bigbuffer+offset,\
			nBytes,stop);


	if (status != 0)
		goto error;

	return status; /* = 0 */

	error:
		i2c_reset(EEPROM_C24X_I2C_DEVICE);
		return status | ee_i2cerror;
}



/** EEPROM blockweise beschreiben, indem passend auf die Speicherseiten verteilt
 * wird
 *
 * Ruft die sehr speicherintensive Funktion ee24Cx_write_page() auf
 *
 * @param[in] buffer: Byte-Array
 * @param[in] addr: Adresse (8 oder 16 Bit)
 * @param[in] size: Zu schreibendes Byte
 * @return Fehlercode */
uint_fast8_t ee24Cx_write_block(uint8_t *buffer, uint16_t addr, uint16_t size)
{
	uint_fast8_t status, blocksize;


	/* Block mit Groesse Null ist immer OK und wird nicht auf den Bus geschickt */
	if(size == 0)
	    return ee_ok;

	do
    {
		blocksize = 16 - (addr & 0x0F); /* Seitenende */
		if( size<blocksize ) blocksize = size;

		status = ee24Cx_write_page(buffer, addr, blocksize);
		if(status)
        {
			return status;
		}

		buffer += blocksize;
		addr += blocksize;
		size -= blocksize;

		/* Wartezeit bei mehreren Seiten noetig. */
		delay_ms(EEPROM_C24X_WRITE_CYCLE_TIME_MS);
	}
	while( blocksize );

	return status;
}



/** EEPROM: Blockweises Update
 *
 * EEPROM blockweise beschreiben, wobei vorher ueberprueft wird, ob nicht
 * schon der passende Inhalt im EEPROM hinterlegt ist.
 *
 * @param[in] buffer: Byte-Array
 * @param[in] addr: Schreib-Adresse
 * @param[in] size: Laenge des Byte-Arrays
 * @return SUCCESS */
uint_fast8_t ee24Cx_update_block(uint8_t *buffer, uint16_t addr, uint16_t size)
{
	uint8_t readbuffer[size];
	uint_fast8_t status;
	int equals; /* memcmp muss in 16-Bit-Variable speichern */

    status = ee24Cx_read_block(readbuffer, addr, size);
    if( status != ee_ok )
    {
    	return status;
    }
    else
    {
    	equals = memcmp(buffer, readbuffer, size);
    }

    if( equals == 0 )
    {
    	return status = ee_ok;
    }
    else
    {
        return ee24Cx_write_block(buffer, addr, size);
	}
}
