/** @file i2c.h
 *
 * I2C-Routinen: Initialisierung, Deinitialisierung, Puffer senden, Puffer empfangen,
 * Reset.
 *
 * Die aktuelle Struktur der I2C-Implementierung ist durch das historische Wachstum bedingt. Damals
 * schien es klug, fuer den STM32F10x und AVR Initalisierung und Rest zu trennen.
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 25.04.2018 @author Nicolas
 */

#ifndef I2C_H_INCLUDED
    #define I2C_H_INCLUDED

    #include "nstdint.h"
    #include "hw_config.h"
    #include <stdbool.h>

    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif


    /** Fehlercodes: I2C-Bus */
    typedef enum i2c_errorcode_e
    {

    	i2c_ok               = 0x0,
        i2c_noStart          = 0x1,
        i2c_startTimeout     = 0x2,
        i2c_noAck            = 0x3,
        i2c_ackTimeout       = 0x4,
        i2c_nAckTimeout      = 0x5,
        i2c_otherTimeout     = 0x6,
        i2c_busyTimeout      = 0x7,
        i2c_off              = 0x8,
        i2c_stop             = 0x9,
        i2c_error            = 0xA,
        i2c_portError        = 0xB,
        i2c_receiveTimeout0  = 0xC,
        i2c_receiveTimeout1  = 0xD,
    }
    i2c_errorcode_t;


    #ifdef AVR
        /* I2C-Konstanten wie beim STM32 fuer den AVR definieren. Da der AVR nur
           einen I2C hat wird diese Einstellung ignoriert

           Die AVR-I2C-Routinen sind in einem beklagenswerten Zustand und vermutlich
           nicht mehr funktionsfaehig, da die I2C-Teile fuer andere Plattformen in der
           Zwischenzeit mehrmals angepasst wurden, ohne dass die AVR-Variante nachgezogen
           wurde. */

        #define  I2C1 0x00
        #define  I2C2 0x00
        typedef const uint8_t I2C_TypeDef;

        #define i2c_reset_AD i2c_reset


    #elif defined(STM32F10x)
        #include "stm32f10x_gpio.h"
        #include "stm32f10x_rcc.h"
        #include "stm32f10x_i2c.h"

        #define i2c_reset_AD i2c_reset

    #elif defined(STM32F4XX)
        #include "stm32f4xx_gpio.h"

        #define i2c_reset_AD i2c_reset

    #elif defined( __MINGW64__ ) || defined( __MINGW32__ )
        // I2C-Konstanten wie beim STM32 definieren. Nur Dummys.
        #define  I2C1 0x00
        #define  I2C2 0x00
        typedef const uint8_t I2C_TypeDef;

        #define i2c_mock_reset i2c_reset

    #else
        #error "unknown MCU"
    #endif




    #ifdef I2C_BITBANG

        i2c_errorcode_t i2c_bitbang_init(I2C_TypeDef* I2Cx);
        ufast8_t i2c_bitbang_deinit(I2C_TypeDef* I2Cx);
        i2c_errorcode_t i2c_bitbang_waitbusy(I2C_TypeDef *I2Cx);
        i2c_errorcode_t I2C_bitbang_start(I2C_TypeDef *I2Cx, uint8_t address, uint8_t direction);
        i2c_errorcode_t I2C_bitbang_write(I2C_TypeDef *I2Cx, uint8_t data);
        i2c_errorcode_t I2C_bitbang_read_ack(I2C_TypeDef *I2Cx, uint8_t *data);
        i2c_errorcode_t I2C_bitbang_read_nack(I2C_TypeDef *I2Cx, uint8_t *data);
        i2c_errorcode_t I2C_bitbang_stop(I2C_TypeDef *I2Cx);

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t i2c_init(I2C_TypeDef *I2Cx)
        {
            return i2c_bitbang_init(I2Cx);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline ufast8_t i2c_deinit(I2C_TypeDef *I2Cx)
        {
            return i2c_bitbang_deinit(I2Cx);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t I2C_waitbusy(I2C_TypeDef *I2Cx)
        {
            return i2c_bitbang_waitbusy(I2Cx);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t I2C_start(I2C_TypeDef* I2Cx, uint8_t address, uint8_t direction)
        {
            return I2C_bitbang_start(I2Cx, address, direction);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t I2C_write(I2C_TypeDef* I2Cx, uint8_t data)
        {
            return I2C_bitbang_write(I2Cx, data);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t I2C_read_ack(I2C_TypeDef* I2Cx, uint8_t *data)
        {
            return I2C_bitbang_read_ack(I2Cx, data);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t I2C_read_nack(I2C_TypeDef* I2Cx, uint8_t *data)
        {
            return I2C_bitbang_read_nack(I2Cx, data);
        }

        /** Wrapper fuer Bitbang-Funktion */
        static_inline i2c_errorcode_t I2C_stop(I2C_TypeDef* I2Cx)
        {
            return I2C_bitbang_stop(I2Cx);
        }

    #else

        /* Plattformspezifische Funktionen */
        ufast8_t I2C_waitbusy(I2C_TypeDef* I2Cx);
        ufast8_t I2C_start(I2C_TypeDef* I2Cx, uint8_t address, uint8_t direction);
        ufast8_t I2C_write(I2C_TypeDef* I2Cx, uint8_t data);
        ufast8_t I2C_read_ack(I2C_TypeDef* I2Cx, uint8_t *data);
        ufast8_t I2C_read_nack(I2C_TypeDef* I2Cx, uint8_t *data);
        ufast8_t I2C_stop(I2C_TypeDef* I2Cx);

    #endif



    /* Plattform- und implementierungsabhaengige Funktionen (muessen extern implementiert sein) */

    /** Stoppbit senden ja/nein */
    enum i2c_stopbit_e {noStop, stop};

    i2c_errorcode_t i2c_init(I2C_TypeDef *I2Cx);
    ufast8_t i2c_deinit(I2C_TypeDef *I2Cx);

    i2c_errorcode_t i2c_sendBuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
       uint8_t* pBuffer, uint8_t nBuffer, enum i2c_stopbit_e stopbit);

    i2c_errorcode_t i2c_receiveBuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
  		uint8_t* pBuffer, uint8_t nByte);

    i2c_errorcode_t i2c_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr);

    ufast8_t i2c_reset(I2C_TypeDef* I2Cx);


#endif /* I2C_H_INCLUDED */
