/** @file i2c_bitbang.c
 *
 * I2C als Bitbanging ohne Verwendung der Spezial-Peripherie
 *
 * Setzt auf i2c_simple.c auf. Benoetigt keine Timer oder Aehnliches
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 01.12.2019 @author Nicolas */

#include "hw_config.h"
#include "i2c_simple.h"
#include "i2c.h"
#include "delay.h"
#include "io.h"
#include "nstdlib.h"
#include "myassert.h"


/* I2C bus specification UM102204 S. 48 */
#define I2C_BUS_FREE_TIME_US 5
#define I2C_HOLD_TIME_START_CONDITION_US 4
#define I2C_DATA_SETUP_TIME_NS 250
#define I2C_TLOW_NS 4700
#define I2C_THIGH_NS 4000



/** I2C (Bitbanging): Initialisierung
 *
 * Pins als Open-Collector-Ausgaenge konfigurieren
 * @param[in] I2Cx: Dummy
 * @return Fehlercode OK */
i2c_errorcode_t i2c_bitbang_init(I2C_TypeDef* I2Cx)
{
    (void) I2Cx;

    /* Pins zuruecksetzen und Pull-Ups aktivieren */
    io_setInputPU(I2C_SDA_PIN);
    io_setInputPU(I2C_SCL_PIN);

    /* Open-Drain-Output */
    io_setOutputOD(I2C_SDA_PIN);
    io_setOutputOD(I2C_SCL_PIN);

    /* Ruhezustand */
    io_setBitOD(I2C_SCL_PIN);
    io_setBitOD(I2C_SDA_PIN);

    return i2c_ok;
}



/** I2C (Bitbanging): Deinitialisierung
 *
 * Pins als Open-Collector-Ausgaenge konfigurieren
 * @param[in] I2Cx: Dummy */
ufast8_t i2c_bitbang_deinit(I2C_TypeDef* I2Cx)
{
    (void) I2Cx;

    io_setInputPU(I2C_SDA_PIN);
    io_setInputPU(I2C_SCL_PIN);

    return EXIT_SUCCESS;
}



/** Busy-Status abwarten, muss vor jedem Start ausgefuehrt werden
 * Die ganze Funktion ist ein Kompatibilitaets-Dummy. Beim Bitbanging sind alle vorherigen
 * Transfers beim Verlassen der Funktion abgeschlossen.
 *
 * @param I2Cx: Kompatibilitaets-Dummy */
i2c_errorcode_t i2c_bitbang_waitbusy(I2C_TypeDef *I2Cx)
{
    (void) I2Cx;
    return i2c_ok;
}



/** Ein Byte an Slave senden (allgemein)
 * @param[in] I2Cx: Dummy
 * @param[in] data: Byte
 * @return i2c_ok|i2c_ackTimeout */
i2c_errorcode_t I2C_bitbang_write(I2C_TypeDef *I2Cx, uint8_t data)
{
    (void) I2Cx;


    /* Byte senden (MSB first) */
    for( uint_fast8_t i = 0; i<8; i++ )
    {
        if( data & 0x80 )
        {
            io_setBitOD(I2C_SDA_PIN);
        }
        else
        {
            io_clearBitOD(I2C_SDA_PIN);
        }
        data <<= 1;

        delay_ns(I2C_TLOW_NS);
        io_setBitOD(I2C_SCL_PIN);
        delay_ns(I2C_THIGH_NS);

        /* Clock stretching erkennen und warten */
        while( !io_bitIsSet(I2C_SCL_PIN) );

        io_clearBitOD(I2C_SCL_PIN);
    }


    /* Ack pruefen */
    io_setBitOD(I2C_SDA_PIN);
    delay_ns(I2C_TLOW_NS);
    io_setBitOD(I2C_SCL_PIN);
    bool ack = !io_bitIsSet(I2C_SDA_PIN);
    delay_ns(I2C_THIGH_NS);
    io_clearBitOD(I2C_SCL_PIN);
    delay_ns(I2C_TLOW_NS);

    if( ack )
        return i2c_ok;
    else
        return i2c_ackTimeout;
}



/** Start condition, Adresse und R/W-Bit senden
 * @param[in] I2Cx: Dummy
 * @param[in] address  : 7-Bit-Adresse linksbuendig
 * @param[in] direction: I2C_Direction_Tranmitter (Master+Senden) oder
 *            I2C_Direction_Receiver (Master+Empfangen) */
i2c_errorcode_t I2C_bitbang_start(I2C_TypeDef *I2Cx, uint8_t address, uint8_t direction)
{
    io_setBitOD(I2C_SDA_PIN);
    io_setBitOD(I2C_SCL_PIN);
    delay_us(I2C_BUS_FREE_TIME_US);


    /* START condition senden */
    io_clearBitOD(I2C_SDA_PIN);
    delay_us(I2C_HOLD_TIME_START_CONDITION_US);

    /* Streng genommen kein Teil der Start condition mehr */
    io_clearBitOD(I2C_SCL_PIN);
    delay_ns(I2C_TLOW_NS);

    /* Slave-Adresse + R/!W-Bit kombinieren */
    if( direction != I2C_Direction_Transmitter )
    {
        address |= 0x01;
    }

    return I2C_bitbang_write(I2Cx, address);
}



/** Ein Byte vom Slave lesen und wahlweise mit ACK besttigen
 * @param[in] I2Cx: Dummy
 * @param[in] data: Byte
 * @param[in] ack: TRUE|FALSE
 * @return Status */
static i2c_errorcode_t I2C_bitbang_read(I2C_TypeDef *I2Cx, uint8_t *data, bool ack)
{
    (void) I2Cx;
    uint8_t in = 0;


    /* 8 Bit einlesen */
    for( uint_fast8_t i = 0; i<8; i++ )
    {
        io_setBitOD(I2C_SCL_PIN);
        in <<= 1;
        bool bit = io_bitIsSet(I2C_SDA_PIN);
        in |= bit;

        delay_ns(I2C_THIGH_NS);
        io_clearBitOD(I2C_SCL_PIN);
        delay_ns(I2C_TLOW_NS);
    }


    /* Ack senden */
    if( ack )
    {
        io_clearBitOD(I2C_SDA_PIN);
    }
    else
    {
        /* SDA ist ja sowieso high */
    }

    io_setBitOD(I2C_SCL_PIN);
    delay_ns(I2C_THIGH_NS);
    io_clearBitOD(I2C_SCL_PIN);
    delay_ns(I2C_TLOW_NS);

    /* Keine Auswirkungen bei NACK */
    io_setBitOD(I2C_SDA_PIN);

    *data = in;

    return i2c_ok;
}



/** Ein Byte vom Slave lesen und mit ACK besttigen = ein weiteres Byte anfordern
 * @param[in] I2Cx: Dummy
 * @param[in] data: Byte
 * @return Status */
i2c_errorcode_t I2C_bitbang_read_ack(I2C_TypeDef *I2Cx, uint8_t *data)
{
    return I2C_bitbang_read(I2Cx, data, true);
}



/** Ein Byte vom Slave lesen und nicht mit ACK besttigen = kein weiteres Byte anfordern
 * @param[in] I2Cx: Dummy
 * @param[in] data: Byte
 * @return Status */
i2c_errorcode_t I2C_bitbang_read_nack(I2C_TypeDef *I2Cx, uint8_t *data)
{
    return I2C_bitbang_read(I2Cx, data, false);
}



/** Sendet STOP condition und gibt den Bus frei
 * @param[in] I2Cx: Dummy
 * @return Status */
i2c_errorcode_t I2C_bitbang_stop(I2C_TypeDef *I2Cx)
{
    (void) I2Cx;

    io_clearBitOD(I2C_SDA_PIN);
    delay_ns(I2C_TLOW_NS);

    io_setBitOD(I2C_SCL_PIN);
    delay_ns(I2C_THIGH_NS);

    io_setBitOD(I2C_SDA_PIN);
    return i2c_stop;
}

