/** @file i2c_reset.c
 *
 * I2C: Freitakten des I2C-Busses per Bitbangig
 *
 * Fuer die Timeouts wird erwartet, dass der Timeout-Timer (setTimeout und pollTimeout) laeuft.
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 06.04.2014 @author Nicolas */


#include "hw_config.h"
#include "i2c.h"
#include "delay.h"
#include "io.h"
#include "wallclock.h"




/** I2C-Reset nach Philips AN10216 I2C Manual Slide 42 ("freitakten")
 *
 * Ein Hardware-I2C wird deaktiviert, das Freitakten wird per Bitbanging erledigt und der
 * urspruengliche I2C wieder aktiviert.
 *
 * @param[in] I2Cx: dummy
 * @return I2C-Status */
uint_fast8_t I2C_reset_Philips(I2C_TypeDef* I2Cx)
{
	uint8_t status, i;

	/* I2C abtrennen, der Rest wird durch Bitbanging mit 50 kHz erledigt */
	i2c_deinit(I2Cx);


	/* Setup SCL und SDA pins */
	io_setOutputOD(I2C_SDA_PIN);
	io_setOutputOD(I2C_SCL_PIN);


	/* Bus recovery
	   1. Send 9 clock pulses on SCL line */
	io_setBit(I2C_SCL_PIN);
	io_setBit(I2C_SDA_PIN);
	for( i = 0; i<9; i++ )
	{
		delay_us(I2C_BITBANGING_DELAY_US);
		io_clearBit(I2C_SCL_PIN);
		delay_us(I2C_BITBANGING_DELAY_US);
		io_setBit(I2C_SCL_PIN);
	}

	{
		/* 2. Ask the master to keep SDA high until the "Slave Transmitter" releases the SDA line
		      to perform the ACK operation */
		wallclock_t wallclock = wallclock_tic();
		while( !io_bitIsSet(I2C_SDA_PIN) )
		{
			if( wallclock_toc_ms(wallclock) > I2C_RESETTIMEOUT_MS )
			{
				return i2c_error;
			}
		}
	}

	/* 3. Keeping SDA high during the ACK means that the "Master Receiver" does
	      not acknowledge the previous byte receive.

       4. The "Slave Transmitter" then goes to an idle state

	   5. The master then sends a STOP command initializing completely the bus */
	io_clearBitOD(I2C_SCL_PIN);
	delay_us(I2C_BITBANGING_DELAY_US);
	io_clearBitOD(I2C_SDA_PIN);
	delay_us(I2C_BITBANGING_DELAY_US);
	io_setBitOD(I2C_SCL_PIN);
	delay_us(I2C_BITBANGING_DELAY_US);
	io_setBitOD(I2C_SDA_PIN);

    /* Pins wieder abgeben */
	io_setInput(I2C_SDA_PIN);
	io_setInput(I2C_SCL_PIN);


	/* Hardware I2C wieder uebernehmen lassen */
	status = i2c_init(I2Cx);
	return status;
}



/** I2C-Reset nach Analog Devices Application Note AN-686s AN10216
 *
 * Ein Hardware-I2C wird deaktiviert, das Freitakten wird per Bitbanging erledigt und der
 * urspruengliche I2C wieder aktiviert.
 *
 * @param[in] I2Cx: dummy
 * @return I2C-Status */
uint_fast8_t i2c_reset_AD(I2C_TypeDef* I2Cx)
{
	uint8_t status, i;


	/* I2C abtrennen, der Rest wird durch Bitbanging mit 50 kHz erledigt */
	i2c_deinit(I2Cx);


	/* Setup SCL und SDA pins */
	io_setOutputOD(I2C_SDA_PIN);
	io_setOutputOD(I2C_SCL_PIN);


	/* Bus recovery:
	   1. Master tries to assert a Logic 1 on the SDA Line */
	io_setBitOD(I2C_SCL_PIN);
	io_setBitOD(I2C_SDA_PIN);
	i = 0;
	while( i < 9 && !io_bitIsSet(I2C_SDA_PIN) )
	{
		i++;
		delay_us(I2C_BITBANGING_DELAY_US);
		io_clearBitOD(I2C_SCL_PIN);
		delay_us(I2C_BITBANGING_DELAY_US);
		io_setBitOD(I2C_SCL_PIN);
	}

	io_clearBitOD(I2C_SCL_PIN);

	/* 4. Generate STOP Condition */
	io_clearBitOD(I2C_SCL_PIN);
	delay_us(I2C_BITBANGING_DELAY_US/2);
	io_clearBitOD(I2C_SDA_PIN);
	delay_us(I2C_BITBANGING_DELAY_US/2);
	io_setBitOD(I2C_SCL_PIN);
	delay_us(I2C_BITBANGING_DELAY_US/2);
	io_setBitOD(I2C_SDA_PIN);

    /* Pins wieder abgeben */
    io_setInput(I2C_SDA_PIN);
    io_setInput(I2C_SCL_PIN);

	/* Hardware I2C wieder uebernehmen lassen */
	status = i2c_init(I2Cx);
	return status;
}



/** Beide I2C-Kanaele freitakten */
void i2c_reset_all(void)
{
	i2c_reset(I2C1);
	i2c_reset(I2C2);
}
