/** @file i2c_simple.c
 *
 * Einfache Variante von I2C-Routinen
 *
 * Diese koennen auf allen Plattformen angewendet werden, wo START/READ_ACK/READ_NACK etc. getrennt
 * hintereinander verarbeitet werden koennen, z.B. AVR, STM32F4xx und Bitbanging. Auf dem
 * STM32F10x sind andere Implementierungen besser.
 *
 * Fuer die Timeouts wird erwartet, dass der Timeout-Timer (setTimeout und
 * pollTimeout) laeuft (sonst wird z.B. der Scan steckenbleiben)
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 22.12.13 @author Nicolas */


#include "i2c_simple.h"
#include "hw_config.h"
#include "i2c.h"

#if I2C_SIMPLE || defined( I2C_BITBANG ) || I2C_MOCK || DOXYGEN

    #define checkforerror(s,l) if(s) {s |= (l<<4); goto error;}



    /** Ganzen Puffer ueber I2C lesen
     *
     * @param[in] I2Cx: I2C1|I2C2
     * @param[in] slaveaddr: I2C-Adresse (linksbuendig)
     * @param[out] *pBuffer : Puffer passender Groesse
     *            wenn pBuffer == NULL werden nBuffer Bytes vom Bus gelesen und verworfen.
     * @param[in] nBuffer  : Puffergroesse
     *            Wenn nBuffer == NULL wird nur Stopbedingung gesendet
     *
     * @return Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
     *            und STOP-Bedingung erzeugt werden */
    i2c_errorcode_t i2c_receiveBuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
            uint8_t* pBuffer, uint8_t nBuffer)
    {
        uint8_t status, i, byte;

        if (nBuffer==0)
        {
            I2C_stop(I2Cx);
            return 0;
        }

        status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Receiver);
        checkforerror(status,5);

        for(i=0;i<nBuffer-1;i++)
        {
            status = I2C_read_ack(I2Cx,&byte);
            checkforerror(status,6);
            if(pBuffer)
            {
                pBuffer[i] = byte;
            }
        }

        status = I2C_read_nack(I2Cx,&byte);
        checkforerror(status,7);
        if(pBuffer)
        {
            pBuffer[nBuffer-1] = byte;
        }

        error:
        I2C_stop(I2Cx);
        return status;
    }



    /** Ganzen Puffer ueber I2C senden
     *
     * @param[in] I2Cx: I2C1|I2C2
     * @param[in] slaveaddr: I2C-Adresse (linksbuendig)
     * @param[in] pBuffer: Puffer passender Groesse
     * @param[in] nBuffer : Puffergroesse
     * @param[in] stopbit: stop|noStop
     *            Stop-Bedingung wird gesendet (normales Senden an Slave)
     *            Stop-Bedingung wird nicht gesendet (Antwort von Slave vorbereiten)
     *
     * @return Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt und STOP-Bedingung
     *         erzeugt werden  */
    i2c_errorcode_t i2c_sendBuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
            uint8_t* pBuffer, uint8_t nBuffer, enum i2c_stopbit_e stopbit )
    {
        ufast8_t status, i;

        status = I2C_waitbusy(I2Cx);
        checkforerror(status,1);

        status = I2C_start(I2Cx, slaveaddr, I2C_Direction_Transmitter);
        checkforerror(status, 2);

        for(i=0; i<nBuffer; i++)
        {
            status = I2C_write(I2Cx, pBuffer[i]);
            checkforerror(status,4);
        }

        switch (stopbit)
        {
            case stop:
                I2C_stop(I2Cx);
                break;
            case noStop:
                ;
        }

        error:
        return status;
    }



    /** I2C-Adresse kurz anpingen um zu sehen, ob Geraet vorhanden
     *
     * @param[in] I2Cx: I2C1|I2C2
     * @param[in] slaveaddr: I2C-Adresse (linksbuendig)
     * @return i2c_ok          : Device vorhanden und hat mit ACK bestaetigt
     *         i2c_busyTimeout : Bus nicht frei
     *         i2c_otherTimeout: Unerwarteter Fehler
     *         i2c_ackTimeout  : NACK, Device nicht vorhanden */
    i2c_errorcode_t i2c_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr)
    {
        uint_fast8_t status;

        /* i2c_ok oder i2c_busyTimeout */
        status = I2C_waitbusy(I2Cx);
        if(status)
        {
            status |= (2<<4);
            goto error;
        }

        /* Startbedingung
           i2c_ok           : ACK empfangen, Teilnehmer vorhanden
           i2c_ackTimeout   : kein ACK empfangen, keine Antwort vom Teilnehmer
           i2c_otherTimeout : Start-Bedingung konnte nicht gesendet werden. Senderfehler. */
        status = I2C_start(I2Cx,slaveaddr,I2C_Direction_Transmitter);
        I2C_stop(I2Cx);

        error:
        return status;
    }



#else
    typedef int this_is_no_empty_translation_unit;

#endif
