/** @file bitmath.h
 *
 * Einfache Funktionen fuer Bitmasken & Co.
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * 02.05.2015 Nicolas */

#ifndef BITMATH_H
    #define BITMATH_H

    #include "myassert.h"
    #include <stdint.h>
    #include <stddef.h>
    #include <stdbool.h>

    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif




	/** Byte strecken, indem jedes Bit verdoppelt wird(8 -> 16 Bit)
	 *
	 * z.B.: 01001100 -> 0011000011110000
	 * Die Funktion sollte mit konstantem Parameter aufgerufen werden, damit sie
	 * statisch wegoptimiert werden kann.
	 *
	 * @param[in] byte: Bitmaske
	 * @return gestreckte Bitmaske */
	static_inline uint16_t stretch2bitmask(const uint8_t byte)
	{
	    uint16_t out = 0;
	    out |= byte & 0x01U ?  0x0003U : 0;
	    out |= byte & 0x02U ?  0x000CU : 0;
	    out |= byte & 0x04U ?  0x0030U : 0;
	    out |= byte & 0x08U ?  0x00C0U : 0;
	    out |= byte & 0x10U ?  0x0300U : 0;
	    out |= byte & 0x20U ?  0x0C00U : 0;
	    out |= byte & 0x40U ?  0x3000U : 0;
	    out |= byte & 0x80U ?  0xC000U : 0;
	    return out;
	}


	/** Byte strecken, indem jedes Bit verdoppelt wird (16 -> 32 Bit)
	 *
	 * z.B.: 01001100 01001100 -> 0011000011110000 0011000011110000
	 * Die Funktion sollte mit konstantem Parameter aufgerufen werden, damit sie
	 * statisch wegoptimiert werden kann.
	 *
	 * @param[in] byte: Bitmaske
	 * @return gestreckte Bitmaske */
	static_inline uint32_t stretch2bitmask16(const uint16_t word)
	{
	    uint32_t out = 0;
	    out |= word & 0x0001U ?  0x00000003U : 0;
	    out |= word & 0x0002U ?  0x0000000CU : 0;
	    out |= word & 0x0004U ?  0x00000030U : 0;
	    out |= word & 0x0008U ?  0x000000C0U : 0;
	    out |= word & 0x0010U ?  0x00000300U : 0;
	    out |= word & 0x0020U ?  0x00000C00U : 0;
	    out |= word & 0x0040U ?  0x00003000U : 0;
	    out |= word & 0x0080U ?  0x0000C000U : 0;
	    out |= word & 0x0100U ?  0x00030000U : 0;
	    out |= word & 0x0200U ?  0x000C0000U : 0;
	    out |= word & 0x0400U ?  0x00300000U : 0;
	    out |= word & 0x0800U ?  0x00C00000U : 0;
	    out |= word & 0x1000U ?  0x03000000U : 0;
	    out |= word & 0x2000U ?  0x0C000000U : 0;
	    out |= word & 0x4000U ?  0x30000000U : 0;
	    out |= word & 0x8000U ?  0xC0000000U : 0;
	    return out;
	}



	/** Byte strecken, indem jedes Bit vervierfacht wird
	 *
	 * z.B.: 01001100 -> 0000111100000000111111110000000000000000
	 * Die Funktion sollte mit konstantem Parameter aufgerufen werden, damit sie
	 * statisch wegoptimiert werden kann.
	 *
	 * @param[in] byte: Bitmaske
	 * @return gestreckte Bitmaske */
	static_inline uint32_t stretch4bitmask(const uint8_t byte)
	{
	    uint32_t out = 0;
	    out |= byte & 0x01U ?  0x0000000FU : 0;
	    out |= byte & 0x02U ?  0x000000F0U : 0;
	    out |= byte & 0x04U ?  0x00000F00U : 0;
	    out |= byte & 0x08U ?  0x0000F000U : 0;
	    out |= byte & 0x10U ?  0x000F0000U : 0;
	    out |= byte & 0x20U ?  0x00F00000U : 0;
	    out |= byte & 0x40U ?  0x0F000000U : 0;
	    out |= byte & 0x80U ?  0xF0000000U : 0;
	    return out;
	}



    /** Bit-Array/Bitmap: Bit an einer einer bestimmten Position in einem Array setzen oder loeschen
     * (einfach indiziert)
     *
     *  @param[in,out] array Byte-Array, interpretiert als Bit-Array
     *  @param[in] bit: Bitposition (einfach indiziert)
     *  @param[in] value: Bitwert */
    static_inline void arrayWriteBit(uint8_t *array, size_t bit, bool value)
    {
        size_t      iByte = bit/8;
        uint_fast8_t iBit = bit%8;
        if( value )
            *(array+iByte) |=  (0x80>>iBit);
        else
            *(array+iByte) &= ~(0x80>>iBit);
    }



    /** Bit-Array/Bitmap: Bit an einer einer bestimmten Position in einem Array lesen
     *  (einfach indiziert)
     *
     *  @param[in] array: Byte-Array, interpretiert als Bit-Array
     *  @param[in] bit: Bitposition (einfach indiziert)
     *  @return Bitwert */
    static_inline bool arrayReadBit(uint8_t *array, size_t bit)
    {
        size_t      iByte = bit/8;
        uint_fast8_t iBit = bit%8;
        return (*(array+iByte) & (0x80>>iBit)) != 0;
    }



    /** Bit-Array/Bitmap: Bit an einer einer bestimmten Position in einem Array setzen oder loeschen.
     *
     *  @param[in,out] array: Byte-Array, interpretiert als Bit-Array
     *  @param[in] i, j: Bitposition (doppelt indiziert)
     *  @param[in] value: [0|1]
     *  @param[in] dimI, dimJ: Dimensionen des zweidimensionalen Arrays */
    static_inline
        void arrayWriteBitIJ(uint8_t *array, size_t i, size_t j, bool value, size_t dimI, size_t dimJ)
    {
        assert( i < dimI ); // 0...dimI
        assert( j < dimJ ); // 0...dimJ
        arrayWriteBit(array, i+dimI*j, value);
    }



    /** Pruefen, ob alle in b gesetzten Bits auch in a gesetzt sind */
    static_inline bool areAllBitsSet(uint32_t a, uint32_t b)
    {
        assert( b != 0 );
        return (a & b) == b;
    }



    /** Pruefen, ob in b gesetzte Bits auch in a gesetzt sind
     *
     * Diese primitive Funktion existiert nur aus Symmetriegruenden zu @see areAllBisSet() */
    static_inline bool areAnyBitsSet(uint32_t a, uint32_t b)
    {
        assert( b!= 0 );
        return (a & b);
    }



    /** Vergleich mit Maske
     *
     * Pruefen, ob fuer alle in der Maske gesetzten Bits die Bits in den Variablen a und b
     * identisch sind
     *
     * @param[in] a: Erste Variable
     * @param[in] b: Zweite Variable
     * @param[in] mask: Bitmaske */
    static_inline bool areBitsEqual(uint32_t a, uint32_t b, uint32_t mask)
    {
        return ((a&mask)^(b&mask)) == 0;
    }



    /** Wieviele Bytes werden benoetigt, um N bits zu speichern ?
     *
     * @param[in] nBit: Anzahl der Bits
     * @return Anzahl der benoetigten Bytes */
    static_inline size_t bytes4bits(unsigned long int nBit)
    {
        size_t nByte = nBit/8;
        if( nBit% 8 )
        {
            nByte += 1;
        }

        return nByte;
    }



    /** Bits in einem Byte umdrehen
     *
     *  - aus den Weiten des Internets -
     *
     * @param[in] b
     * @return b mit umgedrehter Bit-Reihenfolge */
    static_inline uint8_t byteReverseBitOrder(uint8_t b)
    {
       b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
       b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
       b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
       return b;
    }


    /** Nachfolgende Nullen zaehlen (count trailing zeros)
     *
     * @param[in] x: Wert, wird bitweise interpretiert
     * @return Anzahl der End-Nullen */
    static_inline uint8_t ctz(unsigned int x)
    {
        #if 1
            return __builtin_ctz(x);
        #else
            /* Bitwise binary search */
            /* Quelle unbekannt. Entweder Hacker's Delight oder Stackoverflow oder so */
            unsigned int v;      /* 32-bit word input to count zero bits on right  */
            unsigned int c = 32; /* c will be the number of zero bits on the right */
            v &= -signed(v);
            if (v) c--;
            if (v & 0x0000FFFF) c -= 16;
            if (v & 0x00FF00FF) c -= 8;
            if (v & 0x0F0F0F0F) c -= 4;
            if (v & 0x33333333) c -= 2;
            if (v & 0x55555555) c -= 1;
        #endif
    }


#endif

