/** @file intmath.h
 *
 * Einfache Funktionen fuer die Integer-Mathematik
 *
 * z.B.:
 *   - Saettigende Addition, Subtraktion und Multiplikation
 *   - Kleinstes gemeinsames Vielfaches
 *   - Groesster gemeinsamer Teiler
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 08.03.2014 @author Nicolas
 */

#ifndef INTMATH_H_INCLUDED
    #define INTMATH_H_INCLUDED

    #include "nstdint.h"
    #include <limits.h>
    #include "nstdlib.h"
    #include <stdbool.h>
    #include "myassert.h"

    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif

    #if defined(ARM_MATH_CM7) || defined (ARM_MATH_CM4) || defined (ARM_MATH_CM3) || defined (ARM_MATH_CM0)
        #include "arm_math.h"
    #endif



    /* Testen, ob Produkt zweier Variablen in 32/64 Bit passt */
    bool productFits_s32(int32_t a, int32_t b);
    bool productFits_s64_removeme(int64_t a, int64_t b);
    bool productFits_u64(uint64_t a, uint64_t b);

    /* Groesster gemeinsamer Teiler */
    int32_t gcd_s32(int32_t a, int32_t b);
    int64_t gcd_s64(int64_t a, int64_t b);

    /* Kleinstes gemeinsames Vielfaches */
    int32_t lcm_s16(int16_t a, int16_t b );
    int32_t lcm_s32(int32_t a, int32_t b );



    /** Absolutwert
     * undefined behaviour bei INT_MIN */
    #define ABS(a)	   (((a) < 0) ? -(a) : (a))



    /** Vorzeichen */
	#define SIGN(a)  ( (a>0) - (a<0) )



    /** Maximalwert zweier Integer */
	#define MAX(a, b)  ( (a > b) ? (a) : (b) )



    /** Minimalwert zweier Integer */
	#define MIN(a, b)  ( (a < b) ? (a) : (b) )




    /** Pruefen, ob x im Intervall [a...b] liegt. Die Invervallgrenzen sind unsortiert.
     *
     * @param[in] x Vergleichswert
     * @param[in] a obere oder untere Intervallgrenze
     * @param[in] b die andere Intervallgrenze
     * @return a <= x <= b || a <= x <= b */
    static_inline bool inIntervall_s64(int64_t x, int64_t a, int64_t b)
    {
        return ( (a <= x) && (x <= b) ) || ( (a >= x) && (x >= b) );
    }



    /** Compare Threshold: Erweiterte Signum-Funktion (32 Bit)
     *
     * Erweiterte Signum-Funktion mit Toleranzbreite um Nullpunkt
     * @param[in] value:     Eingangswert
     * @param[in] threshold: Vergleichswert
     * @return -1, wenn value < -threshold, 1 wenn value > threshold, sonst 0 */
    static_inline int_fast8_t compareThreshold_s32(int32_t value, int32_t threshold)
    {
        if( value > threshold )
            return 1;
        else if( value < -threshold )
            return -1;
        else
            return 0;
    }



    /** Compare Threshold: Erweiterte Signum-Funktion (64 Bit)
     *
     * Erweiterte Signum-Funktion mit Toleranzbreite um Nullpunkt
     * @param[in] value:     Eingangswert
     * @param[in] threshold: Vergleichswert
     * @return -1, wenn value < -threshold, 1 wenn value > threshold, sonst 0
     *  */
    static_inline int_fast8_t compareThreshold_s64(int64_t value, int64_t threshold)
    {
        if( value > threshold )
            return 1;
        else if( value < -threshold )
            return -1;
        else
            return 0;
    }



    /** Zwei Zahlen nach Groesse sortieren (16 Bit)
     *
     * @param[in] *a, *b: Zwei Zahlen
     * @return EXIT_SUCCESS */
    static_inline exit_t sort_s16(int16_t *a, int16_t *b)
    {
        int16_t tmp = *a;
        if( *b < *a )
        {
            *a = *b;
            *b = tmp;
        }
        return EXIT_SUCCESS;
    }



    /** Saettigung in den Bereich -2^sat1 <= x <= 2^sat1 1.
     * @param[in] x Eingangswert
     * @param[in] sat: Saettigungsbereich (Zeierpotenz) 2^sat1 <= x <= 2^sat1 1, d.h. ssat(x, 9)
     *            laesst einen 8-Bit-Wert zurueck
     * @return gesaettigter Ausgabewert */
    static_inline int32_t sat_s32(int32_t x, int32_t sat)
    {
        #ifdef __SSAT
            // Funktion funktioniert nur als Inline-Funktion, da "sat" als Literal vorliegen
            // muss
            // #include "core_cmInstr.h" muss eingebunden sein
            return  __SSAT(x, sat);
        #else
            const int32_t lim = (1L<<(sat-1));
            return (x < -lim-1 ? -lim-1 : (x > lim ? lim : x  )   );
        #endif
    }




    /** Testen, ob Zahl Zweierpotenz (also genau ein Bit gesetzt) ist */
    #define ISPOWEROFTWO(x) ( (x) && (( ( (x) & ((x) - 1)) ) == 0) )



    /** Testen, ob Zahl Zweierpotenz (also genau ein Bit gesetzt) ist
     *
     * Aufgrund der Darstellung im Zweierkomplement trift dies nur bei positiven Zahlen zu.
     * @param[in] x: Zahl
     * @return x == [1, 2, 4, 8, 16 ... ] */
    static_inline bool ispoweroftwo(int32_t x)
    {
        return x && ((x & (x - 1)) == 0);
    }



    /**Zweier-Logarithmus (32 Bit) */
    static_inline uint8_t log2_u32(uint32_t a)
    {
        if( a == 0 )
            return 0;
        else
            return 31-__builtin_clzl(a);
    }



    /** Integer-Division mit korrektem Runden (64 Bit)
     *
     * @param[in] z: Zaehler
     * @param[in] n: Nenner (Darf nicht Null sein )
     * @return Quotient */
    static_inline int64_t div_s64(int64_t z, int64_t n)
    {
        assert( n != 0 );

        if( (z < 0) ^ (n < 0) )
        {
            /* (z - n/2)/n; */
            int64_t temp;
            assert ( !__builtin_ssubll_overflow(z, n/2, &temp) );
            return temp/n;
        }
        else
        {
            /* (z + n/2)/n; */
            int64_t temp;
            assert ( !__builtin_saddll_overflow(z, n/2, &temp) );
            return temp/n;
        }
    }



    /* ******************************************************************************** *
     *    Clamping                                                                      *
     * ******************************************************************************** */

    /** Generische Makro-Funktion fuer Clamp. Nicht fuer den direkten Aufruf gedacht
     *
     * Ist lo > hi, ist das Ergebnis unerwartet. */
    #define CLAMP(x, lo, hi) \
        (x) > (hi) ? (x)=(hi), 1 : \
        (x) < (lo) ? (x)=(lo), -1 : \
        0\



    /** Clamping (Signed Integer)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int sclamp(int *x, const int lo, const int hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (Long Integer)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int lclamp(long int *x, long int lo, long int hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (Long Long Integer)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int llclamp(long long int *x, long long int lo, long long int hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (Signed Integer)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_s8(int8_t *x, int8_t lo, int8_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (16 Bit)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_s16(int16_t *x, int16_t lo, int16_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (32 Bit)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_s32(int32_t *x, int32_t lo, int32_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (64 Bit)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_s64(int64_t *x, int64_t lo, int64_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (8 Bit unsigned)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_u8(uint8_t *x, uint8_t lo, uint8_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (16 Bit unsigned)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_u16(uint16_t *x, uint16_t lo, uint16_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Clamping (32 Bit unsigned)
     * @param[in,out] *x : Wert
     * @param[in] lo, hi: Untere und obere Grenze, lo < hi
     * @return: -1, 1, 0 (negativ|positiv|nicht begrenzt) */
    static_inline int clamp_u32(uint32_t *x, uint32_t lo, uint32_t hi)
    {
        return CLAMP(*x, lo, hi);
    }



    /** Saettingende Addition (32 Bit)
     *
     * Auf Bitbreite saettigende Addition
     * @param[in] a, b: Summanden
     * @return: Summe/INT32_MAX/INT32_MIN  */
    static_inline int32_t satadd_s32(int32_t a, int32_t b)
    {
        #if defined( __ARM_FEATURE_QBIT )
            return __QADD(a, b);
        #else
            int64_t r = (int64_t) a + b;
            clamp_s64(&r, INT32_MIN, INT32_MAX);
            return r;
        #endif
    }



    /** Saettingende Addition (16 Bit)
     *
     * Auf Bitbreite saettigende Addition
     * @param[in] a, b: Summanden
     * @return: Summe/INT16_MAX/INT16_MIN  */
    static_inline int16_t satadd_s16(int16_t a, int16_t b)
    {
        int_fast24_t r = (int_fast24_t) a + b;
        CLAMP(r, INT16_MIN, INT16_MAX);
        return r;
    }



    /** Saettingende Addition (8 Bit)
     *
     * Auf Bitbreite saettigende Addition
     * @param[in] a, b: Summanden
     * @return: Summe/INT8_MAX/INT8_MIN  */
    static_inline int8_t satadd_s8(int8_t a, int8_t b)
    {
        int_fast16_t r = (int_fast16_t) a + b;
        CLAMP(r, INT8_MIN, INT8_MAX);
        return r;
    }



    /** Saettingende Addition (8 Bit unsigned)
     *
     * Auf Bitbreite saettigende Addition
     * Das Ergebnis ist immer groesser oder gleich dem groesseren der beiden Summanden. Die
     * implizite Subtraktion (durch Cast einer negativen Zahl) funktioniert damit nicht.
     * @param[in] a, b: Summanden
     * @return: Summe|0|UINT8_MAX */
    static_inline uint8_t satadd_u8(uint8_t a, uint8_t b)
    {
        uint8_t r = a + b;
        if( r < a ) /* Nuer bei Ueberlauf moeglich */
        {
            r = UINT8_MAX;
        }
        return r;
    }



    /** Saettingende Addition (16 Bit unsigned)
     *
     * Auf Bitbreite saettigende Addition
     * Das Ergebnis ist immer groesser oder gleich dem groesseren der beiden Summanden. Die
     * implizite Subtraktion (durch Cast einer negativen Zahl) funktioniert damit nicht.
     * @param[in] a, b: Summanden
     * @return: Summe|0|UINT16_MAX */
    static_inline uint16_t satadd_u16(uint16_t a, uint16_t b)
    {
        uint16_t r = a + b;
        if( r < a ) /* Nuer bei Ueberlauf moeglich */
        {
            r = UINT16_MAX;
        }
        return r;
    }



    /** Saettingende Addition (32 Bit unsigned)
     *
     * Auf Bitbreite saettigende Addition
     * Das Ergebnis ist immer groesser oder gleich dem groesseren der beiden Summanden. Die
     * implizite Subtraktion (durch Cast einer negativen Zahl) funktioniert damit nicht.
     * @param[in] a, b: Summanden
     * @return: Summe|0|UINT32_MAX */
    static_inline uint32_t satadd_u32(uint32_t a, uint32_t b)
    {
        uint32_t r = a + b;
        if( r < a ) /* Nuer bei Ueberlauf moeglich */
        {
            r = UINT32_MAX;
        }
        return r;
    }



    /** Saettingende Akkumulation (8 Bit)
     *
     * Auf Bitbreite saettigende Akkumlation einer vorzeichenlosen Zahl mit einem
     * vorzeichenbehafteten Inkrement.
     * @param[in] akku: Summand
     * @param[in] incr: Inkrement
     * @return: Summe|0|UINT8_MAX */
    static_inline uint8_t satakku_u8(uint8_t akku, int8_t incr)
    {
        int_fast16_t r = (int_fast16_t) akku + incr;
        CLAMP(r, 0, UINT8_MAX);
        return r;
    }



    /** Saettingende Akkumulation (16 Bit)
     *
     * Auf Bitbreite saettigende Akkumlation einer vorzeichenlosen Zahl mit einem
     * vorzeichenbehafteten Inkrement.
     * @param[in] akku: Summand
     * @param[in] incr: Inkrement
     * @return: Summe|0|UINT8_MAX */
    static_inline uint16_t satakku_u16(uint16_t akku, int16_t incr)
    {
        int_fast24_t r = (int_fast24_t) akku + incr;
        CLAMP(r, 0, UINT16_MAX);
        return r;
    }



    /** Saettingende Akkumulation (32 Bit)
     *
     * Auf Bitbreite saettigende Akkumlation einer vorzeichenlosen Zahl mit einem
     * vorzeichenbehafteten Inkrement.
     * @param[in] akku: Summand
     * @param[in] incr: Inkrement
     * @return: Summe|0|UINT62_MAX */
    static_inline uint32_t satakku_u32(uint32_t akku, int32_t incr)
    {
        int64_t r = (int64_t) akku + incr;
        CLAMP(r, 0, UINT32_MAX);
        return r;
    }



    /** Saettingende Multiplikation (8 Bit)
     *
     * @return: Produkt/INT8_MAX/INT8_MIN  */
    static_inline int8_t satmul_s8(int8_t fac1, int8_t fac2)
    {
        int_fast16_t p = (int_fast16_t) fac1 * fac2;
        CLAMP(p, INT8_MIN, INT8_MAX);
        return p;
    }



    /** Saettingende Multiplikation (32 Bi)
     *
     * @return: Produkt/INT32_MAX/INT32_MIN  */
    static_inline int32_t satmul_s32(int32_t fac1, int32_t fac2)
    {
        int64_t p = (int64_t) fac1 * fac2;
        CLAMP(p, INT32_MIN, INT32_MAX);
        return p;
    }



    /** Vorzeichenbehafteten Integer beliebiger (kleiner) Breite erweitern
     * (z.B. von 24 Bit auf 32 Bit)
     *
     * @param[in] i: Eingangswert
     * @param[in] bitwidth: Bitbreite des Eingangswerts
     * @return Eingangswert als int32 */
    static_inline int32_t expandInt_s32(int32_t i, uint_fast8_t bitwidth )
    {
        assert( bitwidth <= 32 );

        /* Nach Hacker's Delight S 19f., evtl. ist eine Loesung ueber shiften schneller */
        uint32_t temp = (uint32_t) i;
        const uint32_t sign = 1U<<(bitwidth-1);
        temp = (temp & (sign-1)) - (temp & sign);

        return (int32_t) temp;
    }


#endif

