/** @file menu_color.h
 *
 * Menusystem (Darstellung eines Menues oder Popup-Fensters)
 *
 * Das Menusystem benutzt malloc().
 *
 * $Date: 2020-08-13 11:28:54 +0200 (Do, 13 Aug 2020) $
 * $Revision: 68 $
 *
 * @date 10.07.2013 @author Nicolas  */


#ifndef MENU_COLOR_H
    #define MENU_COLOR_H

    #include "sw_config.h"

    #include "pgmspace.h"
    #include "rect.h"
    #include "color.h"

    #if defined( MENU_SUPPORT_FLOAT )
        #include "floatmath.h"
    #endif



    /** Enum: Menue-Eintrags-Datentyp
     * Dieser Datentyp bestimmt, welche Aktion mit dem ausgewaehlten Menue-Element ausgefuehrt wird
     */
	typedef enum menutype_s
    {
        nothing,    /**< Kein Datentype - keine Aktion */
		voidFcn,    /**< void-Function - wird ausgefuehrt, ohne dass das Menu verlassen wird */
		uint8,      /**< Integer-Datentyp. Laesst sich in gewaehltem Intervall verstellen */
		uint16,     /**< Integer-Datentyp. Laesst sich in gewaehltem Intervall verstellen */
		uint32,     /**< Integer-Datentyp. Laesst sich in gewaehltem Intervall verstellen */
		int8,       /**< Integer-Datentyp. Laesst sich in gewaehltem Intervall verstellen */
		int16,      /**< Integer-Datentyp. Laesst sich in gewaehltem Intervall verstellen */
		int32,      /**< Integer-Datentyp. Laesst sich in gewaehltem Intervall verstellen */
    #if MENU_SUPPORT_FLOAT
        flin32,     /**< Float-Datentyp. Laesst sich in gewaehltem Intervall (linear) verstellen */
        flog32,     /**< Float-Datentyp. Laesst sich in gewaehltem Intervall (logarithmisch) verstellen */
    #endif
        strlist,    /**< Enum-artiger-Datentyp/Auswahlliste */
		exitMenu,   /**< Aktion: Menue beenden. Rueckgabewert ist Index des Menueeintrags */
		endmarker,  /**< Menu-Endmarker (Array-Ende) */
	}
	menutype_t;



    /** @struct Menuitem_s
     * @brief Menue-Eintraege: Zu startende Funktionen, zu aendernde Variablen
     *
     * Die Reihenfolge macht es uebersichtlicher, Menutext und Variableninhalt separat
     * zu speichern.
     * Ein Eintrag sieht dann z.B. wie folgt aus:
     * @example Menuitem1: { FLSTEXT1, .Uint8={ .type=uint8, .ptr=&var, .min=1, .max=255, .inc=5} }
     * @example Menuitem2: { FLSTEXT2, .VoidFcn={ voidFcn, &foo }       }
     *
     * Struct wird terminiert mit einer Zeile, wo in .menutext ein Nulltext ({'\0'}) steht. */
    __extension__ typedef struct Menuitem_s
    {
	    constflash char *menutext;          /**< Menueeintrag/Funktionsname
                                             *
                                             * Der Text wird mit *printf() interpretiert. Die
                                             * entsprechenden Formatierungszeichen muessen zum
                                             * Datentyp des Menue-Eintrags passen.
                                             *
                                             * Ein Nullstring terminiert das struct */
        union
	    {
		    /** Generisch */
		    struct
		    {
			    menutype_t type;                /**< Datentyp:

									            Bei function pointer wird die Funktion
												gestartet, bei Zeigern auf normale Daten
									            der Datentyp im Menue bearbeitet. */
		    }
            All;

		    /** Container void-void-Funktionen */
		    struct
		    {
			    menutype_t type;                ///< Datentyp
			    void (*ptr) (void);         ///< Funktionszeiger
		    }
            VoidFcn;

		    /** Container uint8_t-Variable */
		    struct
		    {
			    menutype_t type;               ///< Datentyp
			    uint8_t *ptr;              ///< Zeiger auf Variablentypen
			    uint8_t min;               ///< Untere Grenze
			    uint8_t max;               ///< Obere Grenze
			    uint8_t inc;               ///< Inkrement
		    }
            Uint8;

		    /** Container int8_t-Variable */
		    struct
		    {
			    menutype_t type;               ///< Datentyp
			    int8_t *ptr;               ///< Zeiger auf Variablentypen
			    int8_t min;                ///< Untere Grenze
			    int8_t max;                ///< Obere Grenze
			    int8_t inc;                ///< Inkrement
		    }
            Int8;

		    /** Container uint16_t-Variable */
		    struct
		    {
			    menutype_t type;               ///< Datentyp
			    uint16_t *ptr;             ///< Zeiger auf Variablentypen
			    uint16_t min;              ///< Untere Grenze
			    uint16_t max;              ///< Obere Grenze
			    uint16_t inc;              ///< Inkrement
		    }
            Uint16;

		    /** Container int16_t-Variable */
		    struct
		    {
			    menutype_t type;               ///< Datentyp
			    int16_t *ptr;              ///< Zeiger auf Variablentypen
			    int16_t min;               ///< Untere Grenze
			    int16_t max;               ///< Obere Grenze
			    int16_t inc;               ///< Inkrement
		    }
            Int16;

		    /** Container uint32_t-Variable */
		    struct
		    {
			    menutype_t type;               ///< Datentyp
			    uint32_t *ptr;             ///< Zeiger auf Variablentypen
			    uint32_t min;              ///< Untere Grenze
			    uint32_t max;              ///< Obere Grenze
			    uint32_t inc;              ///< Inkrement
		    }
            Uint32;

		    /** Container int32_t-Variable */
		    struct
		    {
			    menutype_t type;                ///< Datentyp
			    int32_t *ptr;               ///< Zeiger auf Variablentypen
			    int32_t min;                ///< Untere Grenze
			    int32_t max;                ///< Obere Grenze
			    int32_t inc;                ///< Inkrement
		    }
            Int32;


            #if defined( MENU_SUPPORT_FLOAT )
		    /** Container 32-Bit-float-Variable */
		    struct
		    {
			    menutype_t type;                ///< Datentyp
			    float *ptr;                 ///< Zeiger auf Variablentypen
			    float min;                  ///< Untere Grenze
			    float max;                  ///< Obere Grenze
			    float inc;                  ///< Inkrement
		    }
            Float32;
            #endif

            /** String-Array, das enum-aehnlich genutzt wird, d.h. eine Auswahlliste bildet */
            struct
		    {
			    menutype_t type;                ///< Datentyp
			    uint8_t *idx;               ///< Zeiger auf String-Index (Verstellvariable)
			    const char **str;           ///< Zeiger auf String-Array
			    uint8_t len;                ///< Obere Grenze
			    uint8_t inc;                ///< Inkrement
		    }
            Strlist;
        };
    } Menuitem_t;



    /** Menueinstellung: Farbstruktur */
    typedef struct Menucolor_s
    {
        color_t mainBackground;
        color_t mainText;
        color_t cursorText;
        color_t cursorBackground;
        color_t cursorFrame;
        color_t scrollbarActive;
        color_t scrollbarInactive;
        color_t scrollbarFrame;
        color_t headlineBackground;
        color_t headlineText;
        color_t headlineIcon;
    }
    Menucolor_t;


    typedef struct Menudimension_s
    {
        Rect_t   Area;                 // Position, Groesse (inkl. Kopfzeile und Rand)
        uint8_t  headheight;           // Hoehe der Kopfzeile
        int8_t   headsep;              // Abstand Kopfzeile vom Menuekoerper
        uint16_t scrollbarWidth;       // Breite Scrollbar
        int8_t   margin;               // Breite Rand (links, rechts, unten)
        int8_t   textIndent;           // Texteinzug
        uint8_t  cornerRadius;         // Eckenradius
        uint8_t  textcellwidth[4];     // Breite der einzelnen Textelemente (relativ)
        uint8_t  textalign[4];         // Text-Ausrichtung
    }
    Menudimension_t;


    typedef struct Menucontrol_s
    {
        void (*pInitFcn)(void);        // Zeiger auf Init-Funktion
        void (*pRptFcn)(void);         // Funktionszeiger auf Funktion, die bei
                                       // jedem Schleifendurchlauf aufgerufen
                                       // wird, z.B. zum speziellen Darstellen
                                       // der Kopfzeile.
        int (*pIncrFcn)(void);         // Funktion zum Inkrementieren (z.B. Cursor bewegen)
        bool (*pActnFcn)(void);        // Funktion fuer Aktionen (z.B. Tastendruck)
    }
    Menucontrol_t;

    // Menue-Konfiguration
    typedef struct Menuconfig_s
    {
        Menudimension_t const* Dimension; // Abmessungen
        Menucolor_t const* Color;         // Farbschema
        Menucontrol_t const* Control;     // Steuerfunktionen
        constflash char *headtext;        // Kopfzeile (Text)
        const uint8_t *iconFont;          /**< Schriftart Piktogramme Kopfzeile */
        uint16_t iconLeft;                // Kopfzeile (Piktogramm) als Unicode)
        uint16_t iconRight;
    }
    Menuconfig_t;





    uint_fast8_t menu_standard(constflash Menuitem_t Menudata[],
                        constflash Menuconfig_t *Menuconfig,
                        uint8_t selectedItem);



#endif

