/** @file menu_color_private_draw.c
 *
 * Menusystem (Hilfsfunktionen): Darstellung
 *
 * * Darstellung eines Menues
 *
 * * Darstellung von "Popup-Fenstern"
 *
 * $Date: 2020-08-02 19:08:25 +0200 (So, 02 Aug 2020) $
 * $Revision: 32 $
 *
 * @date 19.03.2018 @author Nicolas  */


#include "menu_color_private_draw.h"
#include "menu_color_private_color.h"
#include "menu_color_private_geometry.h"
#include "menu_color_private_item.h"
#include "text_cursor.h"
#include "text_field.h"
#include "gfx_draw_rectangles.h"
#include "gfx_draw.h"
#include "main.h"


// Aktueller Zustand Menuitem
typedef struct MenuItemState_s
{
    uint8_t selectedItem; // Aktuelle ausgewaehltes Element
    uint8_t scrollPos;    // Aktuelle Scroll-Position
    uint8_t nScrollPos;   // Anzahl der moeglichen Scroll-Positionen
//    uint32_t itemVal;     // Wert des Items
}
MenuItemState_t;


struct Menudrawdata_s
{
    Rect_t Area;                   // Platz fuer Tabelle
    uint16_t tablecellwidth[4];    // Breite der Tabellenelemente
    uint8_t tablecellalign[4];     // Ausrichtung der Tabellenelemente
    uint8_t nVisibleItem;          // Anzahl maximal sichtbarer Menuepunkte
    uint8_t nTotalItem;            // Anzahl vorhandener Menuepunkte
    MenuItemState_t State;
    Menuconfig_t const* MenuCfg;
    Menuitem_t const *MenuItem;
    uint32_t *hashes;              /**< Hashes der Menueinhalte, um Aenderungen zu erkennen. */
};



/** Hilfsfunktion: Mitscrollen (mitschleppen)
 * @param[in] scrollPos: aktuelle Scroll-Position
 * @param[in] selectedItem: Aktives Element
 * @param[in] nVisibleItem: Anzahl der sichbaren Elementpositionen
 * @return: Neue Scroll-Position */
static uint8_t scroll(uint8_t scrollPos, uint8_t selectedItem, uint8_t nVisibleItem)
{
    while( selectedItem >= scrollPos + nVisibleItem )
    {
        scrollPos++;
    }

    while( selectedItem < scrollPos )
    {
        scrollPos--;
    }

    return scrollPos;
}



/* Breite der Tabellenzellen */
static void calctablecellwidth(uint16_t *tablecellwidth, uint8_t const* tableinput, ifast16_t totalWidth, uint8_t n)
{
    assert( tablecellwidth != NULL );

    if( tableinput == NULL )
    {
        // Default
        for( ufast8_t i=0; i<n; i++ )
        {
            tablecellwidth[i] = 1;
        }
    }
    else
    {
        // uint8_t -> uint16_t
        for( ufast8_t i=0; i<n; i++ )
        {
            tablecellwidth[i] = tableinput[i];
        }
    }
    uarray16_stretchDistance(tablecellwidth, n, totalWidth);
}


/* Ausrichtung der Tabellenzellen */
static void calctablecellalign(uint8_t *tablecellalign, uint8_t const* tableinput, uint8_t n)
{
    assert( tablecellalign != NULL );

    if( tableinput == NULL )
    {
        // Default
        tablecellalign[0] = align_left;
        tablecellalign[1] = align_center;
        tablecellalign[2] = align_right;
        tablecellalign[3] = align_left;
    }
    else
    {
        memcpy(tablecellalign, tableinput, n);
    }
}


void menudraw_reset(Menudrawdata_t *Me)
{
    Me->State.selectedItem = 255; // Ungueltiger Wert, um Update beim ersten
                                // Durchlauf zu erzwingen
}



Menudrawdata_t * menudraw_init(constflash Menuitem_t Menuitem[], constflash Menuconfig_t *MCfg)
{
    Menudrawdata_t *Me = malloc(sizeof(Menudrawdata_t));
    assert( Me != NULL );

    // Darstellungsbereich (eine Zeile)
    // Der Komplettbereich kommt durch nVisibleItem Zeilen zustande
    Rect_t R = menu_private_getTextArea(*(MCfg->Dimension));
    ifast16_t totalHeight = R.h;
    ifast16_t lineHeight = window_getLineHeight();
    R.h = lineHeight;
    Me->Area = R;

    // Anzahl der vorhandenen und darstellbaren Elemente
    Me->nTotalItem = menuitem_arraylen(Menuitem);
    Me->nVisibleItem = totalHeight/lineHeight;


    // Tabellenspalten (Breite)
    ifast16_t totalWidth = Me->Area.w;
    calctablecellwidth(Me->tablecellwidth, MCfg->Dimension->textcellwidth, totalWidth, 4);

    // Tabellenspalten (Ausrichtung)
    calctablecellalign(Me->tablecellalign, MCfg->Dimension->textalign, 4);


    Me->MenuCfg = MCfg;
    Me->MenuItem = Menuitem;

    /* Hashes vorbereiten */
    size_t nHashes = Me->nVisibleItem;
    uint32_t *hashes = malloc(sizeof(uint32_t) * nHashes);
    assert( hashes != NULL );
    Me->hashes = hashes;
    memset(hashes, 0xAAAA, nHashes);

    menudraw_reset(Me);

    return Me;
}


void menudraw_deinit(Menudrawdata_t *Me)
{
    free(Me->hashes);
    free(Me);
    Me = NULL;
}




/** Sichtbare Menuetexte darstellen (inkl. Hervorhebung des aktuellen Texts)
 *   firstItem: Anfang des darstellten Bereichs (Index)
 *   selectedItem:   Hervorgehobenes Element (Index)
 *   height:    Hoehe des Dargestellten Bereichs in Zeilen; von unterer
 *              Displaykante
 * @param[in] doNotForceRedraw: Funktion entscheidet selbst, welche Elemente neugezeichnet werden */
void menu_drawtext(
    Menudrawdata_t *Me,
    constflash Menuitem_t menudata[],
    uint8_t selectedItem,
    bool forceRedraw)
{

    ufast8_t firstItem = Me->State.scrollPos;
    ufast8_t nItem = MIN(Me->nTotalItem, Me->nVisibleItem);
    ufast8_t lastItem  = firstItem + nItem;

    Rect_t R = Me->Area;

    Menucolor_t const* Color      = Me->MenuCfg->Color;
    color_t cursorBorderColor     = Color->cursorFrame;
    color_t cursorBackgroundColor = Color->cursorBackground;

    assert(Me->tablecellalign != NULL );
    assert(Me->tablecellwidth != NULL );



    for( ufast8_t i = firstItem; i<lastItem; i++ )
    {
        /* Ausgewaehltes Element wird immer neugezeichnet */
        //bool activeItem = selectedItem == i;

        /* Bei anderen Elementen wird der Hash-Wert verglichen */
        uint32_t *pLastHash = Me->hashes+(i-firstItem);
        uint32_t thisHash = menuitem_getVal(menudata[i]);
        bool newValue = thisHash != *pLastHash;
        *pLastHash = thisHash;


        if( newValue || forceRedraw )
        {
            // Text+Inhalt fuer Darstellung vorbereiten
            menuitem_snprintf(glTextbuffer, TEXTBUFFERSIZE_BYTES-1, &menudata[i]);

            Colors_t CursorColor = {cursorBorderColor, cursorBackgroundColor};


            if( selectedItem == i )
            {
                /* Aktives Element/ Hervorgehobener Menuepunkt */
                Colors_t Temp = textwindow_getColors();
                Colors_t Textcolor = {Color->cursorText, Color->cursorBackground};
                textwindow_setColors(Textcolor);
                drawRectRounded(rect_clip(R, 0, align_all), 3, CursorColor);

                textarray(R, glTextbuffer, Me->tablecellalign, Me->tablecellwidth, 4);

                CursorColor.bg = transparent;
                drawRectRounded(rect_clip(R, 0, align_all), 3, CursorColor);
                textwindow_setColors(Temp);
            }
            else
            {
                /* Normaler Menupunkt */
                textarray(R, glTextbuffer, Me->tablecellalign, Me->tablecellwidth, 4);
            }
        }

        // Fuer naechste Zeile Position inkrementieren
        R.y += R.h;
    }
}








/* Alles zusammen darstellen (Scroll Area)                                   */
/* Alle oder einige Elemente des Menues darstellen/neuzeichnen
 *  menuGfxEle : Auswahl neuzuzeichnender Elemente */
int menudraw_update(Menudrawdata_t *Me, uint8_t selected)
{

    Menudimension_t const* Dimensions = Me->MenuCfg->Dimension;
    Menucolor_t const* Color          = Me->MenuCfg->Color;
    Menuitem_t const* Menuitem        = Me->MenuItem;

    MenuItemState_t Act =
    {
        .selectedItem = selected,
//        .itemVal = menuitem_getVal(Menuitem[selected]),
        .scrollPos = Me->State.scrollPos,
    };

    Act.scrollPos = scroll(Act.scrollPos, Act.selectedItem, Me->nVisibleItem);


    /* Auswahl neuzuzeichnender Elemente */
    enum menuGfxEle_e redrawElement = menuGfx_none;

    // Feststellen, welche Elemente neu gezeichnet werden muessen
    if( Me->State.selectedItem == 255 )
    {
        // Erster Start
        redrawElement = menuGfx_all;
    }

    if( Act.selectedItem != Me->State.selectedItem )
    {
        redrawElement |= menuGfx_textarea;
    }

//    if( Act.itemVal != Me->State.itemVal )
//    {
//        redrawElement |= menuGfx_activeItem;
//    }

    if( Act.scrollPos != Me->State.scrollPos )
    {
        redrawElement |= menuGfx_scrollbar | menuGfx_textarea;
    }

    Me->State = Act;






    // Kopfzeile/Headline darstellen
    if( redrawElement & menuGfx_headline )
    {
        HeadlineColor_t CHl = getHeadlineColor(*(Color));

        // Hintergrund loeschen
        Rect_t HlBg           = menu_private_getHeadBackgroundArea(*Dimensions);

        int16_t radius = Dimensions->cornerRadius;
        drawRectRounded(HlBg, radius, (Colors_t) {transparent, CHl.background});


        Rect_t Hl           = menu_private_getHeadTextArea(*Dimensions);

        HeadlineConfig_t CfgHl =
        {
            .margin = Dimensions->margin,
            .cornerRadius = Dimensions->cornerRadius,
            .text = Me->MenuCfg->headtext,
            .iconLeft = Me->MenuCfg->iconLeft,
            .iconRight = Me->MenuCfg->iconRight,
            .iconFont = Me->MenuCfg->iconFont,
        };
        uiDraw_headline(Hl, &CfgHl, &CHl);
    }

    // Hintergrund loeschen
    if( redrawElement & menuGfx_textarea )
    {

        color_t color = Color->mainBackground;
        if( redrawElement != menuGfx_all )
        {
            // Scrollbar-Bereich aussparen
            Rect_t Bg = menu_private_getTextArea(*Dimensions);
            drawRectF(Bg, color);
        }
        else
        {
            Rect_t Bg = menu_private_getBodyArea(*Dimensions);
            int16_t r = Dimensions->cornerRadius;
            assert( Bg.w >= 2*r && Bg.h >= 2*r );
            drawRectRounded(Bg, r, (Colors_t) {color, color});
        }
    }


    /* Scrollbereich neuzeichnen (nur geaenderte Werte oder alles)
     * Geaenderte Werte erkennt die Zeichenfunktion selbst */
    bool forceRedraw = (redrawElement & menuGfx_textarea);
    menu_drawtext(Me, Menuitem, selected, forceRedraw);


    // Scrollbereich neuzeichnen
    if( redrawElement & menuGfx_textarea )
    {
        //uint16_t x = (MAct->selectedItem-MAct->scrollPos)*Mid->lineheight+Dimensions->headheight;
        //uiDraw_cursor(x);
    }


    /* Scrollbar darstellen */
    if( redrawElement & menuGfx_scrollbar )
    {
        Rect_t Sb            = menu_private_getScrollbarArea(*Dimensions);
        ScrollbarColor_t CSb = getScrollbarColor(*Color);

        ifast8_t nVisibleItem = MIN(Me->nTotalItem, Me->nVisibleItem);
        ifast8_t nScrollpos = Me->nTotalItem - nVisibleItem;
        uiDraw_scrollbar(Sb, Act.scrollPos, nScrollpos, &CSb);
    }



    return EXIT_SUCCESS;
}



