/** @file menu_color_private_item.h
 *
 * Menuesystem (Hilfsfunktionen): Behandung von Menu-Items
 *
 * Darstellung, Inkrement, Vergleich
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * 13.03.2018 Nicolas  */


#ifndef MENU_COLOR_PRIVATE_ITEM_H
    #define MENU_COLOR_PRIVATE_ITEM_H

    #include "menu_color.h"
    #include <stdio.h>
    #include <string.h>
    #include "pgmspace.h"



    /** Menueeintraege zaehlen
     *
     * @param[in] Menudata: Menu-Elemente-Tabelle
     * @return Anzahl der Menueelement (ohne Terminator) */
    static inline int menuitem_arraylen(constflash Menuitem_t Menudata[])
    {
        int n = 0;
        while( 1 )
        {
            const char *s = Menudata[n].menutext;
            if( s == NULL || strlen_P(s) == 0 || Menudata[n].All.type == endmarker )
            {
                break;
            }
            else
            {
                n++;
            }
        }
        return n;
    }



    /** Von Menu-Element ausgeloeste Aktion */
    typedef enum menuitem_caller_action_e
    {
        doReset,
        doNothing,
        doExit
    }
    menuitem_caller_action_t;



    /** Menuezeile ausfuehren:
     * - Funktionen werden ausgefuehrt
     * - Variablen werden mit dem Encoder bearbeitet
     *
     * @param[in] Item: Menuezeile
     * @param[in] enc_diff: EncoderInkrement fuer Variablenbearbeitung
     * @return:   Weitere Aktion in Menue-Hauptschleife ausloesen */
    static inline menuitem_caller_action_t menuitem_execute(Menuitem_t Item, int_fast8_t enc_diff)
    {
        menuitem_caller_action_t callerAction; /* Rueckgabewert */


        switch( Item.All.type )
        {
            case nothing: /* Kann ignoriert werden */
                callerAction = doReset;
                break;

            case voidFcn: /* Funktionsausfuehrung oder Menue-Beenden initiieren */
                {
                    void (*nextfkt)(void) = Item.VoidFcn.ptr;
                    if( nextfkt == NULL )
                    {
                        callerAction = doExit;
                    }
                    else
                    {
                        nextfkt();
                        callerAction = doReset;
                    }
                }
                break;

            case uint8: /* uint8-Variable bearbeiten */
                {
                    uint8_t *pu8 = Item.Uint8.ptr;
                    int_fast8_t incr = satmul_s8(enc_diff, Item.Uint8.inc);
                    *pu8 = satakku_u8(*pu8, incr);
                    clamp_u8(pu8, Item.Uint8.min, Item.Uint8.max);
                }
                callerAction = doNothing;
                break;

            case uint16: /* uint16-Variable bearbeiten */
                {
                    uint16_t *pu16 = Item.Uint16.ptr;
                    int_fast8_t incr = satmul_s8(enc_diff, Item.Uint16.inc);
                    *pu16 = satakku_u16(*pu16, incr);
                    clamp_u16(pu16, Item.Uint16.min, Item.Uint16.max);
                }
                callerAction = doNothing;
                break;

            case uint32: /* uint32-Variable bearbeiten */
                {
                    uint32_t *pu32 = Item.Uint32.ptr;
                    int32_t incr = satmul_s32(enc_diff, Item.Uint32.inc);
                    *pu32 = satakku_u32(*pu32, incr);
                    clamp_u32(pu32, Item.Uint32.min, Item.Uint32.max);
                }
                callerAction = doNothing;
                break;

            case int8: /* int8-Variable bearbeiten */
                {
                    int8_t *pi8 = Item.Int8.ptr;
                    int_fast8_t incr = satmul_s8(enc_diff, Item.Int8.inc);
                    *pi8 = satadd_s8(*pi8, incr);
                    clamp_s8(pi8, Item.Int8.min, Item.Int8.max);

                }
                callerAction = doNothing;
                break;

            case int16: /* int16-Variable bearbeiten */
                {

                    int16_t *pi16 = Item.Int16.ptr;
                    int_fast8_t incr = satmul_s8(enc_diff, Item.Int16.inc);
                    *pi16 = satadd_s16(*pi16, incr);
                    clamp_s16(pi16, Item.Int16.min, Item.Int16.max);
                }
                callerAction = doNothing;
                break;

            case int32: /* int32-Variable bearbeiten */
                {

                    int32_t *pi32 = Item.Int32.ptr;
                    int32_t incr = satmul_s32(enc_diff, Item.Int32.inc);
                    *pi32 = satadd_s32(*pi32, incr);
                    clamp_s32(pi32, Item.Int32.min, Item.Int32.max);
                }
                callerAction = doNothing;
                break;

            #if MENU_SUPPORT_FLOAT
            case flin32:
                if( enc_diff )
                {
                    /* Float-Variable mit linearem Inkrement:
                       Float-Variable nur bearbeiten, wenn ein Inkrement vorliegt. Wird sie
                       bearbeitet, wird sie auch auf dem Raster eingerastet. */
                    float *pf32 = Item.Float32.ptr;
                    float temp = enc_diff;
                    float inc = Item.Float32.inc;
                    temp *= inc;
                    temp += *pf32;
                    fclampf(&temp, Item.Float32.min, Item.Float32.max);

                    /* Auf dem Raster einrasten */
                    temp = round(temp/inc) * inc;

                    *pf32 = temp;
                }
                callerAction = doNothing;
                break;

            case flog32:
                {
                    /* Float-Variable mit logarithmischem/exponenntiellen Inkrement
                       Das exponnentielle Inkrement wird durch mehrfahre Multiplikation
                       ausgefuehrt. */
                    float *pf32 = Item.Float32.ptr;
                    int_fast8_t diff = enc_diff;
                    float inc = Item.Float32.inc;
                    float temp = *pf32;
                    while( diff > 0 )
                    {
                        temp*= inc;
                        diff--;
                    }
                    while( diff < 0 )
                    {
                        temp/= inc;
                        diff++;
                    }
                    fclampf(&temp, Item.Float32.min, Item.Float32.max);
                    *pf32 = temp;
                }
                callerAction = doNothing;
                break;

            #endif

            case strlist: /* String-Listen-Index bearbeiten */
                {
                    /* Die untere Grenze ist immer Null */
                    uint8_t *pu8 = Item.Strlist.idx;
                    int_fast8_t incr = satmul_s8(enc_diff, Item.Strlist.inc);
                    *pu8 = satakku_u8(*pu8, incr);
                    clamp_u8(pu8, 0, Item.Strlist.len-1);
                }
                callerAction = doNothing;
                break;

            case exitMenu:
                callerAction = doExit;
                break;

            default:
                error("unexpected datatype descriptor");
                callerAction = doNothing;
        } /* Ende switch datatype */

        return callerAction;
    }



    /** Inhalt des Menuitem-Zeigers als Bitmuster zurueckgeben
     *
     * Dies eignet sich insbesondere fuer den Vergleich, ob sich der Inhalt geaendert hat.
     * Da externe Funktionen eingebunden werden koennen, kann sich der Wert auch aendern,
     * wenn das Inkrement Null betraegt.
     *
     * @param[in] Item: Menuezeile
     * @return:   Menu-Variablen-Inhalt mit Bitmuster als Ersatz fuer Hashwert */
    static inline uint32_t menuitem_getVal( Menuitem_t Item )
    {
        uint32_t retval = 0;
        switch( Item.All.type )
        {
            case nothing:    /* Fallthru */
            case voidFcn:    /* Fallthru */
            case endmarker:  /* Fallthru */
            case exitMenu:

                break;

            case uint8:
                retval |= *(Item.Uint8.ptr);
                break;

            case uint16:
                retval |= *(Item.Uint16.ptr);
                break;

            case uint32:
                retval |= *(Item.Uint32.ptr);
                break;

            case int8:
                retval |= *(Item.Int8.ptr);
                break;

            case int16:
                retval |= *(Item.Int16.ptr);
                break;

            case int32:
                retval |= *(Item.Int32.ptr);
                break;


            #if MENU_SUPPORT_FLOAT
            case flin32:
            case flog32:
                {
                    /* Type-punning fuer beide Varianten von Float-Item */
                    float tempF = *(Item.Float32.ptr);
                    uint32_t tempI = 0;
                    static_assert( sizeof(tempF) == sizeof(tempI), "Integer size" );
                    memcpy(&tempI, &tempF, sizeof(tempF));
                    retval |= tempI;
                }
                break;

            #endif

            case strlist:
                retval |= *(Item.Strlist.idx);
        }

        return retval;
    }




    /** Inhalt eines Menu-Items als Text interpretieren
     *
     * @param[out] s: Ausgabe-Textstring
     * @param[in] n: Groesse des Ausgabe-Textstrings
     * @param[in] Item: Menuezeile/Menuelement
     * @return:   EXIT_SUCCESS */
    static inline exit_t menuitem_snprintf(char * s, size_t n, constflash Menuitem_t *Item)
    {

        switch( Item->All.type )
        {
            case nothing:  /* Fallthru */
            case exitMenu: /* Fallthru */
            case voidFcn:
                /* Menutexte fuer Funktionen werden einfach dargestellt */
                strncpy_P(s, Item->menutext, n);
                s[n] = '\0';
                break;

            case uint8:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext,	*Item->Uint8.ptr);
                break;

            case uint16:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext, *Item->Uint16.ptr);
                break;

            case uint32:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext, *Item->Uint32.ptr);
                break;

            case int8:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext, *Item->Int8.ptr);
                break;

            case int16:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext, *(Item->Int16.ptr) );
                break;

            case int32:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext, *(Item->Int32.ptr) );
                break;

            #if MENU_SUPPORT_FLOAT
            case flin32: /* fallthru */
            case flog32:
                /* Texte fuer Einstellkonstanten werden mit printf interpretiert */
                wasnprintf_P(s, n, Item->menutext, *(Item->Float32.ptr) );
                break;
            #endif

            case strlist:
                ; /* Enum-artige List: Passendes Listenelement heraussuchen */
                uint8_t idx = *Item->Strlist.idx;
                const char **pss = Item->Strlist.str;
                const char *ss = *(pss+idx);
                wasnprintf_P(s, n, Item->menutext, ss);
                break;

            case endmarker:
                ;

            default:
                /* Unbekannten Typ mit Fehlerdarstellung */
                wasnprintf_P(s, n, "XXX");
        }
        return EXIT_SUCCESS;
    }

#endif

