/** @file wallclock.h
 *
 * Uhrzeiten, Timeouts etc.
 *
 *
 * $Date: 2020-08-12 16:39:52 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 66 $
 *
 * @date 12.02.2018 @author Nicolas */


#ifndef WALLCLOCK_H_INCLUDED
    #define WALLCLOCK_H_INCLUDED


    #include <stdbool.h>
    #include <stdint.h>

    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif



    /** Wallclock: Zeit aufgeloest in Millisekunden, Sekunden, Minuten, Stunde, Tage
     *
     * 2^32-1 Millisekunden sind 49 d, 17 h, 2 min, 47 s und 295 ms */
    typedef struct Wallclock_time_s
    {
        uint16_t ms; ///< Millisekunden
        uint8_t s;   ///< Sekunden
        uint8_t m;   ///< Minuten
        uint8_t h;   ///< Stunden
        uint8_t d;   ///< Tage
    }
    Wallclock_time_t;

    typedef uint32_t wallclock_t;



    /** Der Zaehler wird global (z.B. in main.c) definiert und in einer Timer-ISR im
       Millisekundenabstand inkrementiert. Er laeuft nach 49.71 Tagen zum ersten Mal ueber. */
    extern volatile uint32_t counter_milliseconds;



    /** Tic-toc-Paar: Tic
     * @return Zeitstempel fuer wallclock_toc() */
    static_inline wallclock_t wallclock_tic(void)
    {
        return counter_milliseconds;
    }



    /** Tic-toc-Paar: Toc
     * @param[in] tic_ms: Zeitstempel von wallclock_tic()
     * @return Zeit seit dem Aufruf von tic() in Millisekunden */
    static_inline int32_t wallclock_toc_ms(wallclock_t tic_ms)
    {
        return (int32_t) (counter_milliseconds - tic_ms);
    }



    /** Aktuellen Zeitstempel auslesen
     * @return Zeitstempel in Millisekunden */
    static_inline int32_t wallclock_getMilliseconds(void)
    {
        return counter_milliseconds;
    }



    /** Timeout setzen
     * @param[in] Timeout in Millisekunden
     * @return Zeitstempel des Timeouts */
    static_inline wallclock_t wallclock_setTimeout(uint32_t ms)
    {
        return counter_milliseconds + ms;
    }



    /** Timeout pruefen
     * @param[in] Zeitstempel des Timeouts
     * @return Timeout erreicht true|false */
    static_inline bool wallclock_checkTimeout(uint32_t timeout)
    {
        int32_t diff = timeout - counter_milliseconds;
        return diff <= 0;
    }



    Wallclock_time_t wallclock_ms2wallclocktime(uint32_t ms);


#endif /* WALLCLOCK_H_INCLUDED */
