/** @file mocksdl_input.c
 *
 * SDL-Mockup: Einfache Emulation des Verhaltens der Eingabegeraete ueber SDL2
 *
 * Der Drehgeber wird ueber die Tastatur emuliert. Einer Rechtsdrehung
 * entspricht die Pfeil-Hoch-Taste, einer Linksdrehung die Pfeil-Herunter-Taste.
 * Dem Tastendruck entspricht die Enter-Taste.
 *
 * Der Event-Poll beendet das Programm mit exit(), falls das Fenster geschlossen
 * wird.
 *
 * $Date: 2020-08-12 16:39:52 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 66 $
 *
 * @date 20.04.2015 @author Nicolas */

#include "mocksdl_input.h"
#include "sw_config.h"



#if defined( INCLUDE_MOCKUP_SDL )

    #include <stdint.h>
    #include <stdio.h>
    #include <stdbool.h>
    #include <SDL2/SDL.h>
    #include "sw_config.h"
    #include "hw_config.h"
    #include "io.h"



    /* ****************************** Globale Variablen ***************************************** */

    /** Encoder-Emulation */
    uint_fast8_t glSdlenc = 0;

    /** Taster-Emulation */
    bool glSdlkey = false;

    /** Dauer Tastendruck */
    volatile uint_fast16_t glSdlTime;

    /** Zustandsvariable pushbuttons - Auswirkungen auf Pins koennen hier injiziert werden */
    extern volatile uint_fast16_t pinstate;



    /** SDL-Mockup/Input: Initialisierung
     *
     *  Wenn ein SDL-Fenster offen ist, bedarf die Tastatureingabe keiner  Initialisierung. */
    void mock_input_init(void) { }



    /** SDL-Mockup/Input: Deinitialisierung
     *
     * Es ist keine Deinitialisierung notwendig. */
    void mock_input_deinit(void) { }



    /** Pollen und Interpretation der der Tastatureingaben und Emulation der Auswirkungen */
    void mock_input_poll(void)
    {

        SDL_Event event;
        const uint8_t *keystate = SDL_GetKeyboardState(NULL);

        //SDL_WaitEvent(&event);
        while(SDL_PollEvent(&event))
        {

            switch(event.type)
            {
                case SDL_QUIT:
                    printf("Window closed\n");
                    exit(0);
                    break;

                case SDL_KEYDOWN:

                    /* Keystate wird durch SDL_PumpEvents()/SDL_PollEvent upgedatet */


                    /* Pfeiltasten emulieren Drehgeber */
                    if (keystate[SDL_SCANCODE_UP] || keystate[SDL_SCANCODE_RIGHT])
                    {
                        glSdlenc++;
                    }
                    if (keystate[SDL_SCANCODE_DOWN] || keystate[SDL_SCANCODE_LEFT])
                    {
                        glSdlenc--;
                    }

                    /* Return simuliert kurzen Tastendruck */
                    if (keystate[SDL_SCANCODE_RETURN])
                    {
                        glSdlTime+= 1;
                    }

                    /* Shift+Return simuliert langen Tastendruck */
                    if ( keystate[SDL_SCANCODE_RETURN] && \
                            (keystate[SDL_SCANCODE_LSHIFT] || keystate[SDL_SCANCODE_RSHIFT]) )
                    {
                        glSdlTime+= T_LONGKEYPRESS_MS+1;

                    }

                    /* L : Toggle Locknut-Switch */
                    if (keystate[SDL_SCANCODE_L])
                    {
                        #if N_SWITCH > 1
                            printf("L: Toggle Locknut Switch\n");
                            io_toggleBit(SWITCH4_PIN);
                        #endif
                    }

                    /* E : Toggle E-Stop-Switch */
                    if (keystate[SDL_SCANCODE_E])
                    {
                        #if N_SWITCH > 2
                            printf("E: Toggle ESTOP Switch\n");
                            io_toggleBit(SWITCH3_PIN);
                        #endif
                    }

                    /* 1 : Soft Key 1 */
                    if (keystate[SDL_SCANCODE_1])
                    {
                        #if N_SWITCH > 1
                            printf("1: Soft Key 1 toggeled\n");
                            pinstate ^= SWITCH_KEY1;
                        #endif
                    }

                    /* 2 : Soft Key 2 */
                    if (keystate[SDL_SCANCODE_2])
                    {
                        #if N_SWITCH > 2
                            printf("2: Soft Key 2 toggeled\n");
                            pinstate ^= SWITCH_KEY2;
                        #endif
                    }
                    break;
            }
        }
    }


#else
    void mock_input_poll(void) { }

#endif
