/** @file myassert.c
 *
 * Re-Implementierung der assert()- und error()-Funktionalitaet, um targetspezifische
 * Eigenheiten zu beruecksichtiten.
 *
 * Insbesondere
 *  - Notwendigkeit eines manuellen Refreshs der Ausgabe
 *  - Abschaltmoeglichkeit fuer automatische Modultests
 *
 * @todo: Die Doku koennte sollte ueberarbeitet werden
 * @todo: Der Zweig ohne Einbeziehung der Modultests ist mittlerweile obsolet
 *
 * $Date: 2020-08-13 11:28:54 +0200 (Do, 13 Aug 2020) $
 * $Revision: 68 $
 *
 * @date 30.04.2014 @author Nicolas */


/* Die Funktionen assert_P_dontCallMe() und error_P_dontCallMe() werden nie
 * direkt aufgerufen, sondern nur ueber die Makros assert() und error()!
 *
 * Sie zeigen ein komplett gegenteiliges Verhalten bei Modultests/Unit-Tests:
 * Whrend im Normalfall ein error() oder ein fehlgeschlagenes assert() das letzte ist,
 * was sein System vor dem unwiderbringlichen Halt noch ausgibt, sind die Triggerung
 * von error() und assert() testbare Zustnde.
 */


#include <stdio.h>
#include "pgmspace.h"
#include "myassert.h"


bool errorconditionPrepared = false;
bool errorconditionTriggered = false;

#if( (PC_UNITTEST==1 || TARGET_UNITTEST==1) || 1)
    #include "unity_fixture.h"




    void dealWithPreparedError(void)
    {
        // Erwarteter Fehler im Rahmen eines Modultests aufgetreten
        // Funktioniert nur, wenn vorher mit TEST_PROTECT() eine Sprungmarke gesetzt
        // wurde.
        errorconditionPrepared = false;
        errorconditionTriggered = true;
        TEST_ABORT();
    }

    /* Fuer einen einzigen Fehler (error oder assertion failed) vorbereiten */
    void prepareErrorCondition(void)
    {
        errorconditionPrepared = true;
    }


#else

    // Im Nicht-Modultest-Modus koennen Fehler nicht vorbereitet werden
    // Diese Funktion duerfte niemals nie aufgerufen werden
    // (exploding fake)
    void dealWithPreparedError(void)
    {
        triggerHardFault();
    }

    /* Ein Fehler laesst sich im Zielcode nicht vorbereiten */
    void prepareErrorCondition(void)
    {
    }


#endif





extern void abort(void);




/* Fehlermeldung darstellen und abort()
 * f: Funktionsname. (__func__)
 *    __func__ ist eine builtin-Funktion und kein Makro, kann also nicht mit dem
 *    PSTR-Makro zusammen genutzt werden. Deshalb ist der Datentyp const char *
 *    und nicht constflash char *.
 * l: Zeilennummer (__LINE__)
 * e: Assertion-Text (z.B. " 0==1 ")   */
int assert_P_dontCallMe(const char *f, const int l, constflash char *e)
{
    if( errorconditionPrepared )
    {
        dealWithPreparedError();
        return 0;
    }
    else
    {
        //fprintf_P(stderr,PSTR("%s(),%d assertion failed: '%s' \a"),f,l,e);
        fprintf_P(stderr,PSTR("%s(),%d assertion failed: '"), f, l);
        fputs_P(e,stderr);
        fprintf_P(stderr,PSTR("' \a"));
        abort();
        return 0;
    }
}


/* Fehlermeldung darstellen und abort()
 * f: Funktionsname. (__func__)
 *    __func__ ist eine builtin-Funktion und kein Makro, kann also nicht mit dem
 *    PSTR-Makro zusammen genutzt werden. Deshalb ist der Datentyp const char *
 *    und nicht constflash char *.
 * l: Zeilennummer (__LINE__)
 * e: Fehlertext (z.B. " dumm gelaufen ")   */
void error_P_dontCallMe(const char *f, const int l, constflash char *e)
{
    if( errorconditionPrepared )
    {
        dealWithPreparedError();
    }
    else
    {
        //fprintf_P(stderr,PSTR("%s(),%d error: %s \a"),f,l,e);
        // Ausgabe muss fuer Flash-Strings zerpflueckt werden:
        fprintf_P(stderr,PSTR("%s(),%d error: "), f, l);
        fputs_P(e,stderr);
        fprintf_P(stderr,PSTR(" \a"));
        abort();
    }
}





// Test Spies



bool isErrorConditionTriggered(void)
{
    return errorconditionTriggered;
}

void clearErrorCondition(void)
{
    errorconditionPrepared = false;
    errorconditionTriggered = false;
}



void triggerHardFault(void)
{
    #ifdef __MINGW32__
        int *a = NULL;
        int b = *a;
        b++; // Warning "unused variable" unterdruecken
    #else
        asm volatile("udf");
    #endif
}
