/** @file nvram.c
 *
 * Nichtfluechtiger Speicher
 *
 * Behandlung des nichtfluechtigen Speichers, insbesondere Lesen und Speichern
 * von Konfigurationsdaten. Es werden Pruefsummen gebildet und ueberprueft.
 * Das bedingt einen Overhead von 10 Bytes je Datensatz - bei der Bildung
 * von Adressen ist das zu beruecksichtigen.
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * @date 04.05.2015 @author Nicolas */


#include "nvram.h"
#include "crc.h"
#include <string.h>
#include <stdlib.h>
#include "myassert.h"



#ifdef AVR
    #include <avr/eeprom.h>

#elif defined( STM32F10x ) || defined( STM32F4XX )
    #include "eeprom_24Cx.h"

#elif defined( __MINGW64__ ) || defined( __MINGW32__ )
  	#include "eeprom_24Cx.h"

#else
    #error "unknown Platform"

#endif


#define MAGICNUMBER0 0xDEADu
#define MAGICNUMBER1 0xBEEFu

/* | PREAMBLE | .... Datensatz .... | CRC16 | POSTAMBLE | */
#define NVRAM_OVERHEAD 10U



/** NVRAM: Initialisierung (dummy) */
uint_fast8_t nvram_init(void)
{
    return EXIT_SUCCESS;
}



/** NVRAM: Datensatz aus dem EEPROM lesen und ueberpruefen.
 *
 * Der Puffer ist im nicht-erfolgreichen Fall mit ungueltigen Daten ueberschrieben.
 *
 * @param{out] buffer: Puffer fuer Rueckgabe, min. Groesse "size"
 * @param[in]  address: Speicheradresse des gewuenschten Datensatzes
 * @param[in]  size   : Groesse des gewuenschten Datensatzes
 * @return : EXIT_SUCCESS|EXIT_FAILURE */
uint_fast8_t nvram_read(uint8_t *buffer, uint_fast8_t address, uint_fast8_t size)
{
	/* preamble, postamble, und crc1 werden als Puffer fester Breite verwendet.*/
    uint16_t preamble;
	uint16_t postamble;
	uint16_t crc1, crc2;


    #ifdef AVR
        /* Auf dem AVR sind die EEPROM-Routinen zeigerbasiert */
        uint8_t *addr = (uint8_t*)(size_t) address;

    #else
        /* Ueberall sonst sind die EEPROM-Adressen normale Integer */
        uint_fast8_t addr = address;
        uint_fast8_t status = 0;

    #endif


	/* Daten aus EEPROM auslesen. Fehlercode wird nicht ausgewertet, da:
       - Die AVR-EEPROM-Routinen keinen Rueckgabewert haben und
       - Fehler  beim Auslesen aus dem Datensatz selbst erkannt werden. */
	#ifdef AVR
		eeprom_read_block((uint8_t *) &preamble, addr, 2);
		addr += 2;
		eeprom_read_block(buffer, addr, size);
		addr += size;
		eeprom_read_block((uint8_t *) &crc1, addr, 2);
		addr += 2;
        eeprom_read_block((uint8_t *) &postamble, addr, 2);

	#else
		status |= eeprom_read_block((uint8_t *) &preamble, addr, 2);
		addr += 2;
		status |= eeprom_read_block(buffer, addr, size);
		addr += size;
		status |= eeprom_read_block((uint8_t *) &crc1, addr, 2);
		addr += 2;
		status |= eeprom_read_block((uint8_t *) &postamble, addr, 2);

		if(status)
        {
			return nvram_i2c_failure;
		}

	#endif

	/* Ueberpruefung Praeambel und Postscriptum */
	if( preamble != MAGICNUMBER0 || postamble != MAGICNUMBER1 )
    {
		return nvram_magic_failure;
	}

	/* Ueber den restlichen Teil CRC berechnen. Die letzten zwei Bytes sind die CRC des
	   Datensatzes, also muss der Gesamt-CRC Null ergeben. */
	crc2 = crc16_calc(buffer, size);
	if(crc1 != crc2)
    {
		return nvram_crc_failure;
	}
	else
    {
		return nvram_ok;
	}
}



/** NVRAM: Daten in EEPROM schreiben und ueberpruefen
 *
 * Die EEPROM-Daten belegen etwas mehr Speicherplatz als der eigentliche Datensatz. Layout:
 *  PREAMBLE | DATASET | CRC16 | POSTAMBLE
 * Nach dem Schreiben wird der Inhalt ueberprueft. Wenn das schreiben fehlerhaft war, sind die
 * alten Konfigurationsdaten im EEPROM leider verloren.
 *
 * @param[in] dataset: Zu schreibender Datensatz
 * @param[in] address : gewuenschte Speicheradresse
 * @param[in] size    : Groesse des Datensatzes in Bytes
 * @return : EXIT_SUCCESS|EXIT_FAILURE */
uint_fast8_t nvram_write(uint8_t *dataset, uint_fast8_t address, uint_fast8_t size)
{
	/* preamble, postamble, crc und buffer werden als Puffer fester Breite verwendet. */
	uint16_t preamble = MAGICNUMBER0;
	uint16_t postamble = MAGICNUMBER1;
	uint16_t crc;
    uint8_t buffer[size];
    uint_fast8_t blksize = 0, status;

    #ifdef AVR
        /* Auf dem AVR sind die EEPROM-Routinen zeigerbasiert */
        uint8_t *addr = (uint8_t*)(size_t) address;

    #else
        /* Ueberall sonst sind die EEPROM-Adressen normale Integer */
        uint_fast8_t addr = address;

    #endif

    /* Praeambel schreiben */
	blksize = 2;
    eeprom_update_block((uint8_t *) &preamble, addr, blksize);
 	addr += blksize;

	/* Es folgt der eigentliche Datensatz */
	blksize = size;
	eeprom_update_block(dataset, addr, blksize);
	addr += blksize;

	/* Ueber den Datensatz (ohne Praeambel) wird eine CRC16 berechnet und angehaengt */
	blksize = 2;
    crc = crc16_calc(dataset, size);
	eeprom_update_block((uint8_t *) &crc, addr, blksize);
	addr += blksize;

    /* Postscriptum schreiben */
	blksize = 2;
    eeprom_update_block((uint8_t *) &postamble, addr, blksize);
 	addr += blksize;


	/* Der Erfolg der Schreiboperation wird ueberprueft, indem der Datensatz noch einmal
	   eingelesen wird. */
	status =  nvram_read(buffer, address, size);
	if ( status != EXIT_SUCCESS )
    {
		return EXIT_FAILURE;
	}


	/* Der Check in nvram_read sollte jeden Lesefehler ausschliessen - schlaegt der Vergleich mit
	   dem urspruenglichen Datensatz also fehl, liegt ein Programmfehler vor. */
	status = (memcmp(buffer, dataset, size) != 0);
	if ( status )
    {
		error("Error checking NVRAM data");
	}

    return EXIT_SUCCESS;
}




