/** @file spi_bitbang.c
 *
 * SPI-Master per Bitbanging
 *
 * Verwendbar fuer alle GPIOs anan AVR oder STM32. Der SPI-Treiber kuemmert sich nicht um
 * Chip-Select. Es koennen unmittelbar mehrere SPI-Kanaele mit unterschiedlichen Pinous parallel
 * betrieben werden.
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * 14.02.2017 Nicolas */


#include "spi_bitbang.h"
#include "hw_config.h"
#include "io.h"
#include "delay.h"
#include "myassert.h"
#include "nstdlib.h"




/** SPI Bitbang: Pins initialisieren
 *
 * @param[in] SCK_GPIOx, SCK_Pin: Clock-Pin
 * @param[in] MISO_GPIOx, MISO_Pin: MISO-Pin
 * @param[in] MOSI_GPIOx, MOSI_Pin: MOSI-Pin
 * @return EXIT_SUCCESS */
exit_t spi_bitbang_init(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                        GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                        GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin,
                        enum spi_mode_e mode)
{

    io_setOutput(SCK_GPIOx,  SCK_Pin);
    io_setOutput(MOSI_GPIOx, MOSI_Pin);
    io_setInput(MISO_GPIOx,  MISO_Pin);

    mode &= 0x03; /* MSB, LSB hier uninteressant */
    switch( mode )
    {
        case spi_cpol0_cpha0: // Fallthru
        case spi_cpol0_cpha1:
            io_clearBit(MOSI_GPIOx, MOSI_Pin);
            io_clearBit(SCK_GPIOx,  SCK_Pin);
            break;

        case spi_cpol1_cpha0: // Fallthru
        case spi_cpol1_cpha1:
            io_setBit(MOSI_GPIOx, MOSI_Pin);
            io_setBit(SCK_GPIOx,  SCK_Pin);
            break;
    }

    return EXIT_SUCCESS;
}



/** SPI Bitbang: Pins deinitialisieren (alle Pins auf Eingang schalten)
 *
 * @param[in] SCK_GPIOx, SCK_Pin: Clock-Pin
 * @param[in] MISO_GPIOx, MISO_Pin: MISO-Pin
 * @param[in] MOSI_GPIOx, MOSI_Pin: MOSI-Pin
 * @return EXIT_SUCCESS */
exit_t spi_bitbang_deinit(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                     GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                     GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin)
{

    io_setInput(MOSI_GPIOx, MOSI_Pin);
    io_setInput(SCK_GPIOx,  SCK_Pin);
    io_setInput(MISO_GPIOx, MISO_Pin); /* wobei das noch Input sein sollte */
    return EXIT_SUCCESS;
}



/** SPI Bitbang: Dummy-Clock fuer bestimmte Devices, z.B. ILI9341
 *
 * funktioniert fuer alle CPOLx, CPHAx
 *
 * @param[in] SCK_GPIOx, SCK_Pin: Clock-Pin
 * @param[in] MISO_GPIOx, MISO_Pin: MISO-Pin
 * @param[in] MOSI_GPIOx, MOSI_Pin: MOSI-Pin
 * @return EXIT_SUCCESS */
void spi_bb_dummyclock(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                        GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                        GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin)
{
    (void) MISO_GPIOx;
    (void) MISO_Pin;
    (void) MOSI_GPIOx;
    (void) MOSI_Pin;

    /* nutzt das ueberhaupt wer? */
    io_toggleBit(SCK_GPIOx, SCK_Pin);
    asm volatile("nop");
    asm volatile("nop");
    io_toggleBit(SCK_GPIOx, SCK_Pin);
}
