/** @file spi_bitbang.h
 *
 * SPI-Master per Bitbanging
 *
 * Verwendbar fuer alle GPIOs anan AVR oder STM32. Der SPI-Treiber kuemmert sich nicht um
 * Chip-Select. Es koennen unmittelbar mehrere SPI-Kanaele mit unterschiedlichen Pinous parallel
 * betrieben werden.
 *
 * $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 * $Revision: 12 $
 *
 * 14.02.2017 Nicolas */


#ifndef SPI_BITBANG_H
	#define SPI_BITBANG_H

	#include "nstdlib.h"
	#include "io.h"


    #ifndef static_inline
        #define static_inline static inline __attribute__((always_inline))
    #endif



    /** SPI-Modi */
	enum spi_mode_e
	{
        spi_cpol0_cpha0 = 0, /**< SPI-Mode 0 */
        spi_cpol0_cpha1 = 1, /**< SPI-Mode 1 */
        spi_cpol1_cpha0 = 2, /**< SPI-Mode 2 */
        spi_cpol1_cpha1 = 3, /**< SPI-Mode 3 */
	};



    /** SPI Bitbang - Mode 0 (CPOL0, CPHA0) MSB First - Einzelnes Byte ueber SPI senden und
     * empfangen
     *
     * @param[in] SCK_GPIOx, SCK_Pin: Clock-Pin
     * @param[in] MISO_GPIOx, MISO_Pin: MISO-Pin
     * @param[in] MOSI_GPIOx, MOSI_Pin: MOSI-Pin
     * @return EXIT_SUCCESS */
    static_inline uint8_t spi_bitbang_msb_mode0(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                            GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                            GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin,
                            uint8_t out)
    {

        uint8_t in = 0;

        /* Mode 0 schreiben und lesen */
        for( uint_fast8_t i = 8; i > 0; i-- )
        {
            /* Daten-Bit setzen */
            if( out & 0x80 )
                io_setBit(MOSI_GPIOx, MOSI_Pin);
            else
                io_clearBit(MOSI_GPIOx, MOSI_Pin);

            out<<=1;

            //asm volatile("nop");

            /* Clock erzeugen */
            io_setBit(SCK_GPIOx, SCK_Pin);
            //asm volatile("nop");

            /* Daten einlesen */
            in<<=1;
            in |= io_bitIsSet(MISO_GPIOx, MISO_Pin) != 0;

            //asm volatile("nop");
            io_clearBit(SCK_GPIOx, SCK_Pin);

            /* Noch ein wenig halten */
            //asm volatile("nop");
        }
        return in;
    }



    /** SPI Bitbang - Mode 0 (CPOL0, CPHA0) LSB First - Einzelnes Byte ueber SPI senden
     * @note: Empfang noch nicht implementiert
     *
     * @param[in] SCK_GPIOx, SCK_Pin: Clock-Pin
     * @param[in] MISO_GPIOx, MISO_Pin: MISO-Pin
     * @param[in] MOSI_GPIOx, MOSI_Pin: MOSI-Pin
     * @return EXIT_SUCCESS */
    static_inline uint8_t spi_bitbang_lsb_mode0(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                            GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                            GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin,
                            uint8_t out)
    {
        /* RX noch nicht implementiert */
        (void) MISO_GPIOx;
        (void) MISO_Pin;


        uint8_t in = 0;

        for(uint_fast8_t i = 0; i < 8; i++)
        {
            /* Daten-Bit setzen */
            if( out & 0x01 )
                io_setBit(MOSI_GPIOx, MOSI_Pin);
            else
                io_clearBit(MOSI_GPIOx, MOSI_Pin);

            out>>=1;

            asm volatile("nop");
            asm volatile("nop");

            /* Clock erzeugen */
            io_setBit(SCK_GPIOx, SCK_Pin);
            asm volatile("nop");
            asm volatile("nop");
            io_clearBit(SCK_GPIOx, SCK_Pin);

            /* Noch ein wenig halten */
            asm volatile("nop");
            asm volatile("nop");
        }

        return in;
    }



    /** SPI Bitbang - Mode 3 (CPOL1, CPHA1) MSB First - Einzelnes Byte ueber SPI senden und
     * empfangen
     *
     * @param[in] SCK_GPIOx, SCK_Pin: Clock-Pin
     * @param[in] MISO_GPIOx, MISO_Pin: MISO-Pin
     * @param[in] MOSI_GPIOx, MOSI_Pin: MOSI-Pin
     * @return EXIT_SUCCESS */
    static_inline uint8_t spi_bitbang_msb_mode3(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                                GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                                GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin,
                                uint8_t out)
    {
        uint8_t in = 0;

        /* Mode 3 schreiben und lesen */
        for(uint_fast8_t i = 0; i < 8; i++)
        {
            /* Clock erzeugen */
            io_clearBit(SCK_GPIOx, SCK_Pin);
            asm volatile("nop");

            /* Daten-Bit setzen */
            io_setBitTo(MOSI_GPIOx, MOSI_Pin, out & 0x80);
            out<<=1;

            //asm volatile("nop");
            io_setBit(SCK_GPIOx, SCK_Pin);


            asm volatile("nop");
            asm volatile("nop");

            in<<=1;
            in |= io_bitIsSet(MISO_GPIOx, MISO_Pin) != 0;
        }

        return in;
    }



    exit_t spi_bitbang_init(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                        GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                        GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin,
                        enum spi_mode_e mode);

	exit_t spi_bitbang_deinit(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                     GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                     GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin);

    void spi_bb_dummyclock(GPIO_t * SCK_GPIOx, const IOPin_t SCK_Pin,
                        GPIO_t * MISO_GPIOx, const IOPin_t MISO_Pin,
                        GPIO_t * MOSI_GPIOx, const IOPin_t MOSI_Pin);

#endif
