/** @file Utest_text.c
 *
 * Nutzer-interaktive Tests: Anzeige von Zeichensaetzen, Bounding Boxes, Textfeldern
 *
 * $Date: 2020-08-12 07:47:33 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 59 $
 *
 * @date 16.03.2019 @author Nicolas */



#include "text_cursor.h"
#include "gfx_draw.h"
#include "gfx_draw_rectangles.h"
#include "delay.h"
#include "text_field.h"
#include "debughelp.h"
#include "font.h"
#include "main.h"
#include "sw_config.h"
#include "u8g2_fontdata_icons.h"
#include "u8g2_font_find_glyph.h"
#include "userinterface.h"


/** Hilfsfunktion: Testen, ob Zeichensatz alle ASCII-Grossbuchstaben enthaelt
 *
 * @param[in] font: Zeichensatz
 * @return true|false */
bool font_containsCapitalLetters(const uint8_t *font)
{
    for( int i = 'A'; i <= 'Z'; i++ )
    {
        if( !font_glyphExists(font, i) )
        {
            return false;
        }
    }

    return true;
}



/** Hilfsfunktion: Alle Zeichen eines Zeichensatzes + Textsatz mit Umlauten darstellen
 *
 * @param[in] font: Schriftart
 * @param[in] description: Text, der als Ueberschrift erscheint */
void print_allTextSymbolsBdf(const uint8_t * font, char *description)
{
    printf("Font=%s\n", description);

    Textwindowdump_t Dump = textwindow_dumpSettings();

    textwindow_setColors((Colors_t) {black, white});
    gfx_fillscreen(white);

    /* Ueberschrift noch in Normal-Schriftart ausgeben, damit sie auch auch bei
       Symbol-Schriftarten lesbar ist */
    Rect_t R = text_getWindow();
    cursor_setPosition((Coord2_t) {R.x, R.y});
    tft_putstr(description);
    tft_putstr("\n\n");


    /* Zu testende Schriftart initialisieren */
    exit_t ret =  textwindow_setFont(font);
    assert( EXIT_SUCCESS == ret );


    /* Alle Zeichen */
    uint16_t c = 0;
    while( (c = u8g2_font_findNextChar(font, c)) )
    {
        tft_putChar(c);
    }

    /* Woerter/Saetze mit Umlauten */
    if( font_containsCapitalLetters(font) )
    {
        tft_putChar('\n');
        tft_putstr("Hier gibt's keine braunen Fchse.\n");
        tft_putstr("Hliche Blten stren.\n");
    }

    textwindow_restoreSettings(Dump);

    while( !ui_getAction() );
}



/** Test aller Zeichen aller verwendeten Schriftarten
 *
 * Die Zeichen werden in schwarzer Schrift auf weissem Grund dargestellt. Jede Bildschirm-Seite
 * ist fuer einen Zeichensatz. Mit dem Taster wird die Seite weitergeschaltet. */
void Utest_allCharsOfImportantFonts(void)
{
    Textwindowdump_t Save = textwindow_dumpSettings();

    text_setWindow(gfx_getDisplayDimensions());

    #if defined( font_largestFontAvailable )
        print_allTextSymbolsBdf(font_largestFontAvailable, "Largest font available");
    #endif
    #if defined( font_breadAndButter )
        print_allTextSymbolsBdf(font_breadAndButter, "Bread an butter font");
    #endif
    #if defined( font_menuDefault )
    print_allTextSymbolsBdf(font_menuDefault, "Menu default font");
    #endif
    #if defined( font_monospaceTerminal )
    print_allTextSymbolsBdf(font_monospaceTerminal, "Terminal font");
    #endif
    #if defined( font_smallnice )
    print_allTextSymbolsBdf(font_smallnice, "Small nice font");
    #endif
    #if defined( font_largedigit )
    print_allTextSymbolsBdf(font_largedigit, "Large Digits");
    #endif
    //#if defined(   )
    print_allTextSymbolsBdf(open_iconic_selection_4x, "Icon font");
    //#endif
    //#if defined(   )
    print_allTextSymbolsBdf(open_iconic_selection_2x, "Small Icon font");
    //#endif

    //print_allTextSymbolsBdf(u8g2_font_logisoso92_tn, "u8g2_font_logisoso92_tn");

    textwindow_restoreSettings(Save);
}



/** Test: Textfelder */
void Utest_textfield(void)
{
    Textwindowdump_t Dump = textwindow_dumpSettings();

    exit_t ret =  textwindow_setFont(font_breadAndButter);
    assert( EXIT_SUCCESS == ret );


    Rect_t R = text_getWindow();
    text_setWindow(R);
    gfx_fillscreen(orange);
    drawRectF(R, yellow);

    Rect_t BB;
    const uint8_t *font = textwindow_getFont();

    //Coord2_t C;
    //C = cursor_getPosition();
    //printf("cursor={%i,%i}\n", C.x, C.y);
    char s0[] = "MM";


    BB = font_lineBbx(font, s0);


    Rect_t Rsave = R;
    R.w = R.w/3;
    R.h = R.h/3;
    R.x = R.x + R.w;
    R.y = R.y + R.h;
    BB = textfield(R, s0, align_center | align_middle);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});
    //C = cursor_getPosition();
    //printf("cursor={%i,%i}\n", C.x, C.y);

    R.x = Rsave.x + 2*R.w;
    R.y = Rsave.y + 2*R.h;
    BB = textfield(R, "ur", align_right | align_bottom);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 1*R.w;
    R.y = Rsave.y + 2*R.h;
    BB = textfield(R, "um", align_center | align_bottom);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 0*R.w;
    R.y = Rsave.y + 2*R.h;
    BB = textfield(R, "ul", align_left | align_bottom);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 0*R.w;
    R.y = Rsave.y + 1*R.h;
    BB = textfield(R, "links", align_left | align_middle);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 2*R.w;
    R.y = Rsave.y + 1*R.h;
    BB = textfield(R, "rechts", align_right | align_middle);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 0*R.w;
    R.y = Rsave.y + 0*R.h;
    BB = textfield(R, "ol", align_left | align_top);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 1*R.w;
    R.y = Rsave.y + 0*R.h;
    BB = textfield(R, "om", align_center | align_top);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    R.x = Rsave.x + 2*R.w;
    R.y = Rsave.y + 0*R.h;
    BB = textfield(R, "or", align_right | align_top);
    drawRect(BB, (Colors_t) {red, transparent});
    drawRect(R, (Colors_t) {blue, transparent});

    BB = font_fontBbx(font);
    drawRect(BB, (Colors_t) {yellow, transparent});


{


    cursor_setPosition((Coord2_t) {30, 30});
    tft_putChar(':');
    Rect_t Fbb = font_fontBbx(font);
    Fbb.x += 30;
    Fbb.y += 30;
    drawRect(Fbb, (Colors_t) {red, transparent});
    Rect_t Cbb = rect8ToRect(font_glyphOutline(font, ':').BBx);

    Cbb.x += 30;
    Cbb.y += 30;
    drawRect(Cbb, (Colors_t) {orange, transparent});
    drawPixel((Coord2_t) {30, 30}, blue);
}
{


    cursor_setPosition((Coord2_t) {40, 30});
    tft_putChar('l');
    Rect_t Fbb = font_fontBbx(font);
    Fbb.x += 40;
    Fbb.y += 30;
    drawRect(Fbb, (Colors_t) {red, transparent});
    Rect_t Cbb = rect8ToRect(font_glyphOutline(font, 'l').BBx);
    Cbb.x += 40;
    Cbb.y += 30;
    drawRect(Cbb, (Colors_t) {orange, transparent});
    drawPixel((Coord2_t) {40, 30}, blue);
}

    textwindow_restoreSettings(Dump);

    while( !ui_getAction() );
}



/** Test: Tabelle/Text-Array darstellen */
void Utest_textarray(void)
{
    ufast8_t debugstatus = debugoutput_getstatus();
    debugoutput_enable(debugoutput_boxes);
    Textwindowdump_t WindowSave = textwindow_dumpSettings();

    exit_t ret =  textwindow_setFont(font_menuDefault);
    assert( EXIT_SUCCESS == ret );


    Rect_t R = gfx_getDisplayDimensions();
    text_setWindow(R);
    gfx_fillscreen(0x3186); // dunkles grau
    R.h = 30;
    uint8_t align[4];

    {
        char outputstring[] = "Testbild Tabelle";
        uint16_t cellwidth[] = {1};
        uarray16_stretchDistance(cellwidth, 1, R.w);
        formatString2array(align, "c", 1);
        textarray(R, outputstring, align, cellwidth, 1);
    }

    {
        R.y += R.h;
        char outputstring[] = "Zwei\tSpalten";
        uint16_t cellwidth[] = {1, 1};
        uarray16_stretchDistance(cellwidth, 2, R.w);
        formatString2array(align, "cc", 2);
        textarray(R, outputstring, align, cellwidth, 2);
    }

    {
        R.y += R.h;
        uint16_t cellwidth[] = {1, 1, 1};
        uarray16_stretchDistance(cellwidth, 3, R.w);
        textarray(R, "Drei\tSpalten\tdefault", (uint8_t[]){0, 0, 0}, cellwidth, 3);
    }

    {
        R.y += R.h;
        uint16_t cellwidth[] = {2, 1, 1};
        uarray16_stretchDistance(cellwidth, 3, R.w);
        formatString2array(align, "rcl", 3);
        textarray(R, "rechts (breit)\tmittig\tlinks", align, cellwidth, 3);
    }

    {
        R.y += R.h;
        uint16_t cellwidth[] = {5, 0, 1, 1};
        uarray16_stretchDistance(cellwidth, 4, R.w);
        formatString2array(align, "lcrl", 4);
        textarray(R, "Erstens\t:\t3\t mm", align, cellwidth, 4);

        R.y += R.h;
        const uint8_t align1[] = {align_left, align_center, align_right, align_left};
        textarray(R, "Zweitens\t:\t5\t mm", align1, cellwidth, 4);

        R.y += R.h;
        snprintf(glTextbuffer, TEXTBUFFERSIZE_BYTES-1, "Drittens\t:\t\%i\t kg", 15);
        textarray(R, glTextbuffer, align, cellwidth, 4);
    }

    textwindow_restoreSettings(WindowSave);
    debugoutput_disable(debugoutput_all);
    debugoutput_enable(debugstatus);



    while( !ui_getAction() );
}



/** Wrapper */
static void drawChar(Coord2_t Pos, uint16_t unicode, Colors_t Color)
{
    const uint8_t *font = font_largestFontAvailable;
    font_drawGlyph(Pos, font, unicode, Color, bbx_mode_line);
}




/** Test: Berechnung der Font-Bounding-Box aus den Zeichen-Einzel-Bounding-Boxen
 * und Vergleich mit der hinterlegten Font-Bounding-Box */
void Utest_fontBoundingBox(void)
{
    const uint8_t *font = font_largestFontAvailable;
    exit_t ret =  textwindow_setFont(font);
    assert( EXIT_SUCCESS == ret );



    Coord2_t Pos0 = {30,30};
    Coord2_t Pos1 = {30,90};
    drawChar(Pos0, 'A', (Colors_t) {red, white});
    drawChar(Pos1, 'A', (Colors_t) {red, gray});
    for( uint16_t i = ' '; i<='~'; i++ )
    {
        color_t color = color_rainbow();
        drawChar(Pos0, i, (Colors_t) {color, transparent});
        Rect_t BBfont = font_fontBbx(font);

        /* Eventuell ist die Umrechnung falsch */
        //Rect_t BBchar = convertFontBoundingBox(u8g2glyph_getOutline(Glyph).BBx);
        Rect_t BBchar = rect8ToRect(font_glyphOutline(font, i).BBx);

        BBchar.x += Pos0.x -BBfont.x;
        BBchar.y += Pos0.y -BBfont.y;
        drawDottedRect(BBchar, (Colors_t) {color, transparent});
        //delay_ms(50);
        drawChar(Pos1, i, (Colors_t) {color, transparent});
        //delay_ms(50);
    }



    Rect_t BBfont = font_fontBbx(font);
    BBfont.x += Pos0.x;
    BBfont.y += Pos0.y;
    drawRect(BBfont, (Colors_t) {yellow, transparent});

    Rect_t BBchar = rect8ToRect(font_glyphOutline(font, '~').BBx);
    BBchar.x += Pos0.x;
    BBchar.y += Pos0.y;
    drawRect(BBchar, (Colors_t) {green, transparent});

    drawPixel(Pos0, red);

    /* Die Groesse der Bounding-Box wird ueberwiegend durch die sehr grossen Zeichen wie dem
       Klammeraffen und das "W" dominiert. */

    while( !ui_getAction() );
}
