/** @file text_field.c
 *
 * Textfelder, Tabellen
 *
 * $Date: 2020-08-01 09:01:46 +0200 (Sa, 01 Aug 2020) $
 * $Revision: 27 $
 *
 * @date 14.03.2018 @author Nicolas */

#include "text_field.h"
#include "tft_text.h"
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include "gfx_draw.h"
#include "text_cursor.h"
#include "mystring.h"
#include "debughelp.h"
#include "strfun.h"




/** Textzeile in Feld links- mittig oder rechtsbuendig darstellen
 * @param[in] R: Feldabmessungen (Rechteck)
 * @param[in] s: Zeichenkette
 * @param[in] orientation: Ausrichtung (links/rechts/zentriert)
 * @return Bounding-Box der tatsaechlich dargestellten Zeichenkette */
Rect_t textfieldLine(Rect_t R, char const* s, enum rect_orientation_e orientation )
{
    if( s == NULL )
    {
        return (Rect_t) {0, 0, 0, 0};
    }
    const uint8_t *font = textwindow_getFont();
    assert( font != NULL );

    const char *se = strNextLine(s);
    Rect_t BB = font_lineBbx(font, s);
    BB = rect_align(R, BB, orientation );
    cursor_setPosition((Coord2_t) {BB.x, BB.y});


    tft_putstrn(s, se-s);
    return BB;
}



/** Text in Feld links- mittig oder rechtsbuendig darstellen
 * @param[in] R: Feldabmessungen (Rechteck)
 * @param[in] s: Zeichenkette
 * @param[in] orientation: Ausrichtung (links/rechts/zentriert)
 * @return Bounding-Box der tatsaechlich Textflaechen */
Rect_t textfield(Rect_t R, char const* s, enum rect_orientation_e orientation )
{
    if( s == NULL )
    {
        return (Rect_t) {0, 0, 0, 0};
    }
    const uint8_t *font = textwindow_getFont();
    assert( font != NULL );

    /* Bounding-Box Textblock */
    Rect_t BBblock = font_strBbx(font, s);
    BBblock = rect_align(R, BBblock, orientation );

    /* Texblock zeilenweise darstellen */
    Rect_t BBcpy = BBblock;
    orientation |= align_top;
    orientation &= ~align_middle;
    orientation &= ~align_bottom;
    Rect_t Hl = R, Hr = R; /* Hintergrund links/rechts des Textblocks */
    Colors_t C = textwindow_getColors();
    while( *s != '\0' )
    {
        Rect_t BBr = textfieldLine(BBcpy, s, orientation);
        s = strNextLine(s);


        Hl.w = BBcpy.x - R.x;
        Hr.x = BBcpy.x + BBcpy.w;
        Hr.w = R.x + R.w - Hr.x;

        /* Hintergrund links und rechts des Textblocks mit dem Hintergrund ausmalen */
        if( Hl.w > 0 )
        {
            Hl.y = BBcpy.y;
            Hl.h = BBcpy.h;
            //tft_drawRectTransparent(Hl, green, 50);
            drawRectF(Hl, C.bg);
        }

        if( Hr.w > 0 )
        {
            Hr.y = BBcpy.y;
            Hr.h = BBcpy.h;
            //tft_drawRectTransparent(Hr, orange, 50);
            drawRectF(Hr, C.bg);
        }

        /* Naechste Zeile vorbereiten */
        BBcpy.y += BBr.h;
        BBcpy.h -= BBr.h;
    }

    return BBblock;
}



/** Feldbreiten aufteilen
 * Es wird immer die komplette Feldbreite ausgenutzt.
 * Einzelne Felder mit Nullbreite sind moeglich
 * Bei allen Feldern mit identischer Breite erfolgt automatische Gleichverteilung
 * Bei Gleichverteilung koennen Einzelfelder aufgrund von Rundungsfehlern unterschiedlich
 * gross sein.
 *
 * @param[in, out] widtharray: Feldbreiten unnormalisiert -> normalisiert
 * @param[in] n: Anzahl der Felder
 * @param[in] totalWidth: Gesamtbreite, auf die normiert wird
 * @return EXIT_SUCCESS|EXIT_FAILURE */
exit_t uarray16_stretchDistance(uint16_t *widtharray, uint16_t n, uint16_t totalWidth)
{
    ufast16_t sum = 0;
    if( widtharray == NULL || n == 0 )
    {
        return EXIT_FAILURE;
    }

    for( ufast16_t i = 0; i < n; i++ )
    {
        sum += *(widtharray+i);
    }

    if( sum == 0 )
    {
        /* Alles Null-Felder - diese bleiben so */
        return EXIT_SUCCESS;
    }

    ufast16_t rest = 0;
    for( ufast16_t i = 0; i<n; i++ )
    {
        ufast16_t divisor = (widtharray[i]*totalWidth + rest);
        widtharray[i] = divisor/sum;
        rest += divisor%sum;
    };

    return EXIT_SUCCESS;
}



/** Fotmat-String der Form "lcrrrr" in Array mit den entsprechenden enum-Werten umwandeln
 *
 * @param[out] align: Ausgabe-Format-Array
 * @param[in] formatstring: Formatstring der Form "lcr"
 * @return EXIT_SUCCESS|EXIT_FAILURE */
exit_t formatString2array(uint8_t *align, char *formatstring, ufast8_t n )
{
    if( align == NULL || formatstring == NULL )
    {
        return EXIT_FAILURE;
    }


    for( ufast16_t i = 0; i<n; i++ )
    {
        char format = *formatstring++;
        uint8_t ialign = 0;
        switch( format )
        {
        case 'l':
            ialign = align_left;
            break;

        case 'c':
            ialign = align_center;
            break;

        case 'r':
            ialign = align_right;
            break;

        default:
            error("unknown format");
        }

        *align++ = ialign;
    }

    return EXIT_SUCCESS;
}



/** Tabelle darstellen
 *
 * @param[in] R: aussere Abmessungen
 * @param[in] s: Mit Tabulatoren getrennter Textstring. Anzahl der Tabs n oder n-1
 * @param[in] align: Ausrichtungen der einzelnen Elemente
 * @param[in] celldwidth: Array mit absoluter Textbreite jeder Zelle
 * @param[in] n: Anzahl der Zellen
 * @return EXIT_SUCCESS|EXIT_FAILURE */

exit_t textarray(Rect_t R, char const* s, uint8_t const* align, uint16_t const* cellwidth, uint16_t n)
{
    if( s == NULL || align == NULL || cellwidth == NULL || n == 0 )
    {
        return EXIT_FAILURE;
    }


    /* Zeichenkette kopieren, damit strtok/strsep darin herumwerken kann */
    char *cp = strdup(s);
    char *cpo = cp;
    assert( cp != NULL );
    char *token;
    char delimiters[] = "\n\t";



    token = strsep(&cp, delimiters);
    for( ufast16_t i = 0; i<n; i++ )
    {
        R.w = cellwidth[i];
        debugoutput_showBox(R, orange);
        uint8_t halign = *align++;
        uint8_t valign = align_middle;
        Rect_t Tmp = textfield(R, token, halign | valign);
        R.x += R.w;

        debugoutput_showBox(Tmp, red);

        token = strsep(&cp, delimiters);
        if( token == NULL )
        {
            break;
        }
    }

    free(cpo);
    cpo = NULL;
    return EXIT_SUCCESS;
}
