/** @file tft_text.c
 *
 * Text-Funktionen fuer TFT-GLCD
 *
 * Die Text-Funktione fuer ein TFT-LCD (z.B. ILI9341) sind etwas aufwendiger als fuer die
 * reinen Zeilen-LCDs (z.B. KS0108), da Fonts mit 8 Pixeln Hoehe wenig Sinn ergeben und
 * der Cursor manuell implementiert werden muss.
 *
 * $Date: 2020-08-04 18:20:49 +0200 (Di, 04 Aug 2020) $
 * $Revision: 41 $
 *
 * @date 29.09.2017 @author Nicolas */

#include "tft_text.h"
#include "gfx_draw.h"
#include "gfx_draw_lines.h"
#include "gfx_draw_rectangles.h"
#include "gfx_draw_circles.h"
#include <stddef.h>
#include <stdbool.h>
#include <string.h>
#include "strfun.h"
#include "text_window.h"
#include "bitmath.h"
#include "font.h"

//#include "debughelp.h"
#include "text_cursor.h"




/** Anordnung Zeichen im String, kann auch verODERt werden */
enum txt_char_align
{
    txt_char_middle = 0x0, /**< mittleres Zeichen im String */
    txt_char_first  = 0x1, /**< erstes Zeichen im String */
    txt_char_last   = 0x2, /**< letztes Zeichen im String */
};


/* Funktionsprototypen */
bool str_executeEscape(char c);



/** Einzelnes Zeichen auf dem Display darstellen (mit automatischem Zeichen- und Zeilenvorschub)
 * @param[in] unicode: Zeichen
 * @return EXIT_SUCCESS | EXIT_FAILURE */
exit_t tft_putCharAligned(uint16_t unicode, enum txt_char_align align)
{
    /* Verbleibender Zeichenzwischenraum vom vorherigen Zeichen */
    static int lastSpace = 0;


    /* Pruefen, ob Escape-Sequenz vorliegt. Wenn ja, ausfuehren und beenden */
    if( str_executeEscape(unicode) )
    {
        return EXIT_SUCCESS;
    }


    ifast16_t width; /* Zeichenbreite (fuer Umbruch) */
    ifast16_t pitch; /* Zeichenvorschub nach Zeichen */


    const uint8_t *font = textwindow_getFont();
    Colors_t Color = textwindow_getColors();
    Rect_t Fbb = font_fontBbx(font);
    bool glyphAvailable = font_glyphExists(font, unicode);

    GlyphOutline_t Outline;
    Outline = glyphAvailable ? font_glyphOutline(font, unicode) : font_ersatzOutline(font);


    width = (ifast16_t) Outline.BBx.w + Outline.BBx.x;

    Coord2_t Pos0 = *(cursor_getPtr());
    Coord2_t Pos = cursor_linebreakBeforeChar(width);

    bool linebreak = Pos.y != Pos0.y;
    if( linebreak || (align & txt_char_first) )
    {
        lastSpace = 0;
        if( Outline.BBx.x < 0)
        {
            //Pos.x -= Outline.BBx.x;
            Outline.BBx.x = 0;
        }
    }

    Pos.x += Outline.BBx.x;

    if( glyphAvailable )
    {
        /* Zeichen darstellen */
        font_drawGlyph(Pos, font, unicode, Color, bbx_mode_line);
    }
    else
    {
        /* Ersatz-Rechteck darstellen */
        Rect_t R = {.x = Pos.x, .y = Pos.y, .w = Outline.BBx.w, .h = Outline.BBx.h};
        drawRect(R, Color);
    }


    /* Leeres Rechteck vor dem Zeichen darstellen */
    int we = lastSpace + Outline.BBx.x;
    sclamp(&we, 0, INT_MAX);
    Rect_t R = {.x = Pos.x - we, .y = Pos.y, .w = we, .h = Fbb.h};
    drawRectF(R, Color.bg);


    /* Schmalen Leerraum hinter dem letzten Zeichen darstellen */
    if( align & txt_char_last )
    {
        R.w = Outline.pitch - Outline.BBx.x - Outline.BBx.w;
        R.x = Pos.x + Outline.BBx.w;
        if( R.w > 0 )
        {
            drawRectF(R, Color.bg);
        }
    }


    #if FONT_TEST_GLYPH_REF_POINT
        drawCircle(Pos, 2, orange);
    #endif

    pitch = Outline.pitch;
    lastSpace = pitch - Outline.BBx.x - Outline.BBx.w;


    cursor_advanceAfterChar(pitch);
    return EXIT_SUCCESS;
}



/** Einzelnes Zeichen darstellen
 * Sollte nach Moeglichkeit eher selten verwendet werden, weil die Ausrichtungen nicht besonders
 * gut passen
 * @param[in] unicode: Zeichen
 * @return EXIT_SUCCESS | EXIT_FAILURE */
exit_t tft_putChar(uint16_t unicode)
{
    return tft_putCharAligned(unicode, txt_char_first | txt_char_last);
}



/** Null-terminierten String auf LCD ausgeben
 * @param[in] s: Zeichenkette (interpretiert als 8-Bit-Ascii)
 * @return Anzahl der ausgegebenen Zeichen */
ufast16_t tft_putstr(const char *s)
{
    size_t n = strlen(s);
    return tft_putstrn(s, n);
}



/** String mit definierter Laenge auf LCD ausgeben
 * @param[in] s: Zeichenkette (interpretiert als 8-Bit-Ascii)
 * @param[in] n: maximale Laenge
 * @return Anzahl der ausgegebenen Zeichen */
ufast16_t tft_putstrn(const char *s, size_t n)
{
    for( size_t i = 0; i<n; i++ )
    {
        enum txt_char_align align = 0;
        if( i == 0 )
        {
            align |= txt_char_first;
        }
        if( i == n-1 )
        {
            align |= txt_char_last;
        }
        tft_putCharAligned( (uint8_t) *s++, align );
    }
    return n;
}



/** Auf Escape-Sequenzen pruefen; wenn Escape-Sequenz, wird
 *  1. diese ausgefuehrt und
 *  2. true zurueckgeliefert, damit die Hauptfunktion beendet werden kann.
 * @param[in] c: Einzelnes Zeichen
 * @return true|false: Zeichen als Teil einer Escape-Sequenz erkannt */
bool str_executeEscape(char c)
{
    static uint8_t iscommand = 0; // printable character oder Kommando
    bool retval = false;

    // Falls vorangegangenes Zeichen "Escape": Kommando interpretieren
    if (iscommand)
    {
        // Escape-Zeichen interpretieren
        switch (c)
        {
            case 'f': // feste Schriftweite
                //Textcfg.isFixedWidth = 1;
                break;
            case 'v': // variable Schriftweite
                //Textcfg.isFixedWidth = 0;
                break;
            case 'i': // Invertieren
                //Textcfg.background = ~Textcfg.background;
                //text_invert();
                break;
            case 's': // tiefgestellt, kleiner Zeichensatz
                //Textcfg.pfont = Font36Lookup;
                //Textcfg.downshift = 1;
                break;
            case 'n': // normaler Zeichensatz
                //text_normal();
                //Textcfg.pfont = Font58Lookup;
                //Textcfg.downshift = 0;
                //Textcfg.background = 0x00U;
                //Textcfg.isFixedWidth = 1;
                break;
            case 'b': // fetter Zeichensatz
                //Textcfg.pfont = Font8x8Lookup;
                //Textcfg.downshift = 0;
                break;
            case 'd': // extrem gequetschter Zeichensatz
                //Textcfg.pfont = Font48Lookup;
                //Textcfg.downshift = 0;
                break;
        }
        iscommand = 0;
        return true;
    }

    // Nach Escape-Zeichen suchen:
    switch(c)
    {
        case '\033': // = '\e'
            // Escape: Das folgende Zeichen hat Kommandofunktion
            iscommand = 1;
            retval = true;
            break;
        case '\b': // Backspace
            //backspace();
            retval = true;
            break;
        case '\t': // Tabulator
            //tab();
            retval = true;
            break;
        case '\n': // newline
            text_newline();
            retval = true;
            break;
        case '\f': // Formfeed
            //glcd_clear();
            retval = true;
            break;
        case '\a': // Alert
            //glcd_drawnow();
            retval = true;
            break;
    }

    return retval;
}



/** Bounding-Box eines Text-Strings bis zum Zeilenumbruch, Tabstopp oder Stringende
 * @param[in] s: Zeichenkette
 * @return Umhuellendes Rechteck einer Zeile */
Rect_t font_lineBbx(const uint8_t *font, char const* s)
{
    assert( font != NULL );
    assert( s != NULL );


    Rect_t R = font_fontBbx(font);
    R.w = 0;
    char c;


    /* Fuer das erste Zeichen muss noch der Font-X-Offset hinzugerechnet werden */
    if( *s != '\0' )
    {
        R.w += MAX(-R.x, 0);
    }


    while( true )
    {
        c = *s++;

        switch( c )
        {
            case '\0': /* Fallthru */
            case '\n': /* Fallthru */
            case '\t':
                return R;

            default:
                if( font_glyphExists(font, (uint8_t) c) )
                {
                    R.w += font_glyphOutline(font, (uint8_t) c).pitch;
                }
                else
                {
                    R.w += font_ersatzOutline(font).pitch;
                }
        }
    }
}



/** Bounding-Box einer mehrzeiligen Zeichenkette (manuelle Zeilenumbrueche)
 * @param[in] s: Zeichenkette
 * @return Umhuellendes Rechteck */
Rect_t font_strBbx(const uint8_t *font, char const* s)
{
    assert( s != NULL );
    Rect_t R = font_lineBbx(font, s); // Erste Zeile

    while( *(s = strNextLine(s)) != '\0' )
    {
        Rect_t Ri = font_lineBbx(font, s);
        R.h += Ri.h;
        R.w = MAX(R.w, Ri.w);
    }

    return R;
}
