/** @file Utest_u8g2.c
 *
 * Nutzer-interaktive Tests: U8G2-Zeichensatzfunktionen
 *
 * $Date: 2020-08-12 18:19:08 +0200 (Mi, 12 Aug 2020) $
 * $Revision: 67 $
 *
 * @date 22.05.2018  @author Nicolas */


#include "Utest_all.h"
#include "text_cursor.h"
#include "gfx_draw_rectangles.h"
#include "gfx_draw_circles.h"
#include "gfx_draw_diagrams.h"
#include "delay.h"
#include "text_field.h"
#include "debughelp.h"
#include "sw_config.h"
#include "u8g2_fontdata_icons.h"
#include "font.h"
#include "u8g2_font_decoder.h"
#include "u8g2_font_header_private.h"
#include "u8g2_font_find_glyph.h"
#include "userinterface.h"
#include <inttypes.h>


/** Hilfsfunktion: Zeichen mit Bounding-Box auf dem TFT darstellen und evtl. warten
 *
 * @param[in] font: Zeichensatz-Tabelle
 * @param[in] c: Zeichen
 * @param[in] dowait: Warten ja|nein */
void putcharWithBoundingBoxandWait(const uint8_t *font, uint16_t c, bool dowait)
{
    Coord2_t L = cursor_getPosition();
    Rect_t Fbb = font_fontBbx(font);

    tft_putChar(c);
    Rect_t Gbb    = rect8ToRect(font_glyphOutline(font, c).BBx);

    #if MOCKUP_SDL
        printf("GlyphBB (%c) = ", c);
        rect_print(Gbb);
    #endif

    drawCircle(L, 3, green);
    Coord2_t Bp = {.x = L.x + Gbb.x, .y = L.y + Gbb.y };
    drawCircle(Bp, 2, yellow);

    drawBoundingBox(L, Fbb, (Colors_t) {color_rainbow(), white} );
    Coord2_t Lg = coord2_sub(coord2_add(L, rectToPosition(Gbb)), rectToPosition(Fbb));
    drawBoundingBox(Lg, Gbb, (Colors_t) {red, orange} );

    while( !ui_getAction() && dowait );
}



/** Zeichenkette auf dem TFT mit umgebenden Bounding-Boxen darstellen
 *
 * Das erlaubt eine Beurteilung, ob Zeichenvorschub, Zeilenumbruch und Offsets richtig
 * beruecksichtigt werden. */
void Utest_u8g2_boundingBox(void)
{
    Textwindowdump_t Dump = textwindow_dumpSettings();


    const uint8_t *font = font_largestFontAvailable;
    font = u8g2_font_logisoso50_tr;

    textwindow_setColors((Colors_t) {green, transparent});
    textwindow_setFont(font);
    gfx_fillscreen(0x3186); /* dunkles grau */

    /* Test meiner eignenen Bounding box */
    Rect8_t MyFbb = {0, 0, 0, 0};
    uint16_t code = u8g2_font_findNextChar(font, 0);
    MyFbb = rect8_outline(MyFbb, font_glyphOutline(font, code).BBx);

    while( code )
    {
        code = u8g2_font_findNextChar(font, code);
        Rect8_t ThisGlyphBB = font_glyphOutline(font, code).BBx;
        MyFbb = rect8_outline(MyFbb, ThisGlyphBB);
        //printf("This Glyph (%u) Bounding Box = ", code);
        //rect_print(rect82rect(MyFbb));
    }

    //printf("My Font Bounding Box = ");
    //rect_print(rect82rect(MyFbb));




    drawRectTransparent(Dump.Window, yellow, 50);

    Coord2_t Pos = {0, 00};

    cursor_setPosition(Pos);



    bool dowait = 0;

    char c;
    char *s0 = "Font bounding box test\n";

    while( (c = *s0++) )
    {
        putcharWithBoundingBoxandWait(font, (uint8_t) c, dowait);
    }

    #if 0
    char *s1 = "Agrarkonom([|@01\n";

    while( (c = *s1++) )
    {
        putcharWithBoundingBoxandWait( (uint8_t) c, dowait);
    }
    #endif


    textwindow_restoreSettings(Dump);
    while( !ui_getAction() );

}



/** Test: Header-Daten eines Fonts komplett dekodieren und in der Konsole darstellen */
void Utest_u8g2_decoder(void)
{

    const uint8_t *font = font_largestFontAvailable;
    font =  font_largedigit;
    font = font_monospaceTerminal;

    font = u8g2_font_logisoso50_tr;
    const char *fontstr = "u8g2_font_logisoso50_tr";

    //Glyph_t GlyphData = u8g2glyph_init(font);
    //Glyph_t *Glyph = &GlyphData;


    /* Font Header */
    printf("\n\n");
    printf("--------------------\n");
    printf("%s\n", fontstr);
    printf("Font=%"PRIxPTR", n=%i \n", (uintptr_t) font, u8g2_font_getNumberOfGlyphs(font));
    printf("BBX mode = %i\n", *(font+u8g2_font_header_offset_bounding_box_mode));
    printf("RLE 0|1 = %i|%i \n", u8g2_font_getNumberOfBitsRleZeros(font), u8g2_font_getNumberOfBitsRleOnes(font));
    printf("Glyph Bit width w:%i, h:%i, x:%i, y:%i, dx:%i\n",
           u8g2_font_getNumberOfBitsCharWidth(font),
           u8g2_font_getNumberOfBitsCharHeight(font),
           u8g2_font_getNumberOfBitsCharX(font),
           u8g2_font_getNumberOfBitsCharY(font),
           u8g2_font_getNumberOfBitsDeltaX(font));

    Rect8_t Fbb = u8g2_font_getBoundingBox(font);
    printf("FontBbx={%2i,%2i,%2i,%2i}|\n", Fbb.x, Fbb.y, Fbb.w, Fbb.h);

    printf("'A'-ascent:%i, 'g'-descent:%i, '('-ascent:%i, '('-descent:%i \n",
           u8g2_font_getCapitalAAscent(font),
           (int8_t) *(font+u8g2_font_header_offset_descent_g),
           (int8_t) *(font+u8g2_font_header_offset_ascent_parenthesis),
           (int8_t) *(font+u8g2_font_header_offset_descent_parenthesis)
           );

    printf("Glyph-Offset: 'A'=%i, 'a'=%i, '255'=%i\n",
            u8g2_font_getStartPosUpperA(font),
            u8g2_font_getStartPosLowerA(font),
            u8g2_font_getStartPosUnicode(font)
           );


    /* Glyph-Header */
    printf("--------------------\n");
    uint16_t code = 0;
    while( (code = u8g2_font_findNextChar(font, code)) )
    {

        //u8g2glyph_setCharacter(Glyph, code);

        Rect8_t Gbb = font_glyphOutline(font, code).BBx;
        uint8_t pitch = font_glyphOutline(font, code).pitch;
        printf("code=%4i | '%c' | Bbx={%2i,%2i,%2i,%2i}, %2i |\n",
               code, code, Gbb.x, Gbb.y, Gbb.w, Gbb.h, pitch);

    }

    printf("--------------------\n");
    while( !ui_getAction() );
}


