/** @file font.c
 *
 * Allgemeine Funktionen zu Zeichensatz und Zeichendarstellung
 *
 * $Date: 2020-08-02 19:08:25 +0200 (So, 02 Aug 2020) $
 * $Revision: 32 $
 *
 * @date 08.03.2018 @author Nicolas */

#include "font.h"
#include "u8g2_font_header_private.h"
#include "u8g2_font_decoder.h"
#include "bitmath.h"
#include <string.h>
#include "gfx_draw.h"
#include "gfx_draw_rectangles.h"
#include "main.h"
#include "text_window.h"
#include "strfun.h"



/* Funktionsdeklarationen */
static exit_t u8g2_glyph_drawGlyph(const uint8_t *font, const uint8_t *glyph, uint16_t unicode, Coord2_t Pos, Colors_t Color);



/** Hilfsfunktion: Font-Bounding-Box aus U8G2-Font-Format (Koordinatensystem y Aufwaerts von Basislinie)
 * in passende Bounding Box (y abwaerts von oben links) umrechnen
 * @param[in] BB8: Bounding-Box wie in Glyph-Daten ausgelesen (8 Bit)
 * @return Neue Bounding Box */
static inline Rect_t convertFontBoundingBox(Rect8_t BB8)
{
    Rect_t BB = rect8ToRect(BB8);
    BB.y = -BB.h - BB.y + 1; /*(BB.y < 0) */
    return BB;
}



/** Get Outline Data (Bounding Box + character pitch)
 * @param[in] font: Zeichensatz-Tabelle
 * @param[in] unicode: Zeichen
 * @return Outline-struct
 * @TODO Effizienter programmieren fuer mehrere Aufrufe mit dem gleichen Zeichen */
GlyphOutline_t font_glyphOutline(const uint8_t *font, uint16_t unicode)
{
    assert( font != NULL );
    // @TODO: Das ist eine teure Funktion
    const uint8_t *glyph = font_glyph(font, unicode);
    if( glyph == NULL )
    {
        /* Invalid Character */
        return (GlyphOutline_t) {};
    }
    else
    {
        return u8g2glyph_decodeHeader(font, glyph);
    }
}



/** Ersatz-Get Outline Data (Bounding Box + character pitch) fuer nicht-vorandene Zeichen
 * @param[in] Glyph: Zeichen-Objekt
 * @return Outline-struct */
GlyphOutline_t font_ersatzOutline(const uint8_t *font)
{
    assert( font != NULL );
    Rect_t Fbb = font_fontBbx(font);
    GlyphOutline_t Out = {
        .BBx = {
            .x = 0,
            .y = Fbb.y,
            .h = Fbb.h,
            .w = Fbb.w * 3/4
        },
        .pitch = Fbb.w * 3/4,
    };

    return Out;
}



/** Wrapper-Hilfsfunktion: Zeichen im Zeichensatz suchen
 *
 * @param[in] font: Font-Array
 * @param[in] unicode: 16-Bit-Unicode des Glyphen
 * @return Adresse des Glyphen oder Nullzeiger */
/** Get Glyph Pointer
 * @param[in] font: Zeichen-Tabelle
 * @param[in] unicode: Zeichen
 * @return Outline-struct
 * @TODO Nach Fertigstellung kommt diese Funktion in font.c
 * @TODO Effizienter programmieren fuer mehrere Aufrufe mit dem gleichen Zeichen */
const uint8_t * font_glyph(const uint8_t *font, uint16_t unicode)
{
    static const uint8_t *lastFont = NULL;
    static uint16_t lastUnicode = INT16_MAX; /* Definitiv ungueltig */
    static const uint8_t *lastGlyph = NULL;
    static int64_t match = 0;
    static int64_t new = 0;

    //printf("match:%i, new:%i\n", match, new);

    if( font == lastFont && unicode == lastUnicode )
    {
        match++;
        return lastGlyph;
    }
    else
    {
        lastFont = font;
        lastUnicode = unicode;
        lastGlyph = u8g2_font_findCharacterGlyphPtr(font, unicode);
        new++;
        return lastGlyph;
    }
}



/** Pruefen, ob Zeichen im Zeichensatz existiert
 *
 * @param[in] font: Schrift-Tabelle U8G2
 * @param[in] unicode: Zeichen
 * @return Outline-struct */
bool font_glyphExists(const uint8_t *font, uint16_t unicode)
{
    assert( font != NULL );
    const uint8_t *glyph = font_glyph(font, unicode);
    return (glyph != NULL );
}



/** Zeichen darstellen
 * (ohne Initialisierung, direkt an der Position)
 * @param[in] Pos: X-Y-Koordinaten
 * @param[in] font: Zeichensatz-Tabelle
 * @param[in] unicode: Zeichen
 * @param[in] Color: Farbe (Vordergrund/Hintergrund)
 * @param[in] bbxMode: Ausrichtung BoundingBox (@see glyph_bbx_mode_e) */
void font_drawGlyph(Coord2_t Pos, const uint8_t *font, uint16_t unicode, Colors_t Color,
                         enum glyph_bbx_mode_e bbxMode)
{
    assert( font != NULL );

    Rect8_t FontBB = u8g2_font_getBoundingBox(font);
    const uint8_t *glyph = font_glyph(font, unicode);
    assert( glyph != NULL ); /* Darf nur mit gueltigem Zeichen aufgerufen werden */
    GlyphOutline_t Outline = u8g2glyph_decodeHeader(font, glyph);

    switch( bbxMode )
    {
        case bbx_mode_block:
            Pos.x += Outline.BBx.x;
            break;

        case bbx_mode_tight:
            break;

        case bbx_mode_line:
            break;
    }


    u8g2_glyph_drawGlyph(font, glyph, unicode, Pos, Color);
}



/** Zeichen darstellen (Glyph-Fonts)
 *
 * Zeichen im Glyph-Objekt an einer waehlbaren Position darstellen
 *
 * Die Position beschreibt die obere linke Ecke der Bounding-Box. Diese hat Zeilenhoehe und Zeichen-
 * breite (ohne Zwischenraum), d.h. entspricht den klassischen Bleilettern.
 *
 * Bei transparentem Hintergrund werden die Zeichen Pixelweise geschrieben. Da ein einzelnes Pixel
 * zu schreiben auf einem ILI9341 eine Schreiblast von 13 Bytes erzeugt, ist dies nicht sehr
 * effizient, da uebliche Schriftarten einen Bedeckungsgrad von ca. 28%...40% haben:
 *  Komplett setzen (minimaler Overhead:     100% * 2 Bytes + 1 Byte + 1x10 Bytes
 *  nur aktive Pixel setzen (viel Overhead):   40% * 3 Bytes + 40% * 10 Bytes
 *
 * Bei normaler Hintergrundfarbe werden die Zeichen auf einem Rechteck-Hintergrund (Bounding Box)
 * dargestellt, dessen Groesse so ist, dass alle Zeichen darin Platz finden.
 *
 * @param[in] Glyph: Zeichen-Objekt
 * @param[in] Pos: Koordinate (x/y)
 * @param[in] Color: Farbe (Vordergrund/Hintergrund)
 * @return Breite des ausgegebenen Zeichens
 *
 * @TODO: Merkwuerdig: Funktion hat Glyph und Unicode als Parameter. */
exit_t u8g2_glyph_drawGlyph(const uint8_t *font, const uint8_t *glyph, uint16_t unicode, Coord2_t Pos, Colors_t Color)
{
    assert( font != NULL );
    assert( glyph != NULL );

    /* BoundingBox (Breite und Hoehe) */
    Rect_t FontBB = convertFontBoundingBox(u8g2_font_getBoundingBox(font));
    //printf(" Pos={%i,%i}\n", Pos.x, Pos.y);

    //Rect_t FontBB = (*font_getBoundingBox)();   // Breite und Hoehe
    FontBB.x = Pos.x;  // Internen x-y-Offset durch tatsaechliche Position ersetzen
    FontBB.y = Pos.y;


    /* Im Pixelbuffer liegen die Bitmaps unskaliert */
    assert( PIXELBUFFERSIZE_BYTES >= bytes4bits(FontBB.w * FontBB.h ) );
    //exit_t ret = u8g2glyph_fetchBitmap(font, glyph, glPixelbuffer);
    u8g2glyph_decodeBitmap(font, glyph, glPixelbuffer);


    /* Bitmap auf TFT/LCD darstellen (Skalierungsfaktor 1) */
    //printf("Rect=[%i,%i,%i,%i]\n", FontBB.x, FontBB.y, FontBB.w, FontBB.h);
    Rect8_t GlyphBB = font_glyphOutline(font, unicode).BBx;
    FontBB.w = GlyphBB.w;

    const ufast8_t scale = 1;
    drawBitmap(glPixelbuffer, FontBB, Color, scale);

    return FontBB.w;
}



/** Bounding-Box des Zeichensatzes
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Bounding box (normales Rechteck-Format) */
Rect_t font_fontBbx(const uint8_t *font)
{
    assert( font != NULL );
    return convertFontBoundingBox(u8g2_font_getBoundingBox(font));
}







/** Einzelnes Zeichen beliebig ausgerichtet in Rechteck darstellen
 *
 * Eignet sich insbesondere fuer die Darstellung von Piktogrammen (Schriftarten
 * open_iconic_selection_2x oder open_iconic_selection_4x)
 *
 * vgl. auch @see textfield()
 *
 * Passt das Zeichen nicht in das Rechteck, wird ueber den Rand gemalt. Wird ein Zeichen nicht
 * gefunden, wird an die Stelle ein leeres Rechteck gemalt.
 *
 * @param[in] Pos: Koordinate (x/y)
 * @param[in] font: Font
 * @param[in] c: Unicode
 * @param[in] Color: Farbe (Vordergrund/Hintergrund)
 * return Tatsaechliche Position des dargestellten Zeichens */
Rect_t font_drawGlyphAligned(Rect_t R, const uint8_t *font, uint16_t unicode, Colors_t Color,
                         enum rect_orientation_e orientation)
{
    if( font == NULL )
    {
        return (Rect_t) {0, 0, 0, 0};
    }


    bool isGlyph = font_glyphExists(font, unicode);
    GlyphOutline_t Outline;

    if( isGlyph )
    {
        Outline = font_glyphOutline(font, unicode);
    }
    else
    {
        Outline = font_ersatzOutline(font);
    }

    Rect_t GlyphRect = {.x = 0, .y= 0, .w = Outline.BBx.w, .h = Outline.BBx.h };
    Rect_t Icon = rect_align(R, GlyphRect, orientation);

    if( isGlyph )
    {
        Coord2_t Pos = {.x = Icon.x, .y = Icon.y};
        font_drawGlyph(Pos, font, unicode, Color, bbx_mode_block);
    }
    else
    {
        /* Ersatz-Rechteck */
        drawRect(Icon, Color);
    }
    return Icon;
};

