/** @file u8g2_font_decoder.c

  This Code is based on

  U8g2 for Adafruit GFX Lib (https://github.com/olikraus/U8g2_for_Adafruit_GFX)

  Copyright (c) 2018, olikraus@gmail.com
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list
    of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright notice, this
    list of conditions and the following disclaimer in the documentation and/or other
    materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.




  $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
  $Revision: 12 $

  @date 25.03.2018 @author Nicolas */


#include "u8g2_font_decoder.h"
#include "u8g2_fontdata.h"
#include "u8g2_font_header_private.h"
#include "u8g2_font_find_glyph.h"
#include "font.h"
#include "bitmath.h"
#include <string.h>



/** Font decoder state variables (input state) */
typedef struct FontDecodeState_s
{
    const uint8_t *ptr;         /** pointer to the compressed data */
    uint8_t bitpos;             /** bitpos inside a byte of the compressed data */
}
FontDecodeState_t;



/* Init decode marker by bit offset */
FontDecodeState_t u8g2_font_decode_init(const uint8_t *ptr, ufast16_t offsetBits)
{
    return (FontDecodeState_t)
    {
        .ptr  = ptr + (offsetBits/8),
        .bitpos = offsetBits%8,
    };
}



/* Get byte out of bits of arbitrary offset and width (not bound to byte boundaries) */
uint8_t u8g2_font_decode_get_unsigned_bits(FontDecodeState_t *f, uint8_t cnt)
{
    uint8_t val;
    uint8_t bit_pos = f->bitpos;
    uint8_t bit_pos_plus_cnt;

    //val = *(f->decode_ptr);
    val = u8x8_pgm_read( f->ptr );

    val >>= bit_pos;
    bit_pos_plus_cnt = bit_pos;
    bit_pos_plus_cnt += cnt;
    if( bit_pos_plus_cnt >= 8 )
    {
        uint8_t s = 8;
        s -= bit_pos;
        f->ptr++;
        //val |= *(f->decode_ptr) << (8-bit_pos);
        val |= u8x8_pgm_read( f->ptr ) << (s);
        //bit_pos -= 8;
        bit_pos_plus_cnt -= 8;
    }
    val &= (1U<<cnt)-1;
    //bit_pos += cnt;

    f->bitpos = bit_pos_plus_cnt;
    return val;
}



/*
    2 bit --> cnt = 2
      -2,-1,0. 1

    3 bit --> cnt = 3
      -2,-1,0. 1
      -4,-3,-2,-1,0,1,2,3

      if ( x < 0 )
  r = bits(x-1)+1;
    else
  r = bits(x)+1;

*/
/* optimized */
int8_t u8g2_font_decode_get_signed_bits(FontDecodeState_t *f, uint8_t cnt)
{
    int8_t v, d;
    v = (int8_t)u8g2_font_decode_get_unsigned_bits(f, cnt);
    d = 1;
    cnt--;
    d <<= cnt;
    v -= d;
    return v;
    //return (int8_t)u8g2_font_decode_get_unsigned_bits(f, cnt) - ((1<<cnt)>>1);
}



/* Decode preamble/glyph header
 * Glyph header data is stored independent of byte boundaries, so elements must be decoded
 * one by one in fixed sequence. */
GlyphOutline_t u8g2glyph_decodeHeader(const uint8_t *font, const uint8_t *glyph)
{
    assert( font != NULL );
    assert( glyph != NULL );

    GlyphOutline_t Go;
    FontDecodeState_t Decoder = u8g2_font_decode_init(glyph, 0);

    /* Font Bounding Box */
    Go.BBx.w = u8g2_font_decode_get_unsigned_bits(&Decoder, u8g2_font_getNumberOfBitsCharWidth(font));
    Go.BBx.h = u8g2_font_decode_get_unsigned_bits(&Decoder, u8g2_font_getNumberOfBitsCharHeight(font));
    Go.BBx.x = u8g2_font_decode_get_signed_bits(&Decoder, u8g2_font_getNumberOfBitsCharX(font));
    Go.BBx.y = u8g2_font_decode_get_signed_bits(&Decoder, u8g2_font_getNumberOfBitsCharY(font));

    /* character pitch */
    Go.pitch = u8g2_font_decode_get_signed_bits(&Decoder, u8g2_font_getNumberOfBitsDeltaX(font));

    return Go;
}



/** Font decoder: Bitmap data (Pixels, width and height) */
typedef struct Bitmap_s
{
    uint8_t *data;  /**< Pixels */
    uint16_t w, h;  /**< width and height */
}
Bitmap_t;



/*
  Description:
    Set single segment of n pixels of the glyph in pixel buffer. "nPixel" can have any
    size and the line length will be wrapped at the glyph border.
    Function needs to be called for both zero- and one-bits for the update of local
    co-ordinates. Bitmap will be clipped at boundaries
  Args:
    Bitmap:             Destination
    Pos:                Local co-ordinates of pixel inside Font Bounding Box
    GlyphBB:            Glyph Bounding Box
    nPixel:             Length of the line
    bitval:             Bit value 0/1, set/not set bit
  Return: -

  Implementation is quick & dirty and could be improved by setting byte segments, but it's
  fast enough for the current application  */
static void u8g2_bitmap_set_line_element(Bitmap_t *Bitmap, Coord2_t *Pos, Rect8_t GlyphBB, uint8_t nPixel, bool bitval)
{
    ufast8_t rem;         /* remaining pixel to the right border of the glyph */
    ufast8_t segmentlen;  /* number of pixels to be set in one line */


    /* local coordinates of the glyph (inside character bounding box) */
    ufast8_t lx = Pos->x;
    ufast8_t ly = Pos->y;

    for(;;)
    {
        /* calculate the number of pixel to the right edge of the glyph */
        rem = GlyphBB.w - lx;

        /* calculate how many pixel to draw. This is either to the right edge */
        /* or less, if not enough pixels are left */
        segmentlen = MIN(rem, nPixel);


        /* Target position inside Font Bounding Box */
        ifast8_t x = lx + GlyphBB.x;
        ifast8_t y = ly + GlyphBB.y;

        /* Set segment of bits if active segment (will be called multiple times for
           segments spanning multiple rows), clip y co-ordinate  */
        if( bitval == 1 && y >= 0 && y < Bitmap->h )
        {
            while( segmentlen-- )
            {
                if( 0<= x && x < Bitmap->w ) // clip x co-ordinate
                {
                    // Transposed
                    arrayWriteBitIJ(Bitmap->data, y, x, 1, Bitmap->h, Bitmap->w);
                }
                x++;
            }
        }


        /* check, whether the end of the run length code has been reached */
        if( nPixel < rem )
            break;

        nPixel -= rem;
        lx = 0;
        ly++;
    }
    lx += nPixel;

    Pos->x = lx;
    Pos->y = ly;
}



/** Decode glyph and unpack run-length encoded bitmap to pixel buffer
 *
 * Die Bitmap hat die Form wie klassische Bleilettern, d.h. ueber die ganze Hoehe ohne
 * Beruecksichtigung horizontaler Zwischenraeume.
 *
 * @param[in] font: Font-Tabelle
 * @param[in] unicode: Schriftzeichen
 * @param[out] pixelbuffer: output bitmap pixel buffer */
void u8g2glyph_decodeBitmap(const uint8_t *font, const uint8_t *glyph, uint8_t *pixelbuffer)
{
    assert( font != NULL );
    assert( glyph != NULL );
    assert( pixelbuffer != NULL );


    /* Clear pixel buffer */
    Rect8_t FontBB = u8g2_font_getBoundingBox(font);
    size_t nByte = bytes4bits( (ufast16_t) FontBB.w * FontBB.h);
    memset(pixelbuffer, 0, nByte);


    /* Init decoder with correct offset for glyph bitmap data. This makes decoding glyph
     * bitmap data is stateless */
    ufast16_t offsetBits = u8g2_font_getNumberOfBitsGlyphHeader(font);
    FontDecodeState_t Decoder = u8g2_font_decode_init(glyph, offsetBits);


    /* Character Bounding Box + Offset */
    /* Co-Ordinates of Bitmap and U8G2 Fonts differ:
     * U8G2: y points upwards from baseline
     * Bitmap: y points downwards from upper left corner */


    // @TODO Pruefen: Wird hier nicht der Header unnoetig ein zweites Mal dekodiert?
    Rect8_t GlyphBB = u8g2glyph_decodeHeader(font, glyph).BBx;


    //GlyphBB.x = (ifast16_t) -FontBB.x + GlyphBB.x;
    GlyphBB.x = 0;
    GlyphBB.y = (ifast16_t) FontBB.h + FontBB.y - GlyphBB.h - GlyphBB.y;

    /* Prepare Pixelbuffer */
    //Bitmap_t Bitmap = {.data = pixelbuffer, .w = FontBB.w, .h = FontBB.h};
    Bitmap_t Bitmap = {.data = pixelbuffer, .w = GlyphBB.w, .h = FontBB.h};

    /* Marker: Position of current Pixel inside Font Bounding Box */
    Coord2_t LocalCoord = {0, 0};

    if( GlyphBB.w > 0 )
    {
        /* decode glyph */
        uint8_t bits_per_0 = u8g2_font_getNumberOfBitsRleZeros(font);
        uint8_t bits_per_1 = u8g2_font_getNumberOfBitsRleOnes(font);

        do
        {
            uint8_t rlZeros = u8g2_font_decode_get_unsigned_bits(&Decoder, bits_per_0);
            uint8_t rlOnes  = u8g2_font_decode_get_unsigned_bits(&Decoder, bits_per_1);
            do
            {
                u8g2_bitmap_set_line_element(&Bitmap, &LocalCoord, GlyphBB, rlZeros, 0);
                u8g2_bitmap_set_line_element(&Bitmap, &LocalCoord, GlyphBB, rlOnes, 1);
            }
            while( u8g2_font_decode_get_unsigned_bits(&Decoder, 1) != 0 ); /* Check padding Bit */
        }
        while( LocalCoord.y < GlyphBB.h );
    }
}
