/** @file u8g2_font_find_glyph.c

  Functions for finding glyph data of unicode characters in U8G2 font array



  This Code is derived of

  U8g2 for Adafruit GFX Lib (https://github.com/olikraus/U8g2_for_Adafruit_GFX)

  Copyright (c) 2018, olikraus@gmail.com
  All rights reserved.

  Redistribution and use in source and binary forms, with or without modification,
  are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list
    of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright notice, this
    list of conditions and the following disclaimer in the documentation and/or other
    materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


  $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
  $Revision: 12 $

  @date 29.03.2018 @author Nicolas */




#include "u8g2_font_find_glyph.h"
#include "u8g2_font_header_private.h"


typedef struct GlyphPosition_s
{
    const uint8_t *ptr;
    uint16_t unicode;
}
GlyphPosition_t;



/** Find the starting point of the glyph data. If glyph is found, its unicode and a
    pointer to its data will be return.
    If no glyph is found, unicode and pointer to the next glyph in the 8 bit or 16 bit
    section are returned.
  Args:
    encoding: Encoding (ASCII or Unicode) of the glyph
  Return:
    Address and unicode of the glyph data or NULL, if the encoding is not available in the
    font.

  Glyph data starts at byte-offset 23 in font data array.
  It contains:
   - unicode of character/glyph (8 bit for unicode <= 255, 16 bit for unicode > 255)
   - jump-offset to next character (or glyph size)
   - glyph width (compressed)
   - glyph height (compressed)
   - glyph x-offset (compressed)
   - glyph y-offset (compressed)
   - pixel data for each glyph (encoded)

 For each glyph, unicode will be compared und jumped until next unicode until the glyph
 or termination byte (zero byte) has been found.

 For the important code ranges in 0...'A', 'A'...'a' and 'a'...255 jump-shortcuts are
 implemented.

 Run-length for RLE is stored in font header.
 *
 * @param[in] font: Font data array
 * @param[in] charUnicode: 16-Bit Unicodes des Glyphen
 * @return Zeiger auf gefundenen Glyphen, Unicode des gefundenen Glyphen */
GlyphPosition_t u8g2_font_findCharacterGlyphData(const uint8_t *font, uint16_t charUnicode)
{
    assert( font != NULL );
    GlyphPosition_t G = {.ptr = NULL, .unicode = 0 };
    const uint8_t *font0 = font;
    uint16_t glyphUnicode = 0;
    font += 23;

    /* Look for glyph either in 8 bit or 16 bit section */
    if ( charUnicode <= 255 )
    {
        /* shortcuts to specific ranges */
        if ( charUnicode >= 'a' )
        {
            font += u8g2_font_getStartPosLowerA(font0);
        }
        else if ( charUnicode >= 'A' )
        {
            font += u8g2_font_getStartPosUpperA(font0);
        }

        /* find glyph in 8 bit section */
        for(;;)
        {
            if ( pgm_read_byte( font + 1 ) == 0 )
            {
                /* End of 8 bit range without finding */
                break;
            }
            else if( (glyphUnicode = pgm_read_byte( font )) >= charUnicode )
            {
                /* Glyph found or passed over */
                G.unicode = glyphUnicode;
                G.ptr = font+2;
                break;
            }
            /* jump to next glyph */
            font += pgm_read_byte( font + 1 );
        }
    }
    else
    {
        font += u8g2_font_getStartPosUnicode(font0);

        /* find glyph in 16 bit section */
        for(;;)
        {
            G.unicode = pgm_read_byte( font );
            G.unicode <<= 8;
            G.unicode |= pgm_read_byte( font + 1 );
            if( G.unicode == 0 )
            {
                /* End of 16 bit range without finding */
                break;
            }
            else if( G.unicode >= charUnicode )
            {
                /* Glyph found or passed over */
                G.ptr = font+3;
                break;
            }

            /* jump to next glyph */
            font += pgm_read_byte( font + 2 );
        }
    }
    return G;
}



/** Find glyph array data of unicode. If glyph for character is not found, NULL pointer
 * will be returned
 *
 * @param[in] font: Font-Array
 * @param[in] unicode: 16-Bit-Unicode des Glyphen
 * @return Adresse des Glyphen oder Nullzeiger */
const uint8_t * u8g2_font_findCharacterGlyphPtr(const uint8_t *font, uint16_t unicode)
{
    GlyphPosition_t G = u8g2_font_findCharacterGlyphData(font, unicode);
    if( G.unicode != unicode )
        return NULL;
    else
        return G.ptr;
}



/** Find unicode of next available character/glyph.
 *
 * If no following character exists, return value = 0
 *
 * @param[in] font: Font-Array
 * @param[in] unicode: 16-Bit-Unicode des Glyphen
 * @return 16-Bit-Unicode des nachfolgenden Glyphen oder 0 */
uint16_t u8g2_font_findNextChar(const uint8_t *font, uint16_t unicode)
{
    assert( font != NULL );

    GlyphPosition_t G = u8g2_font_findCharacterGlyphData(font, unicode+1);
    if( G.ptr )
    {
        /* Found next character */
        return G.unicode;
    }
    else if( unicode > 255 )
    {
        /* Searched glyphs in 16 bit region already - there will be no subsequent
           character */
        return 0;
    }
    else
    {
        /* Searched glyph if 8 bit region before and did not find - the first character
           in 16 bit region (if exists) is what we are looking for */
        G = u8g2_font_findCharacterGlyphData(font, 256);
        return G.unicode;
    }
}
