/** @file u8g2_font_header_private.h
 * @ingroup Font_Subsystem
 *
 * Helper functions to read font header
 *
 * Since font header data is uncompressed, direct read of font array causes very little overhead
 *
 *
 *  This Code re-implements elements of
 *
 *  U8g2 for Adafruit GFX Lib (https://github.com/olikraus/U8g2_for_Adafruit_GFX)
 *
 *  Copyright (c) 2018, olikraus@gmail.com
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without modification,
 *  are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list
 *    of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright notice, this
 *    list of conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *  $Date: 2020-06-17 11:20:09 +0200 (Mi, 17 Jun 2020) $
 *  $Revision: 12 $
 *
 *  @date 25.03.2018 @author Nicolas */



#ifndef U8G2_FONT_HEADER_PRIVATE_H_INCLUDED
    #define U8G2_FONT_HEADER_PRIVATE_H_INCLUDED

    #include <stdint.h>
    #include "pgmspace.h"
    #include "rect.h"



    /** Font header data offsets:
       All Font header data is uncompressed */
    enum u8g2_font_header_offset_e
    {
        u8g2_font_header_offset_number_of_glyphs = 0,      /**< Anzahl der Zeichen im Zeichensatz */
        u8g2_font_header_offset_bounding_box_mode = 1,     /**< Bounding-Box-Modus
                                                                (Enum-Wert) */
        u8g2_font_header_offset_bits_per_0 = 2,            /**< Encoding run lenght (Zeros) */
        u8g2_font_header_offset_bits_per_1 = 3,            /**< Encoding run lenght (Ones) */

        /* Encoding lengths of glyph bounding box components */
        u8g2_font_header_offset_bits_per_char_width = 4,  /**< Bits pro Glyph-BB-Breite */
        u8g2_font_header_offset_bits_per_char_height = 5, /**< Bits pro Glyph-BB-Hoehe */
        u8g2_font_header_offset_bits_per_char_x = 6,      /**< Bits pro Glyph-BB-X-Offset */
        u8g2_font_header_offset_bits_per_char_y = 7,      /**< Bits pro Glyph-BB-Y-Offset */
        u8g2_font_header_offset_bits_per_delta_x = 8,     /**< Bits pro Glyph-Zeichenvorschub */

        /* Font Bounding box; bytes are int8_t */
        u8g2_font_header_offset_max_char_width = 9,       /**< Font-bounding-box-Breite (int8) */
        u8g2_font_header_offset_max_char_height = 10,     /**< Font-bounding-box-Hoehe (int8) */
        u8g2_font_header_offset_x_offset = 11,            /**< Font-bounding-box-X-Offset (int8) */
        u8g2_font_header_offset_y_offset = 12,            /**< Font-bounding-box-Y-Offset (int8) */

        /**@{
           Ascent and Descent of selected characters for specific calculations of line pitch.
           Characters are 'A', 'g' and '('; bytes are int8_t */
        u8g2_font_header_offset_ascent_A = 13,            /**< Oberlaenge 'A' (int8) */
        u8g2_font_header_offset_descent_g = 14,           /**< Unterlaenge 'g' (int8) */
        u8g2_font_header_offset_ascent_parenthesis = 15,  /**< Oberlaenge '(' (int8) */
        u8g2_font_header_offset_descent_parenthesis = 16, /**< Unterlaenge '(' (int8) */
        /**@}*/

        /* Start Posistion (Offset) of specific Glyphs ('A', 'a', 255) in Font Array.
           This accelerates search for the correct glyph. Would still work, if all
           zeroes, thus slower */
        u8g2_font_header_offset_start_pos_upper_A = 17,   /**< Glyphs starting from 'A' (uint16) */
        u8g2_font_header_offset_start_pos_lower_a = 19,   /**< Glyphs starting from 'a' (uint16) */
        u8g2_font_header_offset_start_pos_unicode = 21,   /**< Glyphs starting from 255 (uint16) */
    };



    /** Private helper function: Assemble 16 Bit Word from two 8 Bit reads (Big Endian)
     * @param[in] addr: Adresse des oberen Bytes
     * @return Wort */
    static inline uint16_t u8g2_helper_getFontHeaderWord(const uint8_t *addr)
    {
        uint16_t out = 0;
        out |= ( (uint16_t) pgm_read_byte(addr++) ) << 8;
        out |= (uint16_t) pgm_read_byte(addr);
        return out;
    }



    /** Feld auslesen: Anzahl der Zeichen
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfGlyphs(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_number_of_glyphs);
    }



    /** Feld auslesen: Bitbreite Nullen RLE
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsRleZeros(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_0);
    }



    /** Feld auslesen: Bitbreite Einsen RLE
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsRleOnes(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_1);
    }



    /** Feld auslesen: Bitbreite Bounding-Box-Breite
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsCharWidth(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_char_width);
    }



    /** Feld auslesen: Bitbreite Bounding-Box-Hoehe
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsCharHeight(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_char_height);
    }



    /** Feld auslesen: Bitbreite Bounding-Box-Position X
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsCharX(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_char_x);
    }



    /** Feld auslesen: Bitbreite Bounding-Box-Position Y
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsCharY(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_char_y);
    }



    /** Feld auslesen: Bitbreite Zeichenvorschub
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Feldinhalt */
    static inline uint8_t u8g2_font_getNumberOfBitsDeltaX(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_bits_per_delta_x);
    }



    /** Glyph-Header-Groesse / Offset Bitmap im Glyph
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return offset */
    static inline uint8_t u8g2_font_getNumberOfBitsGlyphHeader(const uint8_t *font)
    {
        uint8_t n=0;

        n += u8g2_font_getNumberOfBitsCharWidth(font);
        n += u8g2_font_getNumberOfBitsCharHeight(font);
        n += u8g2_font_getNumberOfBitsCharX(font);
        n += u8g2_font_getNumberOfBitsCharY(font);
        n += u8g2_font_getNumberOfBitsDeltaX(font);

        return n;
    }



    /** Zeichensatz-Bounding-Box
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Bounding box */
    static inline Rect8_t u8g2_font_getBoundingBox(const uint8_t *font)
    {
        Rect8_t Bbx;
        Bbx.x =  (int8_t) pgm_read_byte(font + u8g2_font_header_offset_x_offset);
        Bbx.y =  (int8_t) pgm_read_byte(font + u8g2_font_header_offset_y_offset);
        Bbx.w =  (uint8_t) pgm_read_byte(font + u8g2_font_header_offset_max_char_width);
        Bbx.h =  (uint8_t) pgm_read_byte(font + u8g2_font_header_offset_max_char_height);
        return Bbx;
    }



    /** Hoehe des Grossbuchstabens 'A'
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return Zeichenhoehe */
    static inline uint8_t u8g2_font_getCapitalAAscent(const uint8_t *font)
    {
        return pgm_read_byte(font + u8g2_font_header_offset_ascent_A);
    }



    /** Offset des Grossbuchstabens 'A' in der Glyph-Tabelle
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return offset */
    static inline uint16_t u8g2_font_getStartPosUpperA(const uint8_t *font)
    {
        font += u8g2_font_header_offset_start_pos_upper_A;
        return u8g2_helper_getFontHeaderWord(font);
    }



    /** Offset des Kleinbuchstabens 'a' in der Glyph-Tabelle
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return offset */
    static inline uint16_t u8g2_font_getStartPosLowerA(const uint8_t *font)
    {
        font += u8g2_font_header_offset_start_pos_lower_a;
        return u8g2_helper_getFontHeaderWord(font);
    }



    /** Offset des ersten nicht-ASCII-Uniocodezeichens in der Glyph-Tabelle
     * @param[in] font: Zeiger auf Zeichensatzanfang
     * @return offset */
    static inline uint16_t u8g2_font_getStartPosUnicode(const uint8_t *font)
    {
        font += u8g2_font_header_offset_start_pos_unicode;
        return u8g2_helper_getFontHeaderWord(font);
    }

#endif
